<?php

/**
 * drebedengi.ru web-service description; In this document term "waste" means expense. Please, don't forget process 'SoapFault' messages while programming your interface!
 *
 */
class Soap
{
	public static $isDebug     = false;
	public static $debugInfo   = array();
	public static $lastMsgList = array();
	private       $pushErrors  = array();

	private function login($apiId, $login, $pass)
	{
		/**
		 * Базовые проверки
		 */
		$login = mb_strtolower(trim($login));
		$pass  = trim($pass);
		$apiId = trim($apiId);

		if (mb_strlen($login) < 2 || mb_strlen($pass) < 2 || mb_strlen($apiId) < 2 || mb_strlen($login) > 64 || mb_strlen($pass) > 64 || mb_strlen($apiId) > 32) {
			/**
			 * Пустой пароль частенько бывает от мобильных приложений по причине в них бага.
			 * Считаем это обычным кейсом не верного пароля. Так хоть возможно у юзера запросят
			 * пароль заново.
			 */
			if (empty($pass)) {
				throw new SoapFault('1', 'Login failed');
			}
			throw new SoapFault('4', 'Login failed');
		}

		$authArray = Internal::soapLogin($apiId, $login, $pass);

		/**
		 * Всё ок. Проверяем авторизацию.
		 */
		$budgetApi = BudgetApi::create(null, $apiId);
		if ($budgetApi->get(BudgetApi::NAME) != $apiId || ($apiId == 'demo_api' && $login != 'demo@example.com')) {
			throw new SoapFault('50', "Login failed. API ID '$apiId' is invalid");
		}
		$query = "SELECT * FROM users WHERE trim(mail) = $1".User::passSql();
		$authArray ? '' : $authArray = DAO::fillArray($query, null, array($login, $pass));

		/**
		 * Если не залогинились и такого юзера нет, пытаемся найти и восстановить из архива
		 */
		if (count($authArray) == 0 && DAO::getCount("SELECT id FROM users WHERE trim(mail) = $1", true, array($login)) == null) {
			User::restore(null, $login) ? $authArray = DAO::fillArray($query, null, array($login, $pass)) : '';
		}

		if (sizeof($authArray) == 1) {
			/**
			 * Стопорим синхронизацию в некоторых случаях, когда надо выкачать всё заново:
			 * - Чистка всего в разделе "Начать всё сначала", если удалено слишком много данных
			 * - Чистка всего после импорта вин32
			 * - отключение-подключение-изменение прав пользователя
			 */

			User::lastDate($authArray[0]['id']);
			Session::set(User::ID, $authArray[0]['id']);
			Session::set(Constant::API_AUTHOR_ID, $budgetApi->get(BudgetApi::USER_ID));
			Session::set(Constant::API_ID, $budgetApi->get(BudgetApi::ID));
			Session::set(User::FAMILY_ID, $authArray[0]['family_id']);
			Session::set(User::ACC_LEVEL, $authArray[0]['acc_level']);
			Session::set(User::NICK, $authArray[0]['nick']);
			Session::set(Constant::IS_PREMIUM, true); // Чтобы вытягивалсь записи всех пользователей.

			User::create(User::id())->setMultiple();
			User::setDbId();
			return;
		}

		throw new SoapFault('1', 'Login failed');
	}

	private function checkEmptyArray($value, $faultCode, $elemName, $isCheckArray = true)
	{
		if (empty($value)) {
//			Debug::mail(array($value, $elemName));
			throw new SoapFault($faultCode, "Value of '$elemName' is empty");
		}
		if ($isCheckArray && !is_array($value)) {
			Debug::mail(array($value, $elemName));
			throw new SoapFault($faultCode, "'$elemName' = '$value' is not an array");
		}
	}

	private function checkIdList($idList, $objName)
	{
		if (!empty($idList)) {
			if (!is_array($idList)) {
				throw new SoapFault('86', "$objName: Parameter 'idList' must be an array");
			}
			foreach ($idList as $id) {
				$this->checkNumeric($id, $objName);
			}
			return true;
		}
		return false;
	}

	private function checkNumeric($value, $objName, $isNullAllowed = false)
	{
		if (!$value && $isNullAllowed) {
			return $value;
		}
		if (!is_numeric($value)) {
			throw new SoapFault('45', "$objName: '$value' is not a number");
		}
		return $value;
	}

	private function normalizeString($value, $maxLength, $objName)
	{
		$curLength = mb_strlen($value, 'UTF-8');
		if ($curLength > $maxLength) {
			throw new SoapFault('46', "$objName: Value length($curLength) must be less than $maxLength");
		}
		return $value;
	}

	/**
	 * Проверяем, что все элементы объекта заданы и базово валидны. Если есть лишние поля - просто идут лесом.
	 */
	private function checkObject($tplObject, $givenArray, $objName, $isNullAllowed = false)
	{
		$this->checkEmptyArray($givenArray, '40', "$objName element");
		$fields = $tplObject->getDaoArray();
		foreach ($fields as $name => $field) {
			if ($name == Business::ID && empty($givenArray[Business::ID])) {
				continue;
			}
			/**
			 * Костыль для budget_record.group_id, который был недавно добавлен
			 */
			if ($objName == 'Record' && $name == 'group_id' && !array_key_exists("$name", $givenArray)) {
				$givenArray['group_id'] = '';
				$isNullAllowed          = true;
			}
			if (!array_key_exists("$name", $givenArray)) {
				throw new SoapFault('41', "Element '$name' is not present in $objName array");
			}
			// Поля оканчивающиеся на _id - должны быть числовыми.
			if ((mb_strlen($name) > 2 && mb_strtolower(mb_substr($name, -3)) == '_id' && $givenArray[$name]) || $name == Business::ID) {
				$this->checkNumeric($givenArray[$name], $objName, $isNullAllowed);
			}
			// Поля начинающиеся на is_ должны быть булями
			if (mb_strlen($name) > 2 && mb_strtolower(mb_substr($name, 0, 3)) == 'is_') {
				$value = $givenArray[$name];
				if (!is_bool($value)) {
					throw new SoapFault('43', "Element '$name' = '$value' of $objName is not a boolean");
				}
			}
			$tplObject->set($name, $givenArray[$name]);
		}
	}

	private function setObjectList($apiId, $login, $pass, $list, $type)
	{
		/**
		 * Для функций записи должен быть активен Premium
		 */
		try {
			$this->getAccessStatus($apiId, $login, $pass);
		}
		catch (SoapFault $e) {
			if ($e->faultcode != '87') { // trial
				throw $e;
			}
		}

		$objName = 'Waste category';
		switch ($type) {
			case v2_Budget::TYPE_MOVE :
				$objName = 'Place';
				break;
			case v2_Budget::TYPE_INCOME:
				$objName = 'Income source';
				break;
		}
		/**
		 * Проверки
		 */
		$this->checkEmptyArray($list, '38', $objName);

		try {
			DB::transactionBegin();
			$familyId     = Session::get(User::FAMILY_ID);
			$retArray     = array();
			$dublDebug    = array();
			$lastSyncData = DAO::fillArrayKeyId("
				SELECT client_id AS id, server_id, app_instance_id FROM budget_sync_last
				WHERE user_id = $1 AND api_id = $2 AND object_type = $3",
				array(User::id(), Session::get(Constant::API_ID), $type)
			);
			$appId        = self::getAppId();

			foreach ($list as $category) {
				/**
				 * Первым делом, если это добавление клиентской записи, которая
				 * уже была добавлена в прошлой синхре, то значит был сбой на клиенте
				 * и он шлёт дубль. Ничего не добавляем, а отдаём тот же server_id
				 * что и в прошлый раз.
				 */
				$isClientIdThisInstance = true;
				if (
					abs($appId) > 0 &&
					!empty($category['client_id']) &&
					isset($lastSyncData[$category['client_id']]) &&
					abs($lastSyncData[$category['client_id']]['app_instance_id']) > 0 &&
					$lastSyncData[$category['client_id']]['app_instance_id'] != $appId
				) {
					$isClientIdThisInstance = false;
				}

				if (!empty($category['client_id']) && isset($lastSyncData[$category['client_id']]) && $isClientIdThisInstance) {
					$retArray[]  = array(
						'server_id' => $lastSyncData[$category['client_id']]['server_id'],
						'client_id' => $category['client_id'],
						'status'    => 'inserted'
					);
					$dublDebug[] = array(
						'server_id' => $lastSyncData[$category['client_id']]['server_id'],
						'client_id' => $category['client_id'],
						'type'      => 'category',
					);
					continue;
				}

				$category[v2_BudgetObject::FAMILY_ID]      = $familyId;
				$category[v2_BudgetObject::TYPE]           = $type;
				$category[v2_BudgetObject::IS_CREDIT_CARD] = false;
				if ($type != v2_Budget::TYPE_MOVE) {
					$category[v2_BudgetObject::ICON_ID] = null;
				}
				if ($type != v2_Budget::TYPE_WASTE && $type != v2_Budget::TYPE_INCOME && empty($category[v2_BudgetObject::PARENT_ID])) {
					$category[v2_BudgetObject::PARENT_ID] = -1;
				}

				$categoryId = empty($category['server_id']) ? '' : $category['server_id'];

				isset($category[v2_BudgetObject::DESCRIPTION]) ? '' : $category[v2_BudgetObject::DESCRIPTION] = null;

				/**
				 * Если галки авто-скрытия долгов нет, то это только старые версии
				 * в которых могут приходить только долговые счета, у которых по умолчанию
				 * эта галка всегда включена.
				 * Так же, не выставляем умолчаний для случая редактирования существующего
				 * счёта, галка должна быть такой какой выставлена до этого.
				 */
				isset($category[v2_BudgetObject::IS_AUTOHIDE]) || $categoryId ?
					'' :
					$category[v2_BudgetObject::IS_AUTOHIDE] = true;

				if ($categoryId) {
					$isAutoHide = DAO::getCount("SELECT is_autohide FROM budget_object WHERE id = $1 AND type = $2", true, array($categoryId, $type));
					if (!$isAutoHide) {
						throw new SoapFault('54', "Type = $type is invalid for object id = $categoryId");
					}
					$category[v2_BudgetObject::IS_AUTOHIDE] = $isAutoHide == 't' ? true : false;
				}

				/**
				 * Базовые проверки
				 */
				if (empty($category['server_id']) && empty($category['client_id'])) {
					throw new SoapFault('75', "'server_id' or 'client_id' should be given");
				}
				empty($category['purse_of_nuid']) ? $category['purse_of_nuid'] = null : '';
				$category[v2_BudgetObject::ID] = $categoryId ? $categoryId : null;
				$categoryObj                   = v2_BudgetObject::create();
				$this->checkObject($categoryObj, $category, $objName);

				/**
				 * Какой-то хитрый баг в iPhone приложении. Иногда название долга пустое
				 * и на этом всё ступорится. Заполняем пустое место "хоть как-нибудь", чтобы
				 * процесс прошёл. 23 фев 2013.
				 */
				if (empty($category[v2_BudgetObject::NAME])) {
					$category[v2_BudgetObject::NAME] = 'Noname';
				}

				$categoryObj->
				set(v2_BudgetObject::SORT, $this->checkNumeric($category[v2_BudgetObject::SORT], $objName))->
				set(v2_BudgetObject::NAME, $this->normalizeString($category[v2_BudgetObject::NAME], 128, $objName))->
				daoSave()->
				toServerHistory($categoryId ? 'update' : 'insert');

				$retArrayItem = array(
					'server_id' => $categoryObj->get(v2_BudgetObject::ID),
					'status'    => $categoryId ? 'updated' : 'inserted'
				);
				if (!$categoryId) {
					$retArrayItem['client_id'] = $category['client_id'];
				}
				$retArray[] = $retArrayItem;
			}
			/**
			 * Перед тем, как успешно отправить список связок, надо грохнуть все
			 * клиентско-серверные связки данного типа, от прошлой синхры и добавить текущие.
			 */
			$delAppIdSql = $appId ? "AND app_instance_id = ".DB::esc($appId) : '';
			DB::query("
				DELETE FROM budget_sync_last
	    		WHERE user_id = $1 AND api_id = $2 AND object_type = $3 $delAppIdSql",
				array(User::id(), Session::get(Constant::API_ID), $type)
			);
			foreach ($retArray as $retValue) {
				if ($retValue['status'] != 'inserted') {
					continue;
				}
				BudgetSyncLast::create()->
				set(BudgetSyncLast::USER_ID, User::id())->
				set(BudgetSyncLast::API_ID, Session::get(Constant::API_ID))->
				set(BudgetSyncLast::OBJECT_TYPE, $type)->
				set(BudgetSyncLast::CLIENT_ID, $retValue['client_id'])->
				set(BudgetSyncLast::SERVER_ID, $retValue['server_id'])->
				set(BudgetSyncLast::INSTANCE_ID, $appId)->
				set(BudgetSyncLast::DATE, 'now()')->
				daoSave();
			}
			DB::transactionCommit();
			/**
			 * Для отладки шлём уведомления о дублях.
			 * Когда станет не нужным - убрать.
			 */
			if ($dublDebug) {
				Debug::mail($dublDebug);
			}
		}
		catch (Exception $e) {
			DB::transactionRollback();
			if ('Duty not allowed' == $e->getMessage()) {
				$accName = v2_BudgetObject::create($e->getCode())->get(v2_BudgetObject::NAME);
				$msg     = "Нет доступа к счёту '$accName'\nПопросите главного пользователя дать доступ.";
				throw new SoapFault('188', $msg);
			}
			$conflictMsg = 'Для восстановления синхронизации, ';
			$conflictMsg .= ($apiId == 'vandr_api' ? "необходимо перезайти через меню\n'Ещё->Сменить пользователя'." : 'приложение необходимо переустановить.');
			if (mb_strpos($e->getMessage(), 'create_only_access') !== false) {
				throw new SoapFault('188', "Настройки прав доступа не позволяют изменять операции\n".$conflictMsg);
			}
			$accessErrorMsg = array('forbidden_type_', 'el_forbidden', 'not_owner_obj', 'readonly_access', 'not_owner_rec');
			foreach ($accessErrorMsg as $errorMsg) {
				if (strpos($e->getMessage(), $errorMsg) !== false) {
					throw new SoapFault('188', "Доступ к скрытым элементам запрещён\n".$conflictMsg);
				}
			}
			self::checkNameUniq($e->getMessage());
			Debug::mail($e);
			throw new SoapFault('48', $e->getMessage());
		}
		return $retArray;
	}

	/**
	 * Retrievs currency list (array of arrays) with codes and courses: [id] => Internal currency ID; [name] => Currency name, given by user; [course] => current course from sbrf(dot)ru; [code] => International currency code (for course autoupdating); [family_id] => User family ID (for multiuser mode); [is_default] => is default currency# There should be only one default currency; [is_autoupdate] => autoupdate course once per day, from sbrf(dot)ru; [is_hidden] => is currency hidden in user interface; If parameter [idList] is given, it will be treat as ID list of objects to retrieve# this is used for synchronization;
	 *
	 * @param string $apiId
	 * @param string $login
	 * @param string $pass
	 * @param array $idList
	 * @return array
	 */
	public function getCurrencyList($apiId, $login, $pass, $idList)
	{
		$this->login($apiId, $login, $pass);
		/**
		 * Если это получение списка объектов по ID - обрабатываем первым делом
		 */
		$sqlIdList = '';
		if ($this->checkIdList($idList, 'getCurrencyList')) {
			$sqlIdList = 'id IN ('.DB::in($idList).') AND';
		}
		$list = DAO::fillArray("SELECT * FROM currency WHERE $sqlIdList family_id = $1", null, array(User::familyId()));
		foreach ($list as $key => $item) {
			$list[$key]['code'] = FormUtil::toHtml($item['code']);
			$list[$key]['name'] = FormUtil::toHtml($item['name']);
		}
		return $list;
	}

	/**
	 * Retrievs waste category list (array of arrays): [id] => Internal category ID; [parent_id] => For tree structure; [budget_family_id] => User family ID (for multiuser mode); [type] => Type of object, 3 - waste category; [name] => Category name given by user; [is_hidden] => is category hidden in user interface; [sort] => User sort of category tree; If parameter [idList] is given, it will be treat as ID list of objects to retrieve# this is used for synchronization;
	 *
	 * @param string $apiId
	 * @param string $login
	 * @param string $pass
	 * @param array $idList
	 * @return array
	 */
	public function getCategoryList($apiId, $login, $pass, $idList)
	{
		$this->login($apiId, $login, $pass);
		/**
		 * Если это получение списка объектов по ID - обрабатываем первым делом
		 */
		$sqlIdList = '';
		if ($this->checkIdList($idList, 'getCategoryList')) {
			$sqlIdList = 'id IN ('.DB::in($idList).') AND';
		}
		$deniedList = defined('IS_SERVER') ? array() : Family::getDeniedElems();
		$sqlLimited = Family::getLimitedSql('budget_object.id', v2_Budget::TYPE_WASTE);
		$list       = DAO::fillArray("
			SELECT id,parent_id,budget_family_id,type,name,is_hidden,sort,description
        	FROM budget_object WHERE $sqlIdList type = $1 AND budget_family_id = $2 $sqlLimited ORDER BY parent_id, sort",
			null, array(v2_Budget::TYPE_WASTE, User::familyId())
		);
		foreach ($list as $key => $item) {
			if (isset($deniedList[$item['parent_id']])) {
				$list[$key]['parent_id'] = "-1";
			}
			$list[$key]['description'] = FormUtil::toHtml($item['description']);
			$list[$key]['name']        = FormUtil::toHtml($item['name']);
		}
		return $list;
	}

	/**
	 * Retrievs income source list (array of arrays): [id] => Internal source ID; [parent_id] => For tree structure; [budget_family_id] => User family ID (for multiuser mode); [type] => Type of object, 2 - income source; [name] => Source name given by user; [is_hidden] => is income hidden in user interface; [sort] => User sort of source tree; If parameter [idList] is given, it will be treat as ID list of objects to retrieve# this is used for synchronization;
	 *
	 * @param string $apiId
	 * @param string $login
	 * @param string $pass
	 * @param array $idList
	 * @return array
	 */
	public function getSourceList($apiId, $login, $pass, $idList)
	{
		$this->login($apiId, $login, $pass);
		/**
		 * Если это получение списка объектов по ID - обрабатываем первым делом
		 */
		$sqlIdList = '';
		if ($this->checkIdList($idList, 'getSourceList')) {
			$sqlIdList = 'id IN ('.DB::in($idList).') AND';
		}
		$deniedList = defined('IS_SERVER') ? array() : Family::getDeniedElems();
		$sqlLimited = Family::getLimitedSql('budget_object.id', v2_Budget::TYPE_INCOME);
		$list       = DAO::fillArray("
			SELECT id,parent_id,budget_family_id,type,name,is_hidden,sort,description
        	FROM budget_object WHERE $sqlIdList type = $1 AND budget_family_id = $2 $sqlLimited ORDER BY sort",
			null, array(v2_Budget::TYPE_INCOME, User::familyId())
		);
		foreach ($list as $key => $item) {
			if (isset($deniedList[$item['parent_id']])) {
				$list[$key]['parent_id'] = "-1";
			}
			$list[$key]['description'] = FormUtil::toHtml($item['description']);
			$list[$key]['name']        = FormUtil::toHtml($item['name']);
		}
		return $list;
	}

	/**
	 * Retrievs place list (array of arrays): [id] => Internal place ID; [budget_family_id] => User family ID (for multiuser mode); [type] => Type of object, 4 - places; [name] => Place name given by user; [is_hidden] => is place hidden in user interface; [is_autohide] => debts will auto hide on null balance; [is_for_duty] => Internal place for duty logic, Auto created while user adds "Waste or income duty"; [sort] => User sort of place list; [purse_of_nuid] => Not empty if place is own purse of user# The value is internal user ID; [icon_id] => Place icon ID from http://www(dot)drebedengi(dot)ru/img/pl[icon_id](dot)gif; If parameter [idList] is given, it will be treat as ID list of objects to retrieve# this is used for synchronization; There is may be empty response, if user access level is limited;
	 *
	 * @param string $apiId
	 * @param string $login
	 * @param string $pass
	 * @param array $idList
	 * @return array
	 */
	public function getPlaceList($apiId, $login, $pass, $idList)
	{
		$this->login($apiId, $login, $pass);
		/**
		 * Если это получение списка объектов по ID - обрабатываем первым делом
		 */
		$sqlIdList = '';
		if ($this->checkIdList($idList, 'getPlaceList')) {
			$sqlIdList = 'id IN ('.DB::in($idList).') AND';
		}

		$parArray = array(v2_Budget::TYPE_MOVE, User::familyId());
		/**
		 * С версии 2.3.14 в Андроид появилась поддержка групп счетов.
		 */
		$groupSql = 'type = $1';
		if ('vandr_api' == $apiId && self::isMoreVersion(array(2, 3, 14))) {
			$groupSql   = '(type = $1 OR type = $3)';
			$parArray[] = v2_Budget::TYPE_GROUP_PLACE;
		}

		/**
		 * Права доступа
		 */
		$deniedList = defined('IS_SERVER') ? array() : Family::getDeniedElems();
		$sqlLimited = Family::getLimitedSql('budget_object.id');
		$query      = "
			SELECT id,budget_family_id,type,name,is_hidden,is_for_duty,sort,purse_of_nuid,icon_id, is_autohide, parent_id,
        	description, is_credit_card FROM budget_object WHERE $sqlIdList $groupSql $sqlLimited AND budget_family_id = $2 
        	ORDER BY sort
		";
		$list       = DAO::fillArray($query, null, $parArray);
		foreach ($list as $key => $item) {
			if (isset($deniedList[$item['parent_id']])) {
				$list[$key]['parent_id'] = "-1";
			}
			$list[$key]['description'] = FormUtil::toHtml($item['description']);
			$list[$key]['name']        = FormUtil::toHtml($item['name']);
		}
		return $list;
	}

	/**
	 * Retrievs tag list (array of arrays): [id] => Internal tag ID; [family_id] => User family ID (for multiuser mode); [name] => Tag name given by user; [is_hidden] => is tag hidden in user interface; [is_family] => is tag visible for all family user, or user only; [sort] => User sort of tag list; [parent_id] => For tree view; If parameter [idList] is given, it will be treat as ID list of objects to retrieve# this is used for synchronization;
	 *
	 * @param string $apiId
	 * @param string $login
	 * @param string $pass
	 * @param array $idList
	 * @return array
	 */
	public function getTagList($apiId, $login, $pass, $idList)
	{
		$this->login($apiId, $login, $pass);
		/**
		 * Если это получение списка объектов по ID - обрабатываем первым делом
		 */
		$sqlIdList = '';
		if ($this->checkIdList($idList, 'getTagList')) {
			$sqlIdList = 'id IN ('.DB::in($idList).') AND';
		}
		return DAO::fillArray("
			SELECT * FROM budget_tags WHERE $sqlIdList
	        (
	        	(user_id = $1 AND family_id = $2) OR
	        	(is_family IS TRUE AND family_id = $2)
	        )
	        ORDER BY sort",
			null, array(User::id(), User::familyId())
		);
	}

	/**
	 * Retrievs record list (array of arrays) or report table by parameters; [params] => array of following parameters: 'is_report' [true|false (no default)] - retrievs data for report only or full records (waste, incomes, moves, changes) for export; 'relative_date' [YYYY-MM-DD (NOW by default)] - all data will be retrieved relative to this value, according to 'r_period' value; 'period_to', 'period_from' [YYYY-MM-DD] - custom period, if 'r_period' = 0; 'is_show_duty' [true(default)|false] - whether or not include duty record; 'r_period' [custom period = 0, this month = 1, today = 7, last month = 2, this quart = 3, this year = 4, last year = 5, all time = 6, last 20 record = 8 (default)] - period for which data will be obtained; 'r_what' [income = 2, waste = 3 (default), move = 4, change = 5, all types = 6] - type of data you want to get; 'r_who' [0 (default) - all users, int8 = user ID] - The data of the user to obtain, in the case of multiplayer mode; 'r_how' [show record list by detail = 1 (default), group incomes by source = 2, group wastes by category = 3] - Values 2 and 3 are for 'report' mode only# How to group the result record list; 'r_middle' [No average = 0 (default), Average monthly = 2592000, Average weekly = 604800, Averaged over days = 86400] - How to average the data, if r_how = 2 or 3; 'r_currency' [Original currency = 0 (default), int8 = currency ID] - Convert or not in to given currency; 'r_is_place', 'r_is_tag', 'r_is_category' [Include all = 0 (default), Include only selected = 1, All except selected = 2] - Exclude or include 'r_place', 'r_tag' or 'r_category' respectively; 'r_place', 'r_tag', 'r_category' [Array] - Array of numeric values for place ID, tag ID or category ID respectively; 'is_with_planned' [true|false(default)] - whether or not include planned record; If parameter [idList] is given, it will be treat as ID list of objects to retrieve# this is used for synchronization;
	 *
	 * @param string $apiId
	 * @param string $login
	 * @param string $pass
	 * @param array $params
	 * @param array $idList
	 * @return array
	 */
	public function getRecordList($apiId, $login, $pass, $params, $idList)
	{
		ini_set('memory_limit', '900M');

		$this->login($apiId, $login, $pass);

		// Массив разрешённых параметров, он же задаёт умолчания
		$allowedParams = array(
			'relative_date'   => null, // сегодня
			'period_to'       => 0, // последние 20
			'period_from'     => 0, // последние 20
			'is_show_duty'    => true, // показывать долги
			'is_report'       => null, // умолчания нет, параметр обязательный
			'r_period'        => v2_Budget::PERIOD_LAST_20, // последние 20
			'r_what'          => v2_Budget::TYPE_WASTE, // показывать расходы
			'r_who'           => 0, // Вся семеья
			'r_how'           => v2_Budget::SHOW_BY_DETAIL, // Детально
			'r_middle'        => 0, // Не усреднять
			'r_currency'      => 0, // Исходная валюта
			'r_is_place'      => 0, // Все места хранения
			'r_is_tag'        => 0, // Все теги
			'r_is_category'   => 0, // Все категории
			'r_place'         => null, // Нет умолчания
			'r_tag'           => null, // Нет умолчания
			'r_category'      => null, // Нет умолчания
			'is_only_group'   => false, // Для обновления мобильных клиентов, для поддержки групп операций
			'is_with_planned' => false, // Включить в выборку планируемые операции
		);

		/**
		 * Если это получение списка трат по ID - обрабатываем первым делом, т.к.
		 * параметры при этом не важны.
		 */
		if ($this->checkIdList($idList, 'getRecordList')) {
			/**
			 * Права доступа
			 */
			$recordLimitSql = RecordUtils::getLimitSql();
			$retArray       = DAO::fillArrayKeyId("SELECT budget_record.id, budget_account_id, budget_object_id, difference AS sum,
    			operation_date, comment, currency_id, is_currency_change, is_duty, user_nuid, budget_record.budget_family_id, group_id,
    			(CASE WHEN is_currency_change IS TRUE THEN ".v2_Budget::TYPE_CHANGE." ELSE budget_object.type END) AS operation_type
    			FROM budget_record
    			JOIN budget_object ON (budget_record.budget_object_id = budget_object.id)
    			WHERE budget_record.budget_family_id = $1
    			AND budget_record.id IN (".DB::in($idList).") $recordLimitSql ORDER BY id", array(User::familyId())
			);

			foreach ($retArray as $key => $value) {
				defined('IS_SERVER') ? '' : Family::replaceHidden($retArray[$key]);
				$retArray[$key]['operation_type'] = (int)$retArray[$key]['operation_type'];
				if ($value['is_currency_change'] == 't') {
					$link         = BudgetRecordLinks::createByRecordId($value['id']);
					$serverMoveId = ($link->get(BudgetRecordLinks::RECORD1) == $value['id'] ? $link->get(BudgetRecordLinks::RECORD2) : $link->get(BudgetRecordLinks::RECORD1));

					$retArray[$key]['server_change_id'] = $serverMoveId;
				}
				elseif ($retArray[$key]['operation_type'] == v2_Budget::TYPE_MOVE) {
					$link         = BudgetRecordLinks::createByRecordId($value['id']);
					$serverMoveId = ($link->get(BudgetRecordLinks::RECORD1) == $value['id'] ? $link->get(BudgetRecordLinks::RECORD2) : $link->get(BudgetRecordLinks::RECORD1));

					$retArray[$key]['server_move_id'] = $serverMoveId;
					/**
					 * Если это ограниченный юзер и к одной из половинок у него доступа нет, считаем,
					 * что доступа нет ко всему перемещению и вторую половинку тоже убираем.
					 */
					if (User::isLimited() && !isset($retArray[$serverMoveId])) {
						/**
						 * Такого больше не должно быть, т.к. всегда отдаём либо обе половинки с подменой на
						 * "Скрытый счет" либо ни одну
						 */
						Internal::toLog("User::isLimited() && !isset(retArray[$serverMoveId])");
					}
				}
				$retArray[$key]['comment']  = FormUtil::toHtml($value['comment'], true);
				$retArray[$key]['place_id'] = $retArray[$key]['budget_account_id'];
				unset($retArray[$value['id']]['is_currency_change']);
				unset($retArray[$key]['budget_account_id']);
			}
			return $retArray;
		}

		/**
		 * Проверки
		 */
		$this->checkEmptyArray($params, '5', 'param');
		foreach ($params as $key => $param) {
			if (!array_key_exists("$key", $allowedParams)) {
				throw new SoapFault('7', "parameter '$key' not allowed. params = ".print_r($params, true));
			}
		}
		// is_report - обязательный параметр, должен быть. Остальные все имеют умолчания.
		if (!array_key_exists('is_report', $params)) {
			throw new SoapFault('28', "parameter 'is_report' is mandatory");
		}
		$arrayParams   = array('r_place', 'r_tag', 'r_category');
		$numericParams = array('r_period', 'r_what', 'r_who', 'r_how', 'r_middle', 'r_currency', 'r_is_place', 'r_is_tag', 'r_is_category');
		$boolParams    = array('is_show_duty', 'is_report', 'is_only_group', 'is_with_planned');
		$dateParams    = array('relative_date', 'period_to', 'period_from');
		// Проверяем массивы
		foreach ($arrayParams as $arrayParam) {
			if (array_key_exists($arrayParam, $params)) {
				$this->checkEmptyArray($params[$arrayParam], '8', $arrayParam);
				foreach ($params[$arrayParam] as $arrayParamValue) {
					if (!is_numeric($arrayParamValue)) {
						throw new SoapFault('10', "Value '$arrayParamValue' of '$arrayParam' is not a number");
					}
				}
			}
		}
		// Проверяем були
		foreach ($boolParams as $boolParam) {
			if (array_key_exists($boolParam, $params)) {
				if (!is_bool($params[$boolParam])) {
					throw new SoapFault('11', "'$boolParam' = '$params[$boolParam]' is not a boolean");
				}
			}
		}
		// Проверяем параметры типа "дата"
		foreach ($dateParams as $dateParam) {
			if (array_key_exists($dateParam, $params)) {
				if (!strtotime($params[$dateParam])) {
					Debug::mail($params);
					throw new SoapFault('12', "'$dateParam' = '$params[$dateParam]' is not a date. Valid date format is '2010-11-30'");
				}
			}
		}
		// Проверяем цифровые параметры
		$reportModule = ModuleFactory::spawn('v2_homeBuhPrivateReport');
		foreach ($numericParams as $numericParam) {
			if (array_key_exists($numericParam, $params)) {
				$numParamValue = $params[$numericParam];
				if (!is_numeric($numParamValue)) {
					throw new SoapFault('13', "'$numericParam' = '$numParamValue' is not an integer");
				}
				switch ($numericParam) {
					case 'r_period':
						if (!in_array($numParamValue, array_keys($reportModule->periodList)) && $numParamValue != 0) {
							throw new SoapFault('14', "'$numericParam' = '$numParamValue' should be one of: 0,".implode(",", array_keys($reportModule->periodList)));
						}
						break;
					case 'r_what':
						if (!in_array($numParamValue, array_keys($reportModule->whatList))) {
							throw new SoapFault('15', "'$numericParam' = '$numParamValue' should be one of: ".implode(",", array_keys($reportModule->whatList)));
						}
						break;
					case 'r_how':
						if (!in_array($numParamValue, array_keys(Constant::$howList))) {
							throw new SoapFault('16', "'$numericParam' = '$numParamValue' should be one of: ".implode(",", array_keys(Constant::$howList)));
						}
						break;
					case 'r_middle':
						if (!in_array($numParamValue, array_keys($reportModule->middleList)) && $numParamValue != 0) {
							throw new SoapFault('17', "'$numericParam' = '$numParamValue' should be one of: 0,".implode(",", array_keys($reportModule->middleList)));
						}
						break;
					case 'r_is_place':
					case 'r_is_tag':
					case 'r_is_category':
						if (!in_array($numParamValue, array(0, 1, 2)) && $numParamValue != 0) {
							throw new SoapFault('18', "'$numericParam' = '$numParamValue' should be one of: ".implode(",", array(0, 1, 2)));
						}
						break;
				}
			}
		}
		// Проверяем базовую логическую связность параметров
		if (!array_key_exists("r_how", $params) || (array_key_exists("r_how", $params) && $params['r_how'] == v2_Budget::SHOW_BY_DETAIL && array_key_exists("r_middle", $params))) {
			throw new SoapFault('32', "r_how != 2 or 3, so r_middle should be absent");
		}
		// Проставляем умолчания тем параметрам, которые не заданы и которым надо
		foreach ($allowedParams as $param => $allowedParam) {
			if ($allowedParam !== null && !array_key_exists("$param", $params)) {
				$params[$param] = $allowedParam;
			}
		}
		if ($params['r_period'] == 0 && (empty($params['period_to']) || empty($params['period_from']))) {
			throw new SoapFault('19', "r_period = 0, so period_to and period_from should be given");
		}
		if ($params['r_period'] != 0 && (!empty($params['period_to']) || !empty($params['period_from']))) {
			throw new SoapFault('20', "r_period != 0, so period_to and period_from should be absent");
		}
		// Если нет "только", то не должно быть и списка
		if ((empty($params['r_is_tag']) || $params['r_is_tag'] == 0) && !empty($params['r_tag'])) {
			throw new SoapFault('22', "r_is_tag = 0, so r_tag should be absent");
		}
		if ((empty($params['r_is_place']) || $params['r_is_place'] == 0) && !empty($params['r_place'])) {
			throw new SoapFault('23', "r_is_place = 0, so r_place should be absent");
		}
		if ((empty($params['r_is_category']) || $params['r_is_category'] == 0) && !empty($params['r_category'])) {
			throw new SoapFault('24', "r_is_category = 0, so r_category should be absent");
		}
		// Если есть "только", то должен быть и список
		if (!empty($params['r_is_tag']) && $params['r_is_tag'] != 0 && empty($params['r_tag'])) {
			throw new SoapFault('25', "r_is_tag != 0, so r_tag should be given");
		}
		if (!empty($params['r_is_place']) && $params['r_is_place'] != 0 && empty($params['r_place'])) {
			throw new SoapFault('26', "r_is_place != 0, so r_place should be given");
		}
		if (!empty($params['r_is_category']) && $params['r_is_category'] != 0 && empty($params['r_category'])) {
			throw new SoapFault('27', "r_is_category != 0, so r_category should be given");
		}
		// "Как показать" и "Что показать" - должны быть взаимосвязаны
		if ($params['r_how'] == v2_Budget::SHOW_BY_COSTS && $params['r_what'] != v2_Budget::TYPE_WASTE) {
			throw new SoapFault('29', "r_how = 3, so r_what should be 3");
		}
		if ($params['r_how'] == v2_Budget::SHOW_BY_SOURCE && $params['r_what'] != v2_Budget::TYPE_INCOME) {
			throw new SoapFault('30', "r_how = 2, so r_what should be 2");
		}
		if (!$params['is_report'] && $params['r_how'] != v2_Budget::SHOW_BY_DETAIL) {
			throw new SoapFault('31', "is_report = false, so r_how should be 1");
		}
		// Если отображаем НЕ доходы или расходы - то выбор категорий и тегов должен отсутствовать
		if (!in_array($params['r_what'], array(v2_Budget::TYPE_WASTE, v2_Budget::TYPE_INCOME)) && (array_key_exists("r_tag", $params) || array_key_exists("r_category", $params))) {
			throw new SoapFault('32', "r_what != 2 or 3, so r_tag and r_category should be absent");
		}
		// Дату переводим в timestamp
		if (!empty($params['relative_date'])) {
			$params['relative_timestamp'] = User::getTime(strtotime($params['relative_date']));
		}
		// Дату переводим в timestamp
		if (isset($params['is_with_planned'])) {
			$params['is_with_planned'] = $params['is_with_planned'] ? 'true' : 'false';
		}

		/**
		 * Готовим модуль отчётов к выплонению
		 */
		foreach ($params as $key => $param) {
			$reportModule->form->set($key, $param);
		}
		/**
		 * Когда тянем все записи целиком, то нужны и начальные остатки
		 * Ставим спец. ключ для этого.
		 */
		if ($params['r_period'] == v2_Budget::PERIOD_ALL) {
			$reportModule->form->set('is_with_init_rest', true);
		}

		$reportModule->process();
		$reportTable = $reportModule->reportTable;

		if ($params['is_report']) {
			if ($params['r_how'] == v2_Budget::SHOW_BY_DETAIL) {
				foreach ($reportTable as $key => $value) {
					$reportTable[$key]['operation_type'] = (int)$reportTable[$key]['operation_type'];
					if (array_key_exists("comment", $value)) {
						$reportTable[$key]['comment'] = FormUtil::toHtml($value['comment'], true);
					}
				}
			}
			return $reportTable;
		}

		/**
		 * Раз дошли до сюда - то это вытягивание всех записей при начальной синхре.
		 * Надо также грохнуть всё в budget_sync_last, для случая "Начать всё сначала",
		 * чтобы следующие id-ы, добавленные на клиенте после синхры, не совпали с теми,
		 * которые уже есть в проверочной таблице.
		 */
		DB::query("DELETE FROM budget_sync_last
    		WHERE user_id = $1
			AND api_id = $2",
			array(User::id(), Session::get(Constant::API_ID))
		);

		/**
		 * Надо возвращать список самих трат, а не таблицу с отчётом.
		 */
		$idsArray = array();
		foreach ($reportTable as $tableItem) {
			$idsArray[] = $tableItem['id'];
		}
		if (empty($idsArray)) {
			return array();
		}

		$count          = count($reportTable);
		$newReportTable = array();

		for ($i = 0; $i < $count; $i++) {
			$value                             = $reportTable[$i];
			$reportTable[$i]['operation_type'] = (int)$value['operation_type'];
			$reportTable[$i]['place_id']       = $value['budget_account_id'];
			unset($reportTable[$i]['budget_account_id']);

			$reportTable[$i]['sum'] = (string)$value['difference'];
			unset($reportTable[$i]['difference']);

			unset($reportTable[$i]['name']);
			unset($reportTable[$i]['nick']);

			if ($reportTable[$i]['operation_type'] == v2_Budget::TYPE_CHANGE) {
				$link                                = BudgetRecordLinks::createByRecordId($value['id']);
				$serverMoveId                        = ($link->get(BudgetRecordLinks::RECORD1) == $value['id'] ? $link->get(BudgetRecordLinks::RECORD2) : $link->get(BudgetRecordLinks::RECORD1));
				$reportTable[$i]['server_change_id'] = $serverMoveId;
			}
			if ($reportTable[$i]['operation_type'] == v2_Budget::TYPE_MOVE) {
				$link                              = BudgetRecordLinks::createByRecordId($value['id']);
				$serverMoveId                      = ($link->get(BudgetRecordLinks::RECORD1) == $value['id'] ? $link->get(BudgetRecordLinks::RECORD2) : $link->get(BudgetRecordLinks::RECORD1));
				$reportTable[$i]['server_move_id'] = $serverMoveId;
			}

			$reportTable[$i]['comment'] = FormUtil::toHtml($value['comment'], true);

			$newReportTable[$value['id']] = $reportTable[$i];
			unset($reportTable[$i]);
		}

		return $newReportTable;
	}

	/**
	 * Retrievs balance list (array of arrays) grouped by places; [params] => array of parameters: 'restDate' [YYYY-MM-DD (NOW default)] - rests will be retrieved relative to this value; 'is_with_accum' [true|false (default)] - If true, the the accumulation will be subtracted; 'is_with_duty' [true|false (default)] - If true, duties will be subtracted;
	 *
	 * @param string $apiId
	 * @param string $login
	 * @param string $pass
	 * @param array $params
	 * @return array
	 */
	public function getBalance($apiId, $login, $pass, $params)
	{
		$this->login($apiId, $login, $pass);

		// Массив разрешённых параметров, он же задаёт умолчания
		$allowedParams = array(
			'restDate'       => 0, // сегодня
			'is_with_accum'  => false, // Не вычитать накопления
			'is_with_duty'   => false, // Не вычитать долги
			'is_with_hidden' => false, // Не отображать скрытые места
			'is_with_null'   => false, // Включать счета с нулевыми балансами
		);

		/**
		 * Проверки
		 */
		if (!empty($params) && !is_array($params)) {
			throw new SoapFault('33', 'params is not an array');
		}
		if (empty($params)) {
			$params = array();
		}
		foreach ($params as $key => $param) {
			if (!array_key_exists("$key", $allowedParams)) {
				throw new SoapFault('34', "parameter '$key' not allowed[1], params = ".print_r($params, true));
			}
		}
		$boolParams = array('is_with_accum', 'is_with_duty', 'is_with_hidden', 'is_with_null');
		$dateParams = array('restDate');
		// Проверяем були
		foreach ($boolParams as $boolParam) {
			if (array_key_exists($boolParam, $params)) {
				if (!is_bool($params[$boolParam])) {
					throw new SoapFault('35', "'$boolParam' = '$params[$boolParam]' is not a boolean");
				}
			}
		}
		// Проверяем параметры типа "дата"
		foreach ($dateParams as $dateParam) {
			if (array_key_exists($dateParam, $params)) {
				if (!strtotime($params[$dateParam])) {
					throw new SoapFault('36', "'$dateParam' = '$params[$dateParam]' is not a date. Valid date format is '2010-11-30'");
				}
			}
		}
		// Проставляем умолчания тем параметрам, которые не заданы
		foreach ($allowedParams as $param => $allowedParam) {
			if ($allowedParam !== null && !array_key_exists("$param", $params)) {
				$params[$param] = $allowedParam;
			}
		}

		/**
		 * В iOS версии для корректной сверки балансов нужно отдавать и инвест валюты
		 */
		if (!User::isLimited() && 'vanr_api' != $apiId && !defined('IS_SERVER')) {
			$params['is_with_invest'] = true;
		}

		/**
		 * Готовим модуль к выполнению
		 */
		$restList = Family::getBalance($params);
		return $restList;
	}

	/**
	 * Insert or update currency list; [list] => array of arrays: 'server_id' or 'client_id' [int8] - server or client ID of the record# If client ID is present - try to insert new record, and return server2client correspondence in the result array# If server_id is present - try to update existing record, @param string $apiId
	 * @param string $login
	 * @param string $pass
	 * @param array $list
	 * @return array
	 * @see getCurrencyList description for other detail; Returns the array of server IDs, successfully changed; If 'is_default' flag set to another currency, then previos updated too and sent in return array; The client MUST save server IDs corresponded to client IDs, for subsequent 'update' and 'delete' calls;
	 *
	 */
	public function setCurrencyList($apiId, $login, $pass, $list)
	{
		/**
		 * Для функций записи должен быть активен Premium
		 */
		try {
			$this->getAccessStatus($apiId, $login, $pass);
		}
		catch (SoapFault $e) {
			if ($e->faultcode != '87') { // trial
				throw $e;
			}
		}

		/**
		 * Проверки
		 */
		$this->checkEmptyArray($list, '38', 'Currency list');

		try {
			DB::transactionBegin();
			$retArray     = array();
			$dublDebug    = array();
			$lastSyncData = DAO::fillArrayKeyId("SELECT client_id AS id, server_id, app_instance_id FROM budget_sync_last
    			WHERE user_id = $1
    			AND api_id = $2
    			AND object_type = $3",
				array(User::id(), Session::get(Constant::API_ID), BudgetHistory::TYPE_CURRENCY)
			);
			$appId        = self::getAppId();

			foreach ($list as $currency) {
				/**
				 * Первым делом, если это добавление клиентской записи, которая
				 * уже была добавлена в прошлой синхре, то значит был сбой на клиенте
				 * и он шлёт дубль. Ничего не добавляем, а отдаём тот же server_id
				 * что и в прошлый раз.
				 */
				$isClientIdThisInstance = true;
				if (
					abs($appId) > 0 &&
					!empty($currency['client_id']) &&
					isset($lastSyncData[$currency['client_id']]) &&
					abs($lastSyncData[$currency['client_id']]['app_instance_id']) > 0 &&
					$lastSyncData[$currency['client_id']]['app_instance_id'] != $appId
				) {
					$isClientIdThisInstance = false;
				}
				if (!empty($currency['client_id']) && isset($lastSyncData[$currency['client_id']]) && $isClientIdThisInstance) {
					$retArray[]  = array(
						'server_id' => $lastSyncData[$currency['client_id']]['server_id'],
						'client_id' => $currency['client_id'],
						'status'    => 'inserted'
					);
					$dublDebug[] = array(
						'server_id' => $lastSyncData[$currency['client_id']]['server_id'],
						'client_id' => $currency['client_id'],
						'type'      => 'currency',
					);
					continue;
				}

				/**
				 * Базовые проверки
				 */
				if (empty($currency['server_id']) && empty($currency['client_id'])) {
					throw new SoapFault('74', "'server_id' or 'client_id' should be given");
				}
				$currencyId                       = empty($currency['server_id']) ? '' : $currency['server_id'];
				$currency[v2_Currency::FAMILY_ID] = Session::get(User::FAMILY_ID);
				$currency[v2_Currency::RATIO]     = 1;
				$currency[v2_Currency::ID]        = $currencyId ? $currencyId : null;

				defined('IS_SERVER') ? '' : $currency[v2_Currency::IS_INVESTING] = false;

				$currencyObj = v2_Currency::create();
				$this->checkObject($currencyObj, $currency, 'Currency');

				/**
				 * Кастомные проверки валюты и сохранение.
				 */
				if ($currencyId) {
					$famCount = DAO::getCount("SELECT COUNT(id) FROM currency WHERE id = $1 AND family_id = $2", true,
						array($currencyId, User::familyId())
					);
					if ($famCount != 1) {
						throw new SoapFault('188', "Внимание!\nДля исправления сбоя в синхронизации, приложение необходимо переустановить.");
					}
				}
				/**
				 * Если поставили галку "по умолчанию" - то сначала снимаем со всех других эл-тов эти галки.
				 * Иначе, если это обновление - проверяем, что эта валюта не являлась валютой по умолчанию.
				 */
				if ($currency[v2_Currency::IS_DEFAULT]) {
					$idSql         = $currencyId ? "AND id <> ".DB::esc($currencyId) : '';
					$oldCurrencies = v2_Currency::create()->daoMakeObjectList("SELECT * FROM currency WHERE is_default IS TRUE $idSql AND family_id = $1", false,
						array(User::familyId())
					);
					foreach ($oldCurrencies as $oldCurrency) {
						$oldCurrency->set(v2_Currency::IS_DEFAULT, false);
						$oldCurrency->daoSave();
						$oldCurrency->toServerHistory("update");
						$retArray[] = array(
							'server_id' => $oldCurrency->get(v2_Currency::ID),
							'status'    => 'updated'
						);
					}
				}
				elseif ($currencyId) {
					$wasDef = DAO::getCount("SELECT COUNT(id) FROM currency WHERE is_default IS TRUE AND id = $1 AND family_id = $2", true,
						array($currencyId, User::familyId())
					);
					if ($wasDef == 1) {
						throw new SoapFault('49', "Currency with id = $currencyId is default currency. Set another default currency first.");
					}
				}

				$s = 's';
				$currencyObj->
				set(v2_Currency::COURSE, $this->checkNumeric($currency[v2_Currency::COURSE], "Currency"))->
				set(v2_Currency::CODE, mb_strtoupper($this->normalizeString($currency[v2_Currency::CODE], 16, "Currency")))->
				set(v2_Currency::NAME, FormUtil::minimizeString($currency[v2_Currency::NAME], 16, $s, ''))->
				set(v2_Currency::IS_INVESTING, false)->
				daoSave()->
				toServerHistory($currencyId ? 'update' : 'insert');

				$retArrayItem = array(
					'server_id' => $currencyObj->get(v2_Currency::ID),
					'status'    => $currencyId ? 'updated' : 'inserted'
				);
				if (!$currencyId) {
					$retArrayItem['client_id'] = $currency['client_id'];
				}
				$retArray[] = $retArrayItem;
			}
			/**
			 * Перед тем, как успешно отправить список связок, надо грохнуть все
			 * клиентско-серверные связки данного типа, от прошлой синхры и добавить текущие.
			 */
			$delAppIdSql = $appId ? "AND app_instance_id = ".DB::esc($appId) : '';
			DB::query("DELETE FROM budget_sync_last WHERE user_id = $1
	    		AND api_id = $2 AND object_type = $3 $delAppIdSql",
				array(User::id(), Session::get(Constant::API_ID), BudgetHistory::TYPE_CURRENCY)
			);
			foreach ($retArray as $retValue) {
				if ($retValue['status'] != 'inserted') {
					continue;
				}
				BudgetSyncLast::create()->
				set(BudgetSyncLast::USER_ID, User::id())->
				set(BudgetSyncLast::API_ID, Session::get(Constant::API_ID))->
				set(BudgetSyncLast::OBJECT_TYPE, BudgetHistory::TYPE_CURRENCY)->
				set(BudgetSyncLast::CLIENT_ID, $retValue['client_id'])->
				set(BudgetSyncLast::SERVER_ID, $retValue['server_id'])->
				set(BudgetSyncLast::INSTANCE_ID, $appId)->
				set(BudgetSyncLast::DATE, 'now()')->
				daoSave();
			}
			DB::transactionCommit();
			/**
			 * Для отладки шлём уведомления о дублях.
			 * Когда станет не нужным - убрать.
			 */
			if ($dublDebug) {
				Debug::mail($dublDebug);
			}
		}
		catch (Exception $e) {
			DB::transactionRollback();
			$faultCode = isset($e->faultcode) ? $e->faultcode : 0;
			if ($faultCode == '188') {
				throw $e;
			}
			Debug::mail($e);
			throw new SoapFault('58', $e->getMessage());
		}
		return $retArray;
	}

	/**
	 * Insert or update waste category list; [list] => array of arrays: 'server_id' or 'client_id' [int8] - server or client ID of the record# If client ID is present - try to insert new record, and return server2client correspondence in the result array# If server_id is present - try to update existing record, @param string $apiId
	 * @param string $login
	 * @param string $pass
	 * @param array $list
	 * @return array
	 * @see getCategoryList description for other detail; Returns the array of server IDs, successfully changed; The client MUST save server IDs corresponded to client IDs, for subsequent 'update' and 'delete' calls;
	 *
	 */
	public function setCategoryList($apiId, $login, $pass, $list)
	{
		return $this->setObjectList($apiId, $login, $pass, $list, v2_Budget::TYPE_WASTE);
	}

	/**
	 * Insert or update income source list; [list] => array of arrays: 'server_id' or 'client_id' [int8] - server or client ID of the record# If client ID is present - try to insert new record, and return server2client correspondence in the result array# If server_id is present - try to update existing record, @param string $apiId
	 * @param string $login
	 * @param string $pass
	 * @param array $list
	 * @return array
	 * @see getSourceList description for other detail; Returns the array of server IDs, successfully changed; The client MUST save server IDs corresponded to client IDs, for subsequent 'update' and 'delete' calls;
	 *
	 */
	public function setSourceList($apiId, $login, $pass, $list)
	{
		return $this->setObjectList($apiId, $login, $pass, $list, v2_Budget::TYPE_INCOME);
	}

	/**
	 * Insert or update place list; [list] => array of arrays: 'server_id' or 'client_id' [int8] - server or client ID of the record# If client ID is present - try to insert new record, and return server2client correspondence in the result array# If server_id is present - try to update existing record, @param string $apiId
	 * @param string $login
	 * @param string $pass
	 * @param array $list
	 * @return array
	 * @see getPlaceList description for other detail; Returns the array of server IDs, successfully changed; The client MUST save server IDs corresponded to client IDs, for subsequent 'update' and 'delete' calls;
	 *
	 */
	public function setPlaceList($apiId, $login, $pass, $list)
	{
		return $this->setObjectList($apiId, $login, $pass, $list, v2_Budget::TYPE_MOVE);
	}

	/**
	 * Insert or update tag list; [list] => array of arrays: 'server_id' or 'client_id' [int8] - server or client ID of the record# If client ID is present - try to insert new record, and return server2client correspondence in the result array# If server_id is present - try to update existing record, @param string $apiId
	 * @param string $login
	 * @param string $pass
	 * @param array $list
	 * @return array
	 * @see getTagList description for other detail; Returns the array of server IDs, successfully changed; The client MUST save server IDs corresponded to client IDs, for subsequent 'update' and 'delete' calls;
	 *
	 */
	public function setTagList($apiId, $login, $pass, $list)
	{
		/**
		 * Для функций записи должен быть активен Premium
		 */
		try {
			$this->getAccessStatus($apiId, $login, $pass);
		}
		catch (SoapFault $e) {
			if ($e->faultcode != '87') { // trial
				throw $e;
			}
		}

		/**
		 * Проверки
		 */
		$this->checkEmptyArray($list, '56', 'Tag list');

		$retArray = array();
		try {
			DB::transactionBegin();
			$dublDebug    = array();
			$lastSyncData = DAO::fillArrayKeyId("SELECT client_id AS id, server_id, app_instance_id FROM budget_sync_last
    			WHERE user_id = $1
    			AND api_id = $2
    			AND object_type = $3",
				array(User::id(), Session::get(Constant::API_ID), BudgetHistory::TYPE_TAG)
			);
			$appId        = self::getAppId();

			foreach ($list as $tag) {
				/**
				 * Первым делом, если это добавление клиентской записи, которая
				 * уже была добавлена в прошлой синхре, то значит был сбой на клиенте
				 * и он шлёт дубль. Ничего не добавляем, а отдаём тот же server_id
				 * что и в прошлый раз.
				 */
				$isClientIdThisInstance = true;
				if (
					abs($appId) > 0 &&
					!empty($tag['client_id']) &&
					isset($lastSyncData[$tag['client_id']]) &&
					abs($lastSyncData[$tag['client_id']]['app_instance_id']) > 0 &&
					$lastSyncData[$tag['client_id']]['app_instance_id'] != $appId
				) {
					$isClientIdThisInstance = false;
				}

				if (!empty($tag['client_id']) && isset($lastSyncData[$tag['client_id']]) && $isClientIdThisInstance) {
					$retArray[]  = array(
						'server_id' => $lastSyncData[$tag['client_id']]['server_id'],
						'client_id' => $tag['client_id'],
						'status'    => 'inserted'
					);
					$dublDebug[] = array(
						'server_id' => $lastSyncData[$tag['client_id']]['server_id'],
						'client_id' => $tag['client_id'],
						'type'      => 'tag',
					);
					continue;
				}

				/**
				 * Базовые проверки
				 */
				if (empty($tag['server_id']) && empty($tag['client_id'])) {
					throw new SoapFault('76', "'server_id' or 'client_id' should be given");
				}
				$tagId                     = empty($tag['server_id']) ? '' : $tag['server_id'];
				$tag[BudgetTag::FAMILY_ID] = Session::get(User::FAMILY_ID);
				$tag[BudgetTag::USER_ID]   = User::id();
				$tag[BudgetTag::ID]        = $tagId ? $tagId : null;
				$tagObj                    = BudgetTag::create();
				$this->checkObject($tagObj, $tag, 'Tag');

				/**
				 * Кастомные проверки тега и сохранение.
				 */
				if ($tagId) {
					$famCount = DAO::getCount("SELECT COUNT(id) FROM budget_tags WHERE id = $1 AND family_id = $2", true,
						array($tagId, User::familyId())
					);
					if ($famCount != 1) {
						throw new SoapFault('57', "Tag object with id = $tagId are invalid");
					}
				}

				$tagObj->
				set(BudgetTag::SORT, $this->checkNumeric($tag[BudgetTag::SORT], "Tag"))->
				set(BudgetTag::NAME, $this->normalizeString($tag[BudgetTag::NAME], 64, "Tag"))->
				daoSave();

				$retArrayItem = array(
					'server_id' => $tagObj->get(BudgetTag::ID),
					'status'    => $tagId ? 'updated' : 'inserted'
				);
				if (!$tagId) {
					$retArrayItem['client_id'] = $tag['client_id'];
				}
				$retArray[] = $retArrayItem;
			}
			/**
			 * Перед тем, как успешно отправить список связок, надо грохнуть все
			 * клиентско-серверные связки данного типа, от прошлой синхры и добавить текущие.
			 */
			$delAppIdSql = $appId ? "AND app_instance_id = ".DB::esc($appId) : '';
			DB::query("DELETE FROM budget_sync_last
	    		WHERE user_id = $1
				AND api_id = $2
				AND object_type = $3
				$delAppIdSql",
				array(User::id(), Session::get(Constant::API_ID), BudgetHistory::TYPE_TAG)
			);
			foreach ($retArray as $retValue) {
				if ($retValue['status'] != 'inserted') {
					continue;
				}
				BudgetSyncLast::create()->
				set(BudgetSyncLast::USER_ID, User::id())->
				set(BudgetSyncLast::API_ID, Session::get(Constant::API_ID))->
				set(BudgetSyncLast::OBJECT_TYPE, BudgetHistory::TYPE_TAG)->
				set(BudgetSyncLast::CLIENT_ID, $retValue['client_id'])->
				set(BudgetSyncLast::SERVER_ID, $retValue['server_id'])->
				set(BudgetSyncLast::INSTANCE_ID, $appId)->
				set(BudgetSyncLast::DATE, 'now()')->
				daoSave();
			}
			DB::transactionCommit();
			/**
			 * Для отладки шлём уведомления о дублях.
			 * Когда станет не нужным - убрать.
			 */
			if ($dublDebug) {
				Debug::mail($dublDebug);
			}
		}
		catch (Exception $e) {
			DB::transactionRollback();
			Debug::mail($e);
			throw new SoapFault('59', $e->getMessage());
		}
		return $retArray;
	}

	/**
	 * Insert or update record list; [list] => array (indexes must be 0,1,2...N) of arrays: 'server_id' or 'client_id' [int8] - server or client ID of the record# If client ID is present - try to insert new record, and return server2client correspondence in the result array# If server_id is present - try to update existing record; 'server_move_id' or 'client_move_id' [int8] - for "move" operations only, to identify second part of move# The value must point to first part 'server_id' or 'client_id' respectively; 'server_change_id' or 'client_change_id' [int8] - for "currency change" operations only, to identify second part of change# The value must point to first part 'server_id' or 'client_id' respectively; 'place_id' [int8] - place ID, of which the record; 'budget_object_id' [int8] - object ID of which the record: category ID for waste, source ID for incomes, place ID for moves and currency changes; 'sum' [int8] - absolute value of sum (hundredths); 'operation_date' [YYYY-MM-DD HH:mm:SS] - transaction date; 'comment' [UTF8 text] - the comment of the record, 2048 chars max length; 'currency_id' [int8] - currency ID of the record; 'is_duty' [true|false] - not used; 'operation_type' [income = 2, waste = 3 (default), move = 4, change = 5] - transaction type; Returns the array of server IDs, successfully changed; The client MUST save server IDs corresponded to client IDs, for subsequent 'update' and 'delete' calls;
	 *
	 * @param string $apiId
	 * @param string $login
	 * @param string $pass
	 * @param array $list
	 * @return array
	 */
	public function setRecordList($apiId, $login, $pass, $list)
	{
		/**
		 * Для функций записи должен быть активен Premium
		 */
		try {
			$this->getAccessStatus($apiId, $login, $pass);
		}
		catch (SoapFault $e) {
			if ($e->faultcode != '87') { // trial
				throw $e;
			}
		}

		$retArray = array();
		/**
		 * Проверки
		 */
		$this->checkEmptyArray($list, '60', 'Record list');

		$conflictMsg = 'Для восстановления синхронизации, ';
		$conflictMsg .= ($apiId == 'vandr_api' ? "необходимо перезайти через меню\n'Ещё->Сменить пользователя'." : 'приложение необходимо переустановить.');

		try {
			DB::transactionBegin();
			$moves        = array();
			$changes      = array();
			$dublDebug    = array();
			$lastSyncData = DAO::fillArrayKeyId("SELECT client_id AS id, server_id, app_instance_id FROM budget_sync_last
    			WHERE user_id = $1
    			AND api_id = $2
    			AND object_type = $3",
				array(User::id(), Session::get(Constant::API_ID), BudgetHistory::TYPE_RECORD)
			);
			$appId        = self::getAppId();

			foreach ($list as $key => &$record) {
				/**
				 * Первым делом, если это добавление клиентской записи, которая
				 * уже была добавлена в прошлой синхре, то значит был сбой на клиенте
				 * и он шлёт дубль. Ничего не добавляем, а отдаём тот же server_id
				 * что и в прошлый раз.
				 *
				 * 8 июля 2014. Решено для budget_sync_last идентифицировать само приложение, т.к.
				 * если юзер установить приложение на два девайса, то велика вероятность пересечения
				 * клиентских айдишников. Идентификатор приложения включили в user agent.
				 * В SQL выборку равенство по instance_id добавить не можем из-за старых версий.
				 */
				$isClientIdThisInstance = true;
				if (
					abs($appId) > 0 &&
					!empty($record['client_id']) &&
					isset($lastSyncData[$record['client_id']]) &&
					abs($lastSyncData[$record['client_id']]['app_instance_id']) > 0 &&
					$lastSyncData[$record['client_id']]['app_instance_id'] != $appId
				) {
					$isClientIdThisInstance = false;
				}

				if (
					!empty($record['client_id']) &&
					isset($lastSyncData[$record['client_id']]) &&
					$isClientIdThisInstance
				) {
					$retArray[]  = array(
						'server_id' => $lastSyncData[$record['client_id']]['server_id'],
						'client_id' => $record['client_id'],
						'status'    => 'inserted'
					);
					$dublDebug[] = array(
						'server_id' => $lastSyncData[$record['client_id']]['server_id'],
						'client_id' => $record['client_id'],
						'type'      => 'record',
					);
					continue;
				}

				/**
				 * Базовые проверки
				 */
				$record[v2_BudgetRecord::FAMILY_ID]      = Session::get(User::FAMILY_ID);
				$record[v2_BudgetRecord::NUID]           = User::id();
				$record[v2_BudgetRecord::ACCOUNT_ID]     = $record['place_id'];
				$record[v2_BudgetRecord::SUM]            = $record['sum'];
				$record[v2_BudgetRecord::IS_CURR_CHANGE] = false;
				$recordObj                               = v2_BudgetRecord::create();
				$this->checkObject($recordObj, $record, 'Record');

				if (empty($record['server_id']) && empty($record['client_id'])) {
					throw new SoapFault('64', "'server_id' or 'client_id' should be given");
				}
				$currentId = empty($record['server_id']) ? $record['client_id'] : $record['server_id'];

				/**
				 * Кастомные проверки.
				 */
				empty($record['client_id']) ? '' : $this->checkNumeric($record['client_id'], 'Record');
				$this->normalizeString($record[v2_BudgetRecord::COMMENT], 2048, "Record");
				$this->checkNumeric($record[v2_BudgetRecord::SUM], "Record");
				if (!is_int(abs((string)$record[v2_BudgetRecord::SUM]))) {
					throw new SoapFault('115', "Record: '{$record[v2_BudgetRecord::SUM]}' is not an integer");
				}

				if (!strtotime($record[v2_BudgetRecord::DATE])) {
					throw new SoapFault('62', "'{$record[v2_BudgetRecord::DATE]}' is not a date. Valid date format is '2010-11-30 14:27:01'");
				}

				// С iPhone иногда приходят даты вида 2013-04-22 10:02:35 p.m. Видимо буржуи.
				$record['operation_date'] = date("Y-m-d H:i:s", strtotime($record['operation_date']));
				isset($record['operation_type']) ? '' : $record['operation_type'] = v2_Budget::TYPE_WASTE;

				/**
				 * Если элемента справочника нет на сервере, то чтобы не ступиролась синхронизация,
				 * создаём фихтивную категорию/источник/счёт, с логически понятным для удаления названием
				 * Если элемент есть, но пренадлежит другой семье (напр, после отключения юзера) - стоп
				 * Так же, если нет ВСЕХ сразу элементов (счёт, валюта, категория) в данной семье - стоп, т.к.
				 * скорее всего это кейс когда юзер не перелогинился после объединения/отключения.
				 */
				$objId = $record['budget_object_id'];
				$plcId = $record['place_id'];
				$crnId = $record['currency_id'];

				$isCurrencyEmpty = empty(v2_Currency::getElemById($crnId));
				$isPlaceEmpty    = (Constant::UNDEFINED == Family::getElementName($plcId));
				$isObjEmpty      = (Constant::UNDEFINED == Family::getElementName($objId) && $objId >= 0);

				if ($isCurrencyEmpty && $isPlaceEmpty && $isObjEmpty) {
					throw new SoapFault('188', "Внимание!\n$conflictMsg");
				}

				if ($isCurrencyEmpty) {
					if (DAO::getCount("SELECT id FROM currency WHERE id = $1", true, array($crnId))) {
						throw new SoapFault('188', "Внимание!\n$conflictMsg");
					}
					$curList = v2_AjaxModule::getCurrencyList();

					v2_Currency::create()->
					set(v2_Currency::ID, $crnId)->
					set(v2_Currency::IS_DEFAULT, empty($curList) ? true : false)->
					set(v2_Currency::NAME, Lang::id(100000528))->
					set(v2_Currency::FAMILY_ID, User::familyId())->
					set(v2_Currency::IS_AUTOUPDATE, false)->
					set(v2_Currency::COURSE, 1)->
					set(v2_Currency::RATIO, 1)->
					set(v2_Currency::IS_HIDDEN, false)->
					set(v2_Currency::IS_INVESTING, false)->
					daoSave(true);

					v2_AjaxModule::getCurrencyList(true);
				}
				if ($isPlaceEmpty) {
					if (DAO::getCount("SELECT id FROM budget_object WHERE id = $1", true, array($plcId))) {
						throw new SoapFault('188', "Внимание!\n$conflictMsg");
					}
					v2_BudgetObject::addRootElement(Lang::id(100000528), User::familyId(), v2_Budget::TYPE_MOVE, $plcId);
					Family::getPlaceList(true);
				}
				if ($isObjEmpty) {
					if (DAO::getCount("SELECT id FROM budget_object WHERE id = $1", true, array($objId))) {
						throw new SoapFault('188', "Внимание!\n$conflictMsg");
					}
					v2_BudgetObject::addRootElement(Lang::id(100000528), User::familyId(), $record['operation_type'], $objId);
					if ($record['operation_type'] == v2_Budget::TYPE_WASTE) {
						Family::getCategoryList(true);
					}
					elseif ($record['operation_type'] == v2_Budget::TYPE_INCOME) {
						Family::getSourceList(true);
					}
					elseif ($record['operation_type'] == v2_Budget::TYPE_MOVE) {
						Family::getPlaceList(true);
					}
				}

				/**
				 * Если у перемещения какой-то счёт является источником дохода или категорей расходов,
				 * значит где-то выше косяк, чтобы не стопорить синхронизацию, превращаем перемещение в
				 * корректные расход+доход. На клиенте и сервере будут при этом немного разные данные, но
				 * это меньшее зло, чем сбой в синхронизации.
				 */
				if ($record['operation_type'] == v2_Budget::TYPE_MOVE) {
					$obj = Family::getElementById($objId);
					if ($obj['type'] != v2_Budget::TYPE_MOVE) {
						$record['operation_type'] = $record['sum'] < 0 ? v2_Budget::TYPE_WASTE : v2_Budget::TYPE_INCOME;
						/**
						 * Так же и вторую половинку нужно найти и исправить. При этом в приложении у неё вместо
						 * счёта - категория, а значит остатки она не меняет. Делаем транзакцию с нулевой суммой.
						 */
						$needIdType         = empty($record['server_id']) ? 'client_move_id' : 'server_move_id';
						$needLocalIdType    = empty($record['server_id']) ? 'client_id' : 'server_id';
						$thisRecordClientId = $record[$needLocalIdType];
						$record2            = null;
						for ($i = 0; $i < count($list); $i++) {
							$curRecordClientId = ArrayUtils::safeGet($list[$i], $needIdType);
							if ($curRecordClientId == $thisRecordClientId) {
								$list[$i]['operation_type']   = $list[$i]['sum'] < 0 ? v2_Budget::TYPE_WASTE : v2_Budget::TYPE_INCOME;
								$list[$i]['sum']              = 0;
								$list[$i]['place_id']         = $plcId;
								$list[$i]['budget_object_id'] = $objId;
								$record2                      = $list[$i];
								break;
							}
						}
					}
				}

				/**
				 * Составляем объект формы, для унифицированной передачи в RecordUtils::processRecord();
				 */
				$form = new Form();
				$form->clear();
				switch ($record['operation_type']) {
					case v2_Budget::TYPE_INCOME:
						/**
						 * @todo
						 * Баг обработка СМС об отмене выставляет счёт -1
						 */
						if (-1 == $plcId) {
							$plcId = Family::getFirstPlaceId();
						}
					case v2_Budget::TYPE_WASTE:
						$wiPrefix = $record['operation_type'] == v2_Budget::TYPE_WASTE ? 'w_' : 'i_';
						// Заполняем форму
						if (!empty($record['server_id'])) {
							$form->set(($wiPrefix == 'w_' ? 'waste' : 'income')."Id", $record['server_id']);
						}
						/**
						 * Если у траты категорией является источник дохода, или у дохода источником - категория расходов,
						 * значит где-то выше косяк, чтобы не стопорить синхронизацию исправляем на категорию по умолчанию
						 * На клиенте и сервере будут при этом немного разные данные, но считаем, что это меньшее зло, чем
						 * сбой в синхронизации.
						 */
						$obj = Family::getElementById($objId);
						if ($obj['type'] != $record['operation_type']) {
							/**
							 * А вот если такая белиберда приходит при редактировании операции, значит данные на клиенте
							 * подпорчены уже серьёзно. Заставляем перезайти.
							 */
							if (!empty($record['server_id'])) {
								throw new SoapFault('188', "Внимание!\n$conflictMsg");
							}
							$objId = $wiPrefix == 'w_' ? Family::getFirstCategoryId() : Family::getFirstSourceId();
							Debug::mail("Выставлена категория: $objId. record =".print_r($record, true));
						}
						$sum = abs($record['sum']);
						$form->set("sum_to_post", $sum ? $sum : 1);
						$form->set("date_to_post", $record['operation_date']);
						$form->set($wiPrefix."currency", $crnId);
						$form->set($wiPrefix."from", $plcId);
						$form->set($wiPrefix."category", $objId);
						$form->set($wiPrefix."comment", $record['comment']);
						$form->set("is_".$wiPrefix."duty", $record['is_duty'] ? 'on' : 'off');
						$form->set("is_change_".$wiPrefix."duty", 'off');
						if (!empty($record['group_id'])) {
							$form->set("is_spl_group", "true");
							$form->set("spl_group_id", "{$record['group_id']}");
						}
						/**
						 * Единый мега обработчик ввода траты.
						 */
						try {
							$budgetRecord = RecordUtils::processRecord($form, $wiPrefix == 'w_' ? true : false);
							/**
							 * С 19 дек 2022 на сервере запрещены нулевые суммы. На старых клиентах они могут
							 * появляться из чеков например, позволяем запись сохранить и синхронизацию провести
							 * но затем сразу запись удаляем на сервере. На остатки она не влияет, ничего страшного.
							 */
							$sum ? '' : $budgetRecord->daoDrop();
						}
						catch (BaseException $e) {
							/**
							 * Прикручиваем костыль на следующий случай:
							 * 1. На клиенте вторая фаза (отправка локальных изменений) завершается
							 *    неудачно и следующую синхру будем начинать с неё.
							 * 2. За то время, пока следующая синхра не запускалась, в приложении
							 *    запись отредактирована, а на сервере она же - удалена.
							 * 3. Поскольку клиент в следующей синхре серверные изменения не запрашивает
							 *    то сделанного на сервере удаления не получает, а клиентское изменение
							 *    шлёт. На что сервер и ругается.
							 * Страшного тут ничего нет, проверяем, что запись была действительно удалена
							 * и принадлежит клиенту, после чего отдаём "успешное" редактирование, а в
							 * след. раз запись успешно удалится на клиенте и остатки совпадут.
							 */
							$delCount = 0;
							if (!empty($record['server_id']) && strpos($e->getMessage(), ', does not exist') !== false) {
								$delCount = DAO::getCount("SELECT id FROM budget_history WHERE family_id = $1
									AND object_id = $2 AND object_type_id = '1'
									AND action_id = '3' LIMIT 1", true,
									array(User::familyId(), $record['server_id'])
								);
							}
							if ($delCount) {
								$budgetRecord = v2_BudgetRecord::create()->set(Business::ID, $record['server_id']);
							}
							else {
								throw $e;
							}
						}
						/**
						 * Формируем результирующий массив
						 */
						$retArrayItem = array(
							'server_id' => $budgetRecord->get(v2_BudgetRecord::ID),
							'status'    => 'updated'
						);
						if (!empty($record['client_id'])) {
							$retArrayItem['client_id'] = $record['client_id'];
							$retArrayItem['status']    = 'inserted';
						}
						$retArray[] = $retArrayItem;
						break;
					case v2_Budget::TYPE_MOVE:
						/**
						 * Если эта половинка уже сохранена, то значит и вторая сохранена. Пропускаем
						 */
						if (!empty($list[$key]['is_saved'])) {
							continue 2;
						}

						$needIdType = empty($record['server_id']) ? 'client_move_id' : 'server_move_id';
						$record2id  = null;
						$record2    = array();
						/**
						 * По историческим причинам, клиент вправе присылать ссылку как в обеих половинках
						 * так и только в одной. Оба случая корректны, надо обработать, но и гарантировать
						 * выброс исключения в случае любого нарушения целостности.
						 */
						// Если ссылка на половинку НЕ указана
						if (empty($record[$needIdType])) {
							for ($i = 0; $i < count($list); $i++) {
								if (
									!empty($list[$i][$needIdType]) &&
									$list[$i][$needIdType] == $currentId &&
									v2_Budget::TYPE_MOVE == $list[$i]['operation_type']
								) {
									$record2   = $list[$i];
									$record2id = $i;
									break;
								}
							}
						}
						// Если ссылка на половинку указана
						else {
							$needLocalIdType = empty($record['server_id']) ? 'client_id' : 'server_id';
							for ($i = 0; $i < count($list); $i++) {
								if (
									!empty($list[$i][$needLocalIdType]) &&
									$record[$needIdType] == $list[$i][$needLocalIdType] &&
									v2_Budget::TYPE_MOVE == $list[$i]['operation_type']
								) {
									$record2   = $list[$i];
									$record2id = $i;
									break;
								}
							}
						}

						if (empty($record2)) {
							if (in_array($login, array('mxmiron@gmail.com'))) {
								throw new SoapFault('188', "Внимание!\nДля исправления сбоя в синхронизации, приложение необходимо переустановить.");
							}
							throw new SoapFault('65', "Move failed: '$needIdType' not found for move ID = $currentId");
						}
						if (($record['sum'] + $record2['sum'] != 0 && $record2['sum']) || $record['currency_id'] != $record2['currency_id']) {
							throw new SoapFault('66', "Move failed: '$needIdType' - sum value must be the same as first record, with minus");
						}
						if ($record['place_id'] != $record2['budget_object_id'] || $objId != $record2['place_id']) {
							throw new SoapFault('67', "Move failed: '$needIdType' - place ID must be equal to object ID of first record, and vice versa");
						}
						if (Constant::UNDEFINED == Family::getElementName($objId) && $objId >= 0) {
							if (DAO::getCount("SELECT id FROM budget_object WHERE id = $1", true, array($objId))) {
								throw new SoapFault('63', "Record failed: ".$objId." is not a valid ID");
							}
							v2_BudgetObject::addRootElement(Lang::id(100000528), User::familyId(), $record['operation_type'], $objId);
							$record['operation_type'] == v2_Budget::TYPE_WASTE ? Family::getCategoryList(true) : Family::getSourceList(true);
						}

						/**
						 * Проверки пройдены, составляем форму
						 */
						$move_id = empty($record['server_id']) ? null : $record['server_id'];
						if ($move_id) {
							$form->set("moveId", $move_id);
						}
						$form->set("sum_to_post", abs($record['sum']));
						$form->set("date_to_post", $record['operation_date']);
						$form->set("m_currency", $record['currency_id']);
						$form->set("m_comment", $record['comment']);
						$form->set("m_from", $record['sum'] < 0 ? $record['place_id'] : $record2['place_id']);
						$form->set("m_to", $record['sum'] < 0 ? $objId : $record2['budget_object_id']);
						/**
						 * Запускаем мега обработчик.
						 */
						try {
							$records = RecordUtils::processMove($form);
						}
						catch (BaseException $e) {
							$errMsg = $e->getMessage();
							if (mb_strpos($errMsg, 'должно отличаться от поля') !== false) {
								throw new SoapFault('188', "Ошибка: Исправьте перемещение от '{$record['operation_date']}'\nПоле 'из' должно отличаться от 'в'");
							}
							throw $e;
						}
						/**
						 * Сохраняем историю модификации если надо, для win32 клиентов.
						 * Делаем это в транзакции, что б улучшить целостность данных.
						 */
						if (count($records) != 2 && count($records) != 3) {
							throw new SoapFault('68', "Move failed: can't add two objects");
						}
						$records[0]->toServerHistory($move_id ? 'update' : 'insert');
						$records[1]->toServerHistory($move_id ? 'update' : 'insert');
						/**
						 * Формируем результирующий массив
						 */
						$retArrayItem1 = array(
							'server_id' => $records[0]->get(v2_BudgetRecord::ID),
							'status'    => 'updated'
						);
						$retArrayItem2 = array(
							'server_id' => $records[1]->get(v2_BudgetRecord::ID),
							'status'    => 'updated'
						);

						/**
						 * В ответе клиенту, положительная серверна половинка
						 * должна соответствовать положительной клиентской.
						 */
						if (!empty($record['client_id'])) {
							$retArrayItem1['client_id'] = $record['sum'] > 0 ? $record['client_id'] : $record2['client_id'];
							$retArrayItem1['status']    = 'inserted';
							$retArrayItem2['client_id'] = $record['sum'] < 0 ? $record['client_id'] : $record2['client_id'];
							$retArrayItem2['status']    = 'inserted';
						}
						$retArray[] = $retArrayItem1;
						$retArray[] = $retArrayItem2;

						$list[$key]['is_saved']       = true;
						$list[$record2id]['is_saved'] = true;
						break;
					case v2_Budget::TYPE_CHANGE:
						/**
						 * Если эта половинка уже сохранена, то значит и вторая сохранена. Пропускаем
						 */
						if (!empty($list[$key]['is_saved'])) {
							continue 2;
						}

						$needIdType = empty($record['server_id']) ? 'client_change_id' : 'server_change_id';
						$record2id  = null;
						$record2    = array();
						/**
						 * По историческим причинам, клиент вправе присылать ссылку как в обеих половинках
						 * так и только в одной. Оба случая корректны, надо обработать, но и гарантировать
						 * выброс исключения в случае любого нарушения целостности.
						 */
						// Если ссылка на половинку НЕ указана
						if (empty($record[$needIdType])) {
							for ($i = 0; $i < count($list); $i++) {
								if (
									!empty($list[$i][$needIdType]) &&
									$list[$i][$needIdType] == $currentId &&
									v2_Budget::TYPE_CHANGE == $list[$i]['operation_type']
								) {
									$record2   = $list[$i];
									$record2id = $i;
									break;
								}
							}
						}
						// Если ссылка на половинку указана
						else {
							$needLocalIdType = empty($record['server_id']) ? 'client_id' : 'server_id';
							for ($i = 0; $i < count($list); $i++) {
								if (
									!empty($list[$i][$needLocalIdType]) &&
									$record[$needIdType] == $list[$i][$needLocalIdType] &&
									v2_Budget::TYPE_CHANGE == $list[$i]['operation_type']
								) {
									$record2   = $list[$i];
									$record2id = $i;
									break;
								}
							}
						}

						if (empty($record2)) {
							throw new SoapFault('69', "Change failed: '$needIdType' not found for change ID = $currentId");
						}
						if (!$record['sum'] || $record['currency_id'] == $record2['currency_id']) {
							throw new SoapFault('70', "Change failed: '$needIdType' - currency value must not be the same");
						}
						if (($record['sum'] < 0 && $record2['sum'] < 0) || ($record['sum'] > 0 && $record2['sum'] > 0)) {
							throw new SoapFault('72', "Change failed: First sum must be positive, second sum must be negative");
						}
						if ($record['place_id'] != $objId || $record2['budget_object_id'] != $record2['place_id'] || $record['place_id'] != $record2['place_id']) {
							throw new SoapFault('71', "Change failed: '$needIdType' - place ID must be equal to object ID");
						}
						/**
						 * Проверки пройдены, составляем форму
						 */
						$change_id = empty($record['server_id']) ? null : $record['server_id'];
						if ($change_id) {
							$form->set("changeId", $change_id);
						}
						$form->set("sum_to_post", abs($record['sum']));
						$form->set("date_to_post", $record['operation_date']);
						$form->set("ch_to_currency", $record['sum'] < 0 ? $record2['currency_id'] : $record['currency_id']);
						$form->set("ch_sum_to", $record['sum'] < 0 ? $record2['sum'] : $record['sum']);
						$form->set("ch_from_currency", $record['sum'] < 0 ? $record['currency_id'] : $record2['currency_id']);
						$form->set("ch_sum", $record['sum'] < 0 ? abs($record['sum']) : abs($record2['sum']));
						$form->set("ch_comment", $record['comment']);
						$form->set("ch_from", $record['place_id']);
						/**
						 * Запускаем мега обработчик.
						 */
						$records = RecordUtils::processChange($form);
						/**
						 * Сохраняем историю модификации если надо, для win32 клиентов.
						 * Делаем это в транзакции, что б улучшить целостность данных.
						 */
						if (count($records) != 2 && count($records) != 3) {
							throw new SoapFault('73', "Change failed: can't add two objects");
						}
						$records[0]->toServerHistory($change_id ? 'update' : 'insert');
						$records[1]->toServerHistory($change_id ? 'update' : 'insert');
						/**
						 * Формируем результирующий массив
						 */
						$retArrayItem1 = array(
							'server_id' => $records[0]->get(v2_BudgetRecord::ID),
							'status'    => 'updated'
						);
						$retArrayItem2 = array(
							'server_id' => $records[1]->get(v2_BudgetRecord::ID),
							'status'    => 'updated'
						);

						/**
						 * В ответе клиенту, положительная серверна половинка
						 * должна соответствовать положительной клиентской.
						 */
						if (!empty($record['client_id'])) {
							$retArrayItem1['client_id'] = $record['sum'] > 0 ? $record['client_id'] : $record2['client_id'];
							$retArrayItem1['status']    = 'inserted';
							$retArrayItem2['client_id'] = $record['sum'] < 0 ? $record['client_id'] : $record2['client_id'];
							$retArrayItem2['status']    = 'inserted';
						}
						$retArray[] = $retArrayItem1;
						$retArray[] = $retArrayItem2;

						$list[$key]['is_saved']       = true;
						$list[$record2id]['is_saved'] = true;

						break;
				}
			}
			/**
			 * Перед тем, как успешно отправить список связок, надо грохнуть все
			 * клиентско-серверные связки данного типа, от прошлой синхры и добавить текущие.
			 */
			$delAppIdSql = $appId ? "AND app_instance_id = ".DB::esc($appId) : '';

			DB::query("DELETE FROM budget_sync_last
	    		WHERE user_id = $1
				AND api_id = $2
				AND object_type = $3
				$delAppIdSql",
				array(User::id(), Session::get(Constant::API_ID), BudgetHistory::TYPE_RECORD)
			);
			foreach ($retArray as $retValue) {
				if ($retValue['status'] != 'inserted') {
					continue;
				}
				BudgetSyncLast::create()->
				set(BudgetSyncLast::USER_ID, User::id())->
				set(BudgetSyncLast::API_ID, Session::get(Constant::API_ID))->
				set(BudgetSyncLast::OBJECT_TYPE, BudgetHistory::TYPE_RECORD)->
				set(BudgetSyncLast::CLIENT_ID, $retValue['client_id'])->
				set(BudgetSyncLast::SERVER_ID, $retValue['server_id'])->
				set(BudgetSyncLast::INSTANCE_ID, $appId)->
				set(BudgetSyncLast::DATE, 'now()')->
				daoSave();
			}
			DB::transactionCommit();
		}
		catch (Exception $e) {
			DB::transactionRollback();
			$faultCode = isset($e->faultcode) ? $e->faultcode : 0;
			if ($faultCode == '188') {
				throw $e;
			}
			$dataErrorMsg = array('Transfer failed: account_type', 'Query failed: ERROR:  -9', ', does not exist');
			foreach ($dataErrorMsg as $errorMsg) {
				if (strpos($e->getMessage(), $errorMsg) !== false) {
					throw new SoapFault('188', "Ошибка в структуре данных.\n".$conflictMsg);
				}
			}
			if (mb_strpos($e->getMessage(), 'is_free1000_overed') !== false) {
				throw new SoapFault('188', 'Превышен лимит операций бесплатной версии. См. раздел "Цены" на сайте.');
			}
			if (mb_strpos($e->getMessage(), 'create_only_access') !== false) {
				throw new SoapFault('188', "Настройки прав доступа не позволяют изменять операции\n".$conflictMsg);
			}
			$accessErrorMsg = array('forbidden_type_', 'el_forbidden', 'not_owner_obj', 'readonly_access', 'not_owner_rec');
			foreach ($accessErrorMsg as $errorMsg) {
				if (strpos($e->getMessage(), $errorMsg) !== false) {
					throw new SoapFault('188', "Доступ к скрытым элементам запрещён\n".$conflictMsg);
				}
			}
			$periodClosed = Date::periodClosed($e, false);
			if ($periodClosed) {
				throw new SoapFault('188', $periodClosed);
			}
			self::checkNameUniq($e->getMessage());
			Debug::mail($e);
			throw new SoapFault('61', $e->getMessage());
		}

		return $retArray;
	}

	/**
	 * Delete any object; [id] => ID of the object to delete; [type] => The type of the object, must be one of: 'waste' 'income' 'move' 'change' 'object' 'currency' 'tag' 'accum' 'check'; 'object' is waste category, income source or place; If 'id' identifies 'move' or 'change', both records will be deleted on the server; Returns 1 on success; if an error accures - generates SoapFault message; if there is other object connected to this ID - delete them first;
	 *
	 * @param string $apiId
	 * @param string $login
	 * @param string $pass
	 * @param integer $id
	 * @param string $type
	 * @return integer
	 */
	public function deleteObject($apiId, $login, $pass, $id, $type)
	{
		/**
		 * Для функций записи должен быть активен Premium
		 */
		try {
			$this->getAccessStatus($apiId, $login, $pass);
		}
		catch (SoapFault $e) {
			if ($e->faultcode != '87') { // trial
				throw $e;
			}
		}

		$allowedTypes = array('waste', 'income', 'move', 'change', 'object', 'currency', 'tag', 'accum', 'check');

		if (!in_array($type, $allowedTypes)) {
			throw new SoapFault('77', "Parameter type must be one of: 'waste' 'income' 'move' 'change' 'object' 'currency' 'tag' 'accum' 'check'");
		}
		$this->checkNumeric($id, $type);
		if (!$id) {
			throw new SoapFault('78', "Parameter ID must be positive integer");
		}
		$retStatus = 1;
		/**
		 * Тип тоже определим сами, в Android приложении замечен баг
		 */
		if (in_array($type, array('waste', 'income', 'move', 'change'))) {
			$type = DAO::getCount("SELECT (CASE WHEN budget_record.is_currency_change = 't' THEN 5 ELSE budget_object.type END) 
				FROM budget_record JOIN budget_object ON (budget_object_id = budget_object.id) AND budget_record.id = $1",
				true, array($id)
			);
			if (!$type) {
				return $retStatus;
			}
		}

		$conflictMsg = 'Для восстановления синхронизации, ';
		$conflictMsg .= ($apiId == 'vandr_api' ? "необходимо перезайти через меню\n'Ещё->Сменить пользователя'." : 'приложение необходимо переустановить.');
		/**
		 * Базовые проверки пройдены. Проверки на семью в соотв. обработчиках
		 */
		$form = new Form();
		$form->clear();
		switch ($type) {
			case v2_Budget::TYPE_WASTE:
			case v2_Budget::TYPE_INCOME:
				$form->set(($type == v2_Budget::TYPE_WASTE ? 'waste' : 'income')."Id", $id);
				try {
					DB::transactionBegin();
					$budgetRecord = RecordUtils::deleteRecord($form, $type == v2_Budget::TYPE_WASTE ? true : false);
					$budgetRecord->toServerHistory('delete');
					DB::transactionCommit();
				}
				catch (Exception $e) {
					DB::transactionRollback();
					$errorMessage = $e->getMessage();
					$rep          = array('&#34;', '&quot;', '"');
					$errorMessage = str_replace($rep, '', $errorMessage);
					if (mb_strpos($errorMessage, "budget_record_links")) {
						$errorMessage = "Вы пытаетесь удалить трату, которая является комиссией для перемещения. Она автоматически удаляется, при удалении перемещения.";
					}
					$periodClosed = Date::periodClosed($e, false);
					if ($periodClosed) {
						throw new SoapFault('188', $periodClosed);
					}
					if (mb_strpos($e->getMessage(), 'create_only_access') !== false) {
						throw new SoapFault('188', "Настройки прав доступа не позволяют удалять операции\n".$conflictMsg);
					}
					$accessErrorMsg = array('forbidden_type_', 'el_forbidden', 'not_owner_obj', 'readonly_access', 'not_owner_rec');
					foreach ($accessErrorMsg as $errorMsg) {
						if (strpos($e->getMessage(), $errorMsg) !== false) {
							throw new SoapFault('188', "Доступ к скрытым элементам запрещён\n".$conflictMsg);
						}
					}
					/**
					 * Это значит, что запись уже удалена, ничего страшного.
					 */
					if ('Invalid record' != $errorMessage) {
						Debug::mail($e);
						throw new SoapFault('79', $errorMessage);
					}
				}
				break;
			case v2_Budget::TYPE_MOVE:
			case v2_Budget::TYPE_CHANGE:
				try {
					RecordUtils::deleteMoveNoSync($id, $type == v2_Budget::TYPE_CHANGE ? true : false);
				}
				catch (Exception $e) {
					$errorMessage = $e->getMessage();
					$rep          = array('&#34;', '&quot;', '"');
					$errorMessage = str_replace($rep, '', $errorMessage);
					$periodClosed = Date::periodClosed($e, false);
					if ($periodClosed) {
						throw new SoapFault('188', $periodClosed);
					}
					if (mb_strpos($e->getMessage(), 'create_only_access') !== false) {
						throw new SoapFault('188', "Настройки прав доступа не позволяют удалять операции\n".$conflictMsg);
					}
					$accessErrorMsg = array('forbidden_type_', 'el_forbidden', 'not_owner_obj', 'readonly_access', 'not_owner_rec');
					foreach ($accessErrorMsg as $errorMsg) {
						if (strpos($e->getMessage(), $errorMsg) !== false) {
							throw new SoapFault('188', "Доступ к скрытым элементам запрещён\n".$conflictMsg);
						}
					}
					/**
					 * Это значит, что запись уже удалена, ничего страшного.
					 */
					if ('Invalid record' != $errorMessage) {
						Debug::mail($e);
						throw new SoapFault('80', $e->getMessage());
					}
				}
				break;
			case 'currency':
				try {
					DB::transactionBegin();
					/**
					 * Удаляем начальные остатки, а если есть другие записи по этой валюте - то сработает констрэйнт и транзакция откатится.
					 */
					$currentElement = new v2_Currency($id);
					/**
					 * Если это валюта по умолчанию - не велим
					 */
					if ($currentElement->get(v2_Currency::IS_DEFAULT) == 't') {
						throw new BaseException('Default currency is not allowed to delete');
					}

					DB::query("DELETE FROM budget_record WHERE budget_object_id = -1 AND currency_id = $1 AND budget_family_id = $2",
						array($id, User::familyId())
					);
					$deleteStatus = DB::query("DELETE FROM currency WHERE id = $1 AND family_id = $2",
						array($id, User::familyId())
					);
					$deleteStatus = pg_affected_rows($deleteStatus);

					DB::transactionCommit();
					$currentElement->toServerHistory('delete');
				}
				catch (Exception $e) {
					DB::transactionRollback();
					Debug::mail($e);
					throw new SoapFault('81', $e->getMessage());
				}
				break;
			case 'object':
				try {
					$currentElement = v2_BudgetObject::create($id);
					DB::query("DELETE FROM budget_record WHERE budget_object_id = -1 AND budget_account_id = $1 AND budget_family_id = $2",
						array($id, User::familyId())
					);
					$deleteStatus = DB::query("DELETE FROM budget_object WHERE id = $1 AND budget_family_id = $2",
						array($id, User::familyId())
					);
					$deleteStatus = pg_affected_rows($deleteStatus);
					if ($deleteStatus != 1) {
						throw new BaseException("Кол-во удалённых строк = $deleteStatus, а должно быть 1 !!!");
					}
					$currentElement->toServerHistory('delete');
				}
				catch (Exception $e) {
					throw new SoapFault('82', $e->getMessage());
				}
				break;
			case 'tag':
				/**
				 * Надо из комментов всех трат тоже удалить этот тэг.
				 */
				try {
					DB::transactionBegin();
					$currentElement = new BudgetTag($id);
					$oldName        = $currentElement->get(BudgetTag::NAME);
					$recordList     = v2_BudgetRecord::create()->daoMakeObjectList("SELECT * FROM budget_record WHERE id IN (
							SELECT budget_record_id FROM budget_record_tags WHERE budget_tag_id = $1
						) LIMIT 1000", false,
						array($currentElement->get(BudgetTag::ID))
					);
					foreach ($recordList as $record) {
						$coment = $record->get(v2_BudgetRecord::COMMENT);
						if (($tagPos = mb_strpos(mb_strtolower($coment), mb_strtolower("[".$oldName."]"))) === false) {
							continue;
						}
						$coment = FormUtil::mb_substr_replace($coment, "", $tagPos, mb_strlen("[".$oldName."]"));
						$record->set(v2_BudgetRecord::COMMENT, $coment);
						$record->daoSave();
						$record->toServerHistory('update');
					}

					$deleteStatus = DB::query("DELETE FROM budget_tags WHERE id = $1 AND family_id = $2",
						array($id, User::familyId())
					);
					$deleteStatus = pg_affected_rows($deleteStatus);
					if ($deleteStatus != 1) {
						throw new BaseException("Кол-во удалённых строк = $deleteStatus, а должно быть 1 !!!");
					}
					DB::transactionCommit();
				}
				catch (Exception $e) {
					DB::transactionRollback();
					throw new SoapFault('83', $e->getMessage());
				}
				break;
			case 'accum':
				try {
					DB::query("DELETE FROM budget_accum WHERE id = $1 AND family_id = $2",
						array($id, User::familyId())
					);
				}
				catch (Exception $e) {
					throw new SoapFault('94', $e->getMessage());
				}
				break;
			case 'check':
				try {
					DB::transactionBegin();
					/**
					 * Удалить из комментария всех привязанных трат строку вида [чeк 23232]
					 */
					DB::query("UPDATE budget_record SET comment = regexp_replace(comment, '\[чeк [^\]]+\]', '') WHERE 
						id IN (SELECT record_id FROM budget_check_record WHERE check_id = $1)
						AND budget_family_id = $2",
						array($id, User::familyId())
					);
					DB::query("DELETE FROM budget_check WHERE id = $1 AND user_id IN (SELECT id FROM users WHERE family_id = $2)",
						array($id, User::familyId())
					);
					DB::transactionCommit();
				}
				catch (Exception $e) {
					DB::transactionRollback();
					Debug::mail($e);
					throw new SoapFault('118', $e->getMessage());
				}
				break;
		}
		return $retStatus;
	}

	/**
	 * Delete all data (transactions, waste category, income source, place list, currency and tags; Used for synchronization in case of the conflicts; Returns void if success, or throws SoapFault exception;
	 *
	 * @param string $apiId
	 * @param string $login
	 * @param string $pass
	 * @return void
	 */
	public function deleteAll($apiId, $login, $pass)
	{
		/**
		 * Для функций записи должен быть активен Premium
		 */
		try {
			$this->getAccessStatus($apiId, $login, $pass);
		}
		catch (SoapFault $e) {
			if ($e->faultcode != '87') { // trial
				throw $e;
			}
		}

		try {
			$familyId = Session::get(User::FAMILY_ID);
			$userId   = User::id();

			if (User::isLimited()) {
				throw new SoapFault('95', 'User is not accessed to delete all');
			}

			DB::transactionBegin();

			$users     = User::create()->daoMakeObjectList("SELECT * FROM users WHERE family_id = $1", false, array($familyId));
			$userCount = count($users);

			/**
			 * Футболим Демо юзера
			 */
			if ($userId == Constant::DEMO_USER_ID) {
				throw new BaseException("Извините, пользователю demo не разрешено всё чистить.");
			}

			// Удаляем все связки половинок перемещений и обменов для данного юзера
			DB::query("DELETE FROM budget_record_links WHERE record1 IN (
					SELECT id FROM budget_record WHERE user_nuid = $1 AND budget_family_id = $2
				) OR record2 IN (
					SELECT id FROM budget_record WHERE user_nuid = $1 AND budget_family_id = $2
				)",
				array($userId, $familyId)
			);

			// удаляем планы
			DB::query("DELETE FROM budget_plan WHERE family_id = $1 AND parent_id = -1", array($familyId));

			// Чистим все записи
			DB::query("DELETE FROM budget_record WHERE user_nuid = $1 AND budget_family_id = $2", array($userId, $familyId));

			// Удаляем инфу из облаков тэгов.. пускай пока только для того юзера семьи кот. удаляет
			DB::query("DELETE FROM budget_object_count WHERE nuid = $1", array($userId));

			// Удаляем данные из кэширующий синхронизацию таблицы
			DB::query("DELETE FROM budget_sync_last WHERE user_id = $1", array($userId));

			/**
			 * Если есть нес-ко членов семьи и выбрано "удалить только мой аккаунт", оставляем
			 * справочники и т.д.
			 */
			if ($userCount > 1) {
				foreach ($users as $user) {
					// Удаляем все связки половинок перемещений и обменов для данного юзера
					DB::query("DELETE FROM budget_record_links WHERE record1 IN (
							SELECT id FROM budget_record WHERE user_nuid = $1 AND budget_family_id = $2
						) OR record2 IN (
							SELECT id FROM budget_record WHERE user_nuid = $1 AND budget_family_id = $2
						)",
						array($user->get(User::ID), $familyId)
					);

					// Чистим все записи
					DB::query("DELETE FROM budget_record WHERE user_nuid = $1 AND budget_family_id = $2", array($user->get(User::ID), $familyId));
					// Удаляем инфу из облаков тэгов.. пускай пока только для того юзера семьи кот. удаляет
					DB::query("DELETE FROM budget_object_count WHERE nuid = $1", array($user->get(User::ID)));
				}
			}
			// Грохаем справочники строго после гроханья всех записей.
			DB::query("DELETE FROM budget_object WHERE budget_family_id = $1", array($familyId));

			DB::transactionCommit();
		}
		catch (Exception $e) {
			DB::transactionCommit();
			throw new SoapFault('88', 'asas');
		}
	}

	/**
	 * Get current payment status; Returns 1 if Ok, throw SoapFault otherwise.
	 *
	 * @param string $apiId
	 * @param string $login
	 * @param string $pass
	 * @return integer
	 */
	public function getAccessStatus($apiId, $login, $pass)
	{
		$this->login($apiId, $login, $pass);

		/**
		 * Иногда для отладки нужно разрешить синхронизацию демо-пользователю
		 */
		if ($login == 'demo@example.com') {
//			return strtotime("2025-01-01");
		}

		if ($apiId == 'demo_api' && $login != 'demo@example.com') {
			throw new SoapFault('85', "No 'demo_api' allowed to real users");
		}

		if (
			User::id() != Session::get(Constant::API_AUTHOR_ID) &&
			!Billing::checkAccess(User::id(), $apiId)
		) {
			throw new SoapFault('84', "No payment");
		}

		/**
		 * В новых версиях в случае успеха отдаём дату окончания подписки.
		 * Это выгоднее.
		 * Новые версии это:
		 * Android >= 2.1.12
		 * iPhone > ...
		 * WP7 > ...
		 */
		$isReturnExpireDate = ($apiId == 'vandr_api' && self::isMoreVersion(array(2, 1, 12)) ? true : false);

		if ($isReturnExpireDate) {
			/**
			 * Для автора приложения - вечная подписка
			 */
			if (User::id() == Session::get(Constant::API_AUTHOR_ID)) {
				return time() + 3600 * 24 * 365;
			}

			$date = $this->getExpireDate($apiId, $login, $pass);
			if ($date <= 0) {
				Debug::mail($login.":".$date);
			}
			$date > 0 ? $date = strtotime($date) : '';
			return $date;
		}

		return 1;
	}

	/**
	 * Get all changes (array of arrays) from server relative to given revision: [revision] => the revision of the change, [action_id] => the action of the change '1' - add, '2' - update, '3' - delete'; [object_type_id] => type of the object changed '1' - any record (transction), '2' - income source, '3' - waste category, '4' - place, '5' - currency, '6' - budget_tags, '7' - budget_accum, '8' - budget_accum_order; [object_id] => ID of the object for subsequent calls getRecordList, getCategoryList etc; [date] => the date of the change; Parameter [revision] => int8 number, usually saved on the client from last successfull sync.
	 *
	 * @param string $apiId
	 * @param string $login
	 * @param string $pass
	 * @param string $revision
	 * @return array
	 */
	public function getChangeList($apiId, $login, $pass, $revision)
	{
		$this->login($apiId, $login, $pass);
		$this->checkNumeric($revision, 'getChanges');

		/**
		 * С версии 2.3.0 в Андроид накопления разрешаем.
		 */
		$isAccumAllow = false;
		if ('vandr_api' == $apiId && self::isMoreVersion(array(2, 3, 0))) {
			$isAccumAllow = true;
		}
		$objTypeSql = ($apiId == 'vandr_api' && !$isAccumAllow ? "AND object_type_id NOT IN ('7', '8')" : '');

		/**
		 * С версии 2.5.9 Андроид отдаём чеки и связки чеков с тратами (там только QR).
		 */
		$isQrAllow = false;
		if ('vandr_api' == $apiId && self::isMoreVersion(array(2, 5, 9))) {
			$isQrAllow = true;
		}
		$objQrSql = ($isQrAllow ? '' : "AND object_type_id NOT IN ('10', '11')");

		/**
		 * Группы счетов пока никому не отдаём, а будем отдавать по мере добавления поддержки
		 * в каждом из клиентов.
		 */
		$isPlaceGroupAllow = false;
		if ('vandr_api' == $apiId && self::isMoreVersion(array(2, 3, 14))) {
			$isPlaceGroupAllow = true;
		}
		$placeGroupSql = $isPlaceGroupAllow ? '' : "AND object_type_id <> '".v2_Budget::TYPE_GROUP_PLACE."'";

		/**
		 * Эти запросы долгие, пока не понятно как их ускорить. Не мониторим.
		 */
		DB::$minDuration = 30;

		$tmpChanges = DAO::fillArray("SELECT id AS revision,action_id,
    		(CASE WHEN object_type_id = '".v2_Budget::TYPE_GROUP_PLACE."' THEN '".BudgetHistory::TYPE_PLACE."' ELSE trim(object_type_id) END) AS object_type_id,
    		object_id,date
    		FROM budget_history WHERE
    		id > $1 AND family_id = $2
    		$objTypeSql
    		$objQrSql
    		$placeGroupSql
    		ORDER BY id", null,
			array($revision, User::familyId())
		);

		/**
		 * Надо выбросить взаимоисключающие "удаления" и "добавления", из-за архивирования и последующего
		 * восстановления юзера из архива
		 */
		$returnChanges = array();
		$idxToIgnore   = array();
		$changesByType = array();
		foreach ($tmpChanges as $idx => $change) {
			$oType = $change['object_type_id'];
			$oId   = $change['object_id'];
			$aId   = $change['action_id'];
			isset($changesByType[$oType]) ? '' : $changesByType[$oType] = array();
			isset($changesByType[$oType][$oId]) ? '' : $changesByType[$oType][$oId] = array();

			$lastActionId = ArrayUtils::safeGet($changesByType[$oType][$oId], 'a_id');
			if ('3' == $lastActionId && '1' == $aId) {
				$idxToIgnore[$changesByType[$oType][$oId]['idx']] = '';
				$idxToIgnore[$idx]                                = '';
			}

			$changesByType[$oType][$oId]['a_id'] = $aId;
			$changesByType[$oType][$oId]['idx']  = $idx;
		}

		if (empty($idxToIgnore)) {
			return $tmpChanges;
		}

		foreach ($tmpChanges as $idx => $change) {
			isset($idxToIgnore[$idx]) ? '' : $returnChanges[] = $change;
		}

//		Debug::mail("idxToIgnore count:\t".count($idxToIgnore)."\nreturnChanges count:\t".count($returnChanges)."\n");

		return $returnChanges;
	}

	/**
	 * Get current server revision number. This method must be called at the end of synchronization, after all client changes saved on server; The number of current server revision must be saved on the client, for the next synchronization;
	 *
	 * @param string $apiId
	 * @param string $login
	 * @param string $pass
	 * @return integer
	 */
	public function getCurrentRevision($apiId, $login, $pass)
	{
		$this->login($apiId, $login, $pass);
		$rev = DAO::getCount("SELECT cur_revision FROM budget_family WHERE id = $1", true, array(User::familyId()));
		$rev = empty($rev) ? 0 : $rev;
		return $rev;
	}

	/**
	 * Get internal user ID by login;
	 *
	 * @param string $apiId
	 * @param string $login
	 * @param string $pass
	 * @return string
	 */
	public function getUserIdByLogin($apiId, $login, $pass)
	{
		$this->login($apiId, $login, $pass);
		return User::id();
	}

	/**
	 * Get user name by user ID; For multi user mode;
	 *
	 * @param string $apiId
	 * @param string $login
	 * @param string $pass
	 * @return string
	 */
	public function getUserNameById($apiId, $login, $pass, $userId)
	{
		$this->login($apiId, $login, $pass);
		$this->checkNumeric($userId, 'users');
		$name = DAO::getCount("SELECT name FROM users WHERE id = $1 AND family_id = $2", true, array($userId, User::familyId()));
		return FormUtil::toHtml($name);
	}

	public function getExpireDate($apiId, $login, $pass)
	{
		$this->login($apiId, $login, $pass);
		return Internal::getExpireDate($apiId, $login, $pass);
	}

	/**
	 * Reqister new user; Return user password on successfully register; If an error accur return -1 (such email already exist), -2 (register not allowed), -3 (email incorrect), -4 (other error); [lang] - 'en' or 'ru';
	 *
	 * @param string $apiId
	 * @param string $login
	 * @param string $name
	 * @param string $lang
	 * @return string
	 */
	public function userRegister($apiId, $login, $name, $lang)
	{
		if (!in_array($apiId, array('ktt_api', 'ktt_ipad_api', 'vandr_api', 'wp_api'))) {
			return '-2';
		}
		$newpass = "-4";
		try {
			$refId = $this->getRefId();

			Lang::setLang($lang == 'en' ? Lang::EN : Lang::RU);
			DB::transactionBegin();

			$user = User::create();

			$login = trim($login);
			$login = $this->normalizeString(mb_strtolower($login), 64, "userName");

			$user->set(User::MAIL, $login);
			$user->set(User::NICK, $name);
			$user->set(User::NAME, $name);
			$user->set(User::IS_SEND_NEWS, true);
			$user->set(User::IS_SEND_REMIND, true);
			$user->set(User::REG_DATE, 'now()');
			$user->set(User::LANG, Lang::getLang());
			$user->set(User::MESSAGE_COUNT, 0);
			$seq_val = DB::query("SELECT nextval('users_id_seq')");
			$seq_val = pg_fetch_result($seq_val, 0, 0);
			$user->set(User::ID, $seq_val);

			$user->set(User::REF_INIT, $apiId.' '.$refId);
			$user->set(User::SOURCE_ID, Constant::getSourceId('www.drebedengi.ru'));

			// Генерация пароля
			$newpass    = User::genPass();
			$crypt_pass = User::cryptPass($newpass);
			$user->set(User::PASSWORD, $crypt_pass);

			// Отправка его на мыло
			// Проверим-ка мыло на корректность (базово)
			$at_pos = mb_strpos($login, "@");
			if (!(mb_strlen($login) > 5) || !($at_pos > 0) || mb_strpos($login, ' ')) {
				return '-3';
			}
			$regModule = ModuleFactory::spawn('v2_register');
			$regModule->regUser($user, $newpass);

			DB::transactionCommit();

			$refId && 'google' != $refId ? Debug::mail("$refId: $login, $name, $lang") : '';
		}
		catch (Exception $e) {
			DB::transactionRollback();
			if (mb_strpos($e->getMessage(), 'mail_uniq') !== false) {
				return '-1';
			}
			if (mb_strpos($e->getMessage(), 'Почтовое сообщение не было отправлено') !== false) {
				return '-3';
			}
			Debug::mail($e);
			return '-4';
		}
		return $newpass;
	}

	/**
	 * Gets list of orders for purchases; If parameter [idList] is given, it will be treat as ID list of objects to retrieve# this is used for synchronization;
	 *
	 * @param string $apiId
	 * @param string $login
	 * @param string $pass
	 * @param array $idList
	 * @return array
	 */
	public function getOrderList($apiId, $login, $pass, $idList)
	{
		$this->login($apiId, $login, $pass);

		$sqlIdList = '';
		if ($this->checkIdList($idList, 'getOrderList')) {
			$sqlIdList = 'id IN ('.DB::in($idList).') AND';
		}

		$orderList = DAO::fillArray("SELECT id, name, user_id, is_family, is_mobile FROM budget_accum_order
    		WHERE $sqlIdList (
				(user_id = $1 AND family_id = $2) OR
				(is_family IS TRUE AND family_id = $2)
			)
    		AND is_mobile IS TRUE
    		ORDER BY id", null,
			array(User::id(), User::familyId())
		);
		foreach ($orderList as $key => $order) {
			$orderList[$key]['name'] = FormUtil::toHtml($order['name']);
		}
		return $orderList;
	}

	/**
	 * Gets list of purchases (accums); For purchase list logic is only needed fields 'id, name, sum_need, sum_currency_id, user_id, is_family, created, comment, priority, order_id', other fields is for accum logic and should not be changed; ; If parameter [idList] is given, it will be treat as ID list of objects to retrieve# this is used for synchronization;
	 *
	 * @param string $apiId
	 * @param string $login
	 * @param string $pass
	 * @param array $idList
	 * @return array
	 */
	public function getAccumList($apiId, $login, $pass, $idList)
	{
		$this->login($apiId, $login, $pass);

		$sqlIdList = '';
		if ($this->checkIdList($idList, 'getAccumList')) {
			$sqlIdList = 'id IN ('.DB::in($idList).') AND';
		}

		$limitedSql = defined('IS_SERVER') ? '' : BudgetAccum::getLimitSql();

		$accumList = DAO::fillArray("
			SELECT * FROM budget_accum WHERE $sqlIdList (
				(user_id = $1 AND family_id = $2) OR
				(is_family IS TRUE AND family_id = $2)
			)
			$limitedSql
    		ORDER BY is_accum DESC, priority", null,
			array(User::id(), User::familyId())
		);

		foreach ($accumList as $key => $accum) {
//     		$accumList[$key]['name'] = FormUtil::toHtml($accum['name']);
			$accumList[$key]['comment'] = FormUtil::toHtml($accum['comment'], true);
		}

		return $accumList;
	}

	/**
	 * Insert or update accum list; [list] => array of arrays: 'server_id' or 'client_id' [int8] - server or client ID of the record# If client ID is present - try to insert new record, and return server2client correspondence in the result array# If server_id is present - try to update existing record, @param string $apiId
	 * @param string $login
	 * @param string $pass
	 * @param array $list
	 * @return array
	 * @see getAccumList description for other detail; Returns the array of server IDs, successfully changed; The client MUST save server IDs corresponded to client IDs, for subsequent 'update' and 'delete' calls;
	 *
	 */
	public function setAccumList($apiId, $login, $pass, $list)
	{
		/**
		 * Для функций записи должна быть активна подписка
		 */
		try {
			$this->getAccessStatus($apiId, $login, $pass);
		}
		catch (SoapFault $e) {
			if ($e->faultcode != '87') { // trial
				throw $e;
			}
		}

		/**
		 * Проверки
		 */
		$this->checkEmptyArray($list, '89', 'Accum list');

		$retArray = array();

		try {
			DB::transactionBegin();
			$dublDebug    = array();
			$lastSyncData = DAO::fillArrayKeyId("SELECT client_id AS id, server_id, app_instance_id FROM budget_sync_last
    			WHERE user_id = $1
    			AND api_id = $2
    			AND object_type = $3",
				array(User::id(), Session::get(Constant::API_ID), BudgetHistory::TYPE_ACCUM)
			);
			$appId        = self::getAppId();

			foreach ($list as $accum) {
				/**
				 * Первым делом, если это добавление клиентской записи, которая
				 * уже была добавлена в прошлой синхре, то значит был сбой на клиенте
				 * и он шлёт дубль. Ничего не добавляем, а отдаём тот же server_id
				 * что и в прошлый раз.
				 */
				$isClientIdThisInstance = true;
				if (
					abs($appId) > 0 &&
					!empty($accum['client_id']) &&
					isset($lastSyncData[$accum['client_id']]) &&
					abs($lastSyncData[$accum['client_id']]['app_instance_id']) > 0 &&
					$lastSyncData[$accum['client_id']]['app_instance_id'] != $appId
				) {
					$isClientIdThisInstance = false;
				}

				if (!empty($accum['client_id']) && isset($lastSyncData[$accum['client_id']]) && $isClientIdThisInstance) {
					$retArray[]  = array(
						'server_id' => $lastSyncData[$accum['client_id']]['server_id'],
						'client_id' => $accum['client_id'],
						'status'    => 'inserted'
					);
					$dublDebug[] = array(
						'server_id' => $lastSyncData[$accum['client_id']]['server_id'],
						'client_id' => $accum['client_id'],
						'type'      => 'accum',
					);
					continue;
				}

				/**
				 * Базовые проверки
				 */
				if (empty($accum['server_id']) && empty($accum['client_id'])) {
					throw new SoapFault('90', "'server_id' or 'client_id' should be given");
				}
				$accumId                          = empty($accum['server_id']) ? '' : $accum['server_id'];
				$accum[BudgetAccum::FAMILY_ID]    = Session::get(User::FAMILY_ID);
				$accum[BudgetAccum::USER_ID]      = User::id();
				$accum[BudgetAccum::ID]           = $accumId ? $accumId : null;
				$accum[BudgetAccum::DATE_CREATED] = str_replace(array(' PM', ' AM'), '', $accum[BudgetAccum::DATE_CREATED]);

				$accumObj = BudgetAccum::create();
				$this->checkObject($accumObj, $accum, 'BudgetAccum', true);

				/**
				 * Кастомные проверки накопления и сохранение.
				 */
				if ($accumId) {
					$famCount = DAO::getCount("SELECT COUNT(id) FROM budget_accum WHERE id = $1 AND family_id = $2", true,
						array($accumId, User::familyId())
					);
					if ($famCount == 0) { // Удалили на сервере
						$retArray[] = array(
							'server_id' => $accumId,
							'status'    => 'updated'
						);
						continue;
					}
					elseif ($famCount != 1) {
						throw new SoapFault('91', "Accum object with id = $accumId is invalid");
					}
					$accumObj->get(BudgetAccum::SORT) ? '' : $accumObj->set(BudgetAccum::SORT, "0");
				}
				/**
				 * Для новых покупок по умолчанию выставляем, что они семейные.
				 * Вне зависимости от того, что там на клиенте
				 */
				else {
					$accumObj->set(BudgetAccum::IS_FAMILY, true);
				}

				if ($accum[BudgetAccum::SUM_CURRENCY_ID]) {
					$curId    = $this->checkNumeric($accum[BudgetAccum::SUM_CURRENCY_ID], "BudgetAccum");
					$famCount = DAO::getCount("SELECT family_id FROM currency WHERE id = $1 AND family_id = $2", true,
						array($curId, User::familyId())
					);
					if ($famCount != Session::get(User::FAMILY_ID)) {
						throw new SoapFault('92', "Accum object with id = $accumId is invalid");
					}
					$accumObj->
					set(BudgetAccum::SUM_CURRENCY_ID, $curId)->
					set(BudgetAccum::SUM_NEED, $this->checkNumeric($accum[BudgetAccum::SUM_NEED], "BudgetAccum"));

					$accumObj->get(BudgetAccum::SUM_NEED) ? '' : $accumObj->set(BudgetAccum::SUM_NEED, "0");
				}
				else {
					$accumObj->set(BudgetAccum::SUM_NEED, null);
				}

				/**
				 * Проверим если очередь уже удалили на сервере, создадим заново "очередь заглушку",
				 * во избежание конфликта синхронизации
				 */
				$orderId = $accumObj->get(BudgetAccum::ORDER_ID);
				if (!DAO::getCount("SELECT id FROM budget_accum_order WHERE id = $1", true, array($orderId))) {
					BudgetAccumOrder::create()->
					set(Business::ID, $orderId)->
					set(BudgetAccumOrder::USER_ID, User::id())->
					set(BudgetAccumOrder::FAMILY_ID, User::familyId())->
					set(BudgetAccumOrder::IS_FAMILY, true)->
					set(BudgetAccumOrder::IS_MOBILE, true)->
					set(BudgetAccumOrder::NAME, 'Удалённая очередь')->
					daoSave(true);
					Debug::mail("Создана удалённая очередь id = $orderId");
				}

				$accumObj->
				set(BudgetAccum::COMMENT, $this->normalizeString(str_replace('<br />', "\n", $accum[BudgetAccum::COMMENT]), 2048, "BudgetAccum"))->
				set(BudgetAccum::NAME, mb_substr($accum[BudgetAccum::NAME], 0, 128))->
				daoSave();

				$retArrayItem = array(
					'server_id' => $accumObj->get(BudgetAccum::ID),
					'status'    => $accumId ? 'updated' : 'inserted'
				);
				if (!$accumId) {
					$retArrayItem['client_id'] = $accum['client_id'];
				}
				$retArray[] = $retArrayItem;
			}
			/**
			 * Перед тем, как успешно отправить список связок, надо грохнуть все
			 * клиентско-серверные связки данного типа, от прошлой синхры и добавить текущие.
			 */
			$delAppIdSql = $appId ? "AND app_instance_id = ".DB::esc($appId) : '';
			DB::query("DELETE FROM budget_sync_last
	    		WHERE user_id = $1
				AND api_id = $2
				AND object_type = $3
				$delAppIdSql",
				array(User::id(), Session::get(Constant::API_ID), BudgetHistory::TYPE_ACCUM)
			);
			foreach ($retArray as $retValue) {
				if ($retValue['status'] != 'inserted') {
					continue;
				}
				BudgetSyncLast::create()->
				set(BudgetSyncLast::USER_ID, User::id())->
				set(BudgetSyncLast::API_ID, Session::get(Constant::API_ID))->
				set(BudgetSyncLast::OBJECT_TYPE, BudgetHistory::TYPE_ACCUM)->
				set(BudgetSyncLast::CLIENT_ID, $retValue['client_id'])->
				set(BudgetSyncLast::SERVER_ID, $retValue['server_id'])->
				set(BudgetSyncLast::INSTANCE_ID, $appId)->
				set(BudgetSyncLast::DATE, 'now()')->
				daoSave();
			}
			DB::transactionCommit();
			/**
			 * Для отладки шлём уведомления о дублях.
			 * Когда станет не нужным - убрать.
			 */
			if ($dublDebug) {
				Debug::mail($dublDebug);
			}
		}
		catch (Exception $e) {
			DB::transactionRollback();
			Debug::mail($e);
			throw new SoapFault('93', $e->getMessage());
		}
		return $retArray;
	}

	/**
	 * Gets list of checks; Only QR checks is synced; Return array of arrays: [id] => Internal check ID; [ext] => qr-url; [state] => the state of qr check process; [qr_sum] => the sum of check from QR; [qr_date] => the date of check from QR; If parameter [idList] is given, it will be treat as ID list of objects to retrieve# this is used for synchronization;
	 *
	 * @param string $apiId
	 * @param string $login
	 * @param string $pass
	 * @param array $idList
	 * @return array
	 */
	public function getCheckList($apiId, $login, $pass, $idList)
	{
		/**
		 * @todo Добавить поддержку серверной версии
		 */
		if (defined('IS_SERVER')) {
			return array();
		}

		$this->login($apiId, $login, $pass);

		$sqlIdList = '';
		if ($this->checkIdList($idList, 'getCheckList')) {
			$sqlIdList = 'id IN ('.DB::in($idList).') AND';
		}

		$limitSql = User::isLimited() ? ' AND user_id = '.User::id() : '';

		$checkList = DAO::fillArray("
			SELECT * FROM budget_check WHERE $sqlIdList user_id IN (
				SELECT id FROM users WHERE family_id = $1
			)
			$limitSql
			AND qr_sum > 0 ORDER BY qr_date",
			null, array(User::familyId())
		);

		return $checkList;
	}

	/**
	 * Gets list of check to record link; Only for QR checks; Return array of arrays: [id] => Internal link ID; [check_id] => ID of the check; [record_id] => ID of the record; If parameter [idList] is given, it will be treat as ID list of objects to retrieve# this is used for synchronization;
	 *
	 * @param string $apiId
	 * @param string $login
	 * @param string $pass
	 * @param array $idList
	 * @return array
	 */
	public function getCheckToRecordList($apiId, $login, $pass, $idList)
	{
		/**
		 * @todo Добавить поддержку серверной версии
		 */
		if (defined('IS_SERVER')) {
			return array();
		}

		$this->login($apiId, $login, $pass);

		$sqlIdList = '';
		if ($this->checkIdList($idList, 'getCheckToRecordList')) {
			$sqlIdList = 'id IN ('.DB::in($idList).') AND';
		}

		$limitSql = User::isLimited() ? ' AND user_id = '.User::id() : '';

		$checkToRecordList = DAO::fillArray("
			SELECT * FROM budget_check_record WHERE $sqlIdList check_id IN (
				SELECT budget_check.id FROM budget_check WHERE user_id IN (
					SELECT users.id FROM users WHERE family_id = $1
				)
				$limitSql
				AND qr_sum > 0 
			)",
			null, array(User::familyId())
		);

		return $checkToRecordList;
	}

	/**
	 * Insert or update check list; [list] => array of arrays: 'server_id' or 'client_id' [int8] - server or client ID of the check# If client ID is present - try to insert new check, and return server2client correspondence in the result array# If server_id is present - try to update existing check, @param string $apiId
	 * @param string $login
	 * @param string $pass
	 * @param array $list
	 * @return array
	 * @see getCheckList description for other detail; Returns the array of server IDs, successfully changed; The client MUST save server IDs corresponded to client IDs, for subsequent 'update' and 'delete' calls;
	 *
	 */
	public function setCheckList($apiId, $login, $pass, $list)
	{
		/**
		 * Для функций записи должна быть активна подписка
		 */
		try {
			$this->getAccessStatus($apiId, $login, $pass);
		}
		catch (SoapFault $e) {
			if ($e->faultcode != '87') { // trial
				throw $e;
			}
		}

		/**
		 * Проверки
		 */
		$this->checkEmptyArray($list, '119', 'Check list');

		$retArray = array();

		try {
			DB::transactionBegin();
			$dublDebug    = array();
			$lastSyncData = DAO::fillArrayKeyId("SELECT client_id AS id, server_id, app_instance_id FROM budget_sync_last
    			WHERE user_id = $1
    			AND api_id = $2
    			AND object_type = $3",
				array(User::id(), Session::get(Constant::API_ID), BudgetHistory::TYPE_CHECK)
			);
			$appId        = self::getAppId();

			foreach ($list as $check) {
				/**
				 * Первым делом, если это добавление клиентской записи, которая
				 * уже была добавлена в прошлой синхре, то значит был сбой на клиенте
				 * и он шлёт дубль. Ничего не добавляем, а отдаём тот же server_id
				 * что и в прошлый раз.
				 */
				$isClientIdThisInstance = true;
				if (
					abs($appId) > 0 &&
					!empty($check['client_id']) &&
					isset($lastSyncData[$check['client_id']]) &&
					abs($lastSyncData[$check['client_id']]['app_instance_id']) > 0 &&
					$lastSyncData[$check['client_id']]['app_instance_id'] != $appId
				) {
					$isClientIdThisInstance = false;
				}

				if (!empty($check['client_id']) && isset($lastSyncData[$check['client_id']]) && $isClientIdThisInstance) {
					$retArray[]  = array(
						'server_id' => $lastSyncData[$check['client_id']]['server_id'],
						'client_id' => $check['client_id'],
						'status'    => 'inserted'
					);
					$dublDebug[] = array(
						'server_id' => $lastSyncData[$check['client_id']]['server_id'],
						'client_id' => $check['client_id'],
						'type'      => 'check',
						'check'     => $check,
					);
					continue;
				}

				/**
				 * Проверки
				 */
				if (empty($check['server_id']) && empty($check['client_id'])) {
					throw new SoapFault('120', "'server_id' or 'client_id' should be given");
				}

				$isChkRec = false;
				$origQr   = str_replace(array('&useSmsPlace=1', '&dontUseTagPlace=1'), '', $check['filename']);
				$origQr   = mb_substr($origQr, 0, 128);
				$checkId  = empty($check['server_id']) ? '' : $check['server_id'];

				/**
				 * Кейсы, когда такой QR уже есть в данной семье
				 */
				$existedCheckId = DAO::getCount("SELECT id FROM budget_check 
					WHERE user_id IN (SELECT id FROM users WHERE family_id = $1)
					AND ext = $2 LIMIT 1",
					true, array(User::familyId(), $origQr)
				);
				if ($existedCheckId) {
					/**
					 * Если у присылаемого чека уже есть server_id и он НЕ равен server_id
					 * существующего на сервере, то генерим исключение, не должно быть двух разных чеков с одним QR.
					 */
					if ($checkId && $existedCheckId != $checkId) {
						throw new SoapFault('127', "Invalid check '$checkId'. There is already exist check with id = $existedCheckId and QR '$origQr'");
					}
					/**
					 * Если у существующего чека нет привязанных операций, а у присланного не выставлен server_id,
					 * то существующий удаляем.
					 */
					$isChkRec = DAO::getCount(
						"SELECT record_id FROM budget_check_record WHERE check_id = $1 LIMIT 1",
						true, array($existedCheckId)
					);
					if (!$isChkRec && !$checkId) {
						DB::query("DELETE FROM budget_check WHERE id = $1", array($existedCheckId));
					}
				}

				/**
				 * Проверка даты в QR
				 */
				$qrData = array();
				parse_str($check['filename'], $qrData);
				$qrDateStm = strtotime(ArrayUtils::safeGet($qrData, 't'));
				if ($qrDateStm === false || $qrDateStm < 946674000) { // Меньше 2000-01-01
					$qrDateStm = User::getTime(time());
//					Debug::mail("Not valid qr date: ".print_r($qrData, true));
				}

				/**
				 * Всё нормально, создаём объект
				 */
				$checkObj = BudgetCheck::create($checkId)->
				set(BudgetCheck::STATE, $check['process'])->
				set(BudgetCheck::QR_DATE, date("Y-m-d H:i:s", $qrDateStm))->
				set(BudgetCheck::QR_SUM, $check['sum']);

				if (!$checkObj->get(BudgetCheck::ID)) {
					$checkObj->
					set(BudgetCheck::USER_ID, User::id())->
					set(BudgetCheck::CH_PLACE_ID, ArrayUtils::safeGet($qrData, 'useSmsPlace') == 1 ? -1 : null)->
					set(BudgetCheck::QR_USE_TAG, ArrayUtils::safeGet($qrData, 'dontUseTagPlace') == 1 ? false : true)->
					set(BudgetCheck::DATE, 'now()')->
					set(BudgetCheck::EXT, $origQr);
				}

				$isChecked = $checkObj->get(BudgetCheck::ID) ? $checkObj->get(BudgetCheck::IS_CHECKED) : false;

				$checkObj->
				set(BudgetCheck::IS_CHECKED, isset($check['is_checked']) ? $check['is_checked'] : $isChecked)->
				daoSave();

				/**
				 * При создании нового чека в веб-версии нужно насильно отобразить блок с чеками
				 * А то шлют чеки из мобильных приложений и не видят их.
				 */
				if (BudgetSettings::getValue(BudgetSettings::IS_HIDDEN_CHKBLK) == 'on') {
					BudgetSettings::save(BudgetSettings::IS_HIDDEN_CHKBLK, 'off');
				}

				$retArrayItem = array(
					'server_id' => $checkObj->get(BudgetCheck::ID),
					'status'    => $checkId ? 'updated' : 'inserted'
				);
				if (!$checkId) {
					$retArrayItem['client_id'] = $check['client_id'];
				}
				$retArray[] = $retArrayItem;
			}
			/**
			 * Перед тем, как успешно отправить список связок, надо грохнуть все
			 * клиентско-серверные связки данного типа, от прошлой синхры и добавить текущие.
			 */
			$delAppIdSql = $appId ? "AND app_instance_id = ".DB::esc($appId) : '';
			DB::query("DELETE FROM budget_sync_last
	    		WHERE user_id = $1
				AND api_id = $2
				AND object_type = $3
				$delAppIdSql",
				array(User::id(), Session::get(Constant::API_ID), BudgetHistory::TYPE_CHECK)
			);
			foreach ($retArray as $retValue) {
				if ($retValue['status'] != 'inserted') {
					continue;
				}
				BudgetSyncLast::create()->
				set(BudgetSyncLast::USER_ID, User::id())->
				set(BudgetSyncLast::API_ID, Session::get(Constant::API_ID))->
				set(BudgetSyncLast::OBJECT_TYPE, BudgetHistory::TYPE_CHECK)->
				set(BudgetSyncLast::CLIENT_ID, $retValue['client_id'])->
				set(BudgetSyncLast::SERVER_ID, $retValue['server_id'])->
				set(BudgetSyncLast::INSTANCE_ID, $appId)->
				set(BudgetSyncLast::DATE, 'now()')->
				daoSave();
			}
			DB::transactionCommit();
		}
		catch (Exception $e) {
			DB::transactionRollback();
			if (!in_array(User::id(), array('1000000225667'))) {
				Debug::mail($e);
			}
			throw new SoapFault('123', $e->getMessage());
		}
		return $retArray;
	}

	/**
	 * Insert or update check to record link list; [list] => array of arrays: 'server_id' or 'client_id' [int8] - server or client ID of the check# If client ID is present - try to insert new check, and return server2client correspondence in the result array# If server_id is present - try to update existing check, @param string $apiId
	 * @param string $login
	 * @param string $pass
	 * @param array $list
	 * @return array
	 * @see getCheckToRecordList description for other detail; Returns the array of server IDs, successfully changed; The client MUST save server IDs corresponded to client IDs, for subsequent 'update' and 'delete' calls;
	 *
	 */
	public function setCheckToRecordList($apiId, $login, $pass, $list)
	{
		/**
		 * Для функций записи должна быть активна подписка
		 */
		try {
			$this->getAccessStatus($apiId, $login, $pass);
		}
		catch (SoapFault $e) {
			if ($e->faultcode != '87') { // trial
				throw $e;
			}
		}

		/**
		 * Проверки
		 */
		$this->checkEmptyArray($list, '124', 'Check to record link list');

		$retArray = array();

		try {
			DB::transactionBegin();
			$dublDebug    = array();
			$lastSyncData = DAO::fillArrayKeyId("SELECT client_id AS id, server_id, app_instance_id FROM budget_sync_last
    			WHERE user_id = $1
    			AND api_id = $2
    			AND object_type = $3",
				array(User::id(), Session::get(Constant::API_ID), BudgetHistory::TYPE_CHECK_TO_REC)
			);
			$appId        = self::getAppId();

			foreach ($list as $checkToRec) {
				/**
				 * Первым делом, если это добавление клиентской записи, которая
				 * уже была добавлена в прошлой синхре, то значит был сбой на клиенте
				 * и он шлёт дубль. Ничего не добавляем, а отдаём тот же server_id
				 * что и в прошлый раз.
				 */
				$isClientIdThisInstance = true;
				if (
					abs($appId) > 0 &&
					!empty($checkToRec['client_id']) &&
					isset($lastSyncData[$checkToRec['client_id']]) &&
					abs($lastSyncData[$checkToRec['client_id']]['app_instance_id']) > 0 &&
					$lastSyncData[$checkToRec['client_id']]['app_instance_id'] != $appId
				) {
					$isClientIdThisInstance = false;
				}

				if (!empty($checkToRec['client_id']) && isset($lastSyncData[$checkToRec['client_id']]) && $isClientIdThisInstance) {
					$retArray[]  = array(
						'server_id'      => $lastSyncData[$checkToRec['client_id']]['server_id'],
						'client_id'      => $checkToRec['client_id'],
						'deleteSmsRecId' => "0",
						'status'         => 'inserted'
					);
					$dublDebug[] = array(
						'server_id' => $lastSyncData[$checkToRec['client_id']]['server_id'],
						'client_id' => $checkToRec['client_id'],
						'type'      => 'check',
						'data'      => $checkToRec,
					);
					continue;
				}

				/**
				 * Проверки
				 * Связь может только создаваться/удаляться и не может изменяться.
				 */
				if (empty($checkToRec['client_id'])) {
					throw new SoapFault('126', "'client_id' of the check to record link should be given");
				}
				/**
				 * Если чека или операции уже (или ещё?) нет на сервере, то связку создать нельзя
				 * Отдаём клиенту фихтивное значение server_id, связка должна грохнуться там каскадно в БД
				 */
				$deleteSmsRecId = "0";
				$dbCheckId      = DAO::getCount("SELECT id FROM budget_check WHERE id = $1", true, array($checkToRec['check_id']));
				$dbRecordId     = DAO::getCount("SELECT id FROM budget_record WHERE id = $1", true, array($checkToRec['record_id']));
				if (!$dbCheckId || !$dbRecordId) {
					$nextIdVal = "1";
				}
				else {
					$nextIdVal = DAO::getCount("SELECT id FROM budget_check_record WHERE check_id = $1 AND record_id = $2",
						true, array($checkToRec['check_id'], $checkToRec['record_id'])
					);

					$nextIdVal = BudgetCheckRecord::create($nextIdVal)->
					set(BudgetCheckRecord::CHECK_ID, $checkToRec['check_id'])->
					set(BudgetCheckRecord::RECORD_ID, $checkToRec['record_id'])->
					daoSave()->
					get(Business::ID);

					/**
					 * Завершаем логику создания чека
					 */
					$checkObj = BudgetCheck::create($dbCheckId);
					$isChkRec = DAO::getCount(
						"SELECT record_id FROM budget_check_record WHERE check_id IN (
							SELECT budget_check.id FROM budget_check 
							WHERE user_id IN (SELECT users.id FROM users WHERE family_id = $1)
							AND budget_check.id <> $3 AND ext = $2					
						) LIMIT 1",
						true, array(User::familyId(), $checkObj->get(BudgetCheck::EXT), $dbCheckId)
					);
					if ($isChkRec) {
						/**
						 * Если есть другой чек с таким же QR и у него есть привязанные операции, то обрабатываемый
						 * чек удаляем, QR операцию у него удаляем и связка тоже удалится базой.
						 */
						DB::query("DELETE FROM budget_record WHERE id IN (
							SELECT record_id FROM budget_check_record WHERE check_id = $1)",
							array($checkObj->get(BudgetCheck::ID))
						);
						$checkObj->daoDrop();
					}
					else {
						/**
						 * Считаем, что в этой точке оказываемся только при создании нового чека. Если при этом есть
						 * расходная операция созданная из Пуш.СМС на такую же сумму и дату (плюс-минус MINUTE_DIFF мин),
						 * то её (другую) удаляем, её поля переносим (какие надо) в QR-овскую операцию, и меняем record_id
						 * в таблицах "push_history", "budget_record_tags", "budget_record_planned"
						 */
						$curRecord   = v2_BudgetRecord::create($dbRecordId);
						$curRecDate  = $curRecord->get(v2_BudgetRecord::DATE);
						$nearRecList = RecordUtils::getNearRecList($curRecord->get(v2_BudgetRecord::SUM), $curRecDate, $dbRecordId);
						$nearRecord  = null;
						foreach ($nearRecList as $nearRec) {
							$isAnyCheck = DAO::getCount(
								"SELECT check_id FROM budget_check_record WHERE record_id = $1 LIMIT 1",
								true, array($nearRec->get(Business::ID))
							);
							$isFromPush = DAO::getCount(
								"SELECT record_id FROM push_history WHERE record_id = $1 LIMIT 1",
								true, array($nearRec->get(Business::ID))
							);
							/**
							 * Склеивать можно только пушовую операцию, которая ещё не привязана ни к какому чеку
							 */
							if (!$isFromPush || $isAnyCheck) {
								continue;
							}

							/**
							 * Исключаем ситуацию, когда пользователем было создано две операции на одну сумму
							 * и к одной из них базовая операция по чеку уже была приклеена на стороне Андроид
							 * клиента. Второй раз склейку не делаем, вторую операцию не повреждаем.
							 *
							 * Критерем считаем одновременное изменение базовой расходной операции по чеку и создание
							 * самого чека, т.к. это происходит в одной синхронизации. С разницей в несколько сек.
							 */
							$psbCheckId = DAO::getCount("
								SELECT object_id FROM  budget_history bh where family_id = $1 and object_type_id = '10' 
								and abs(extract(epoch from (date - current_timestamp::timestamp(0)))) < 10 and action_id ='1'
								and object_id = $2",
								true, array(User::familyId(), $dbCheckId)
							);
							$psbRecId   = DAO::getCount("
								SELECT object_id FROM budget_history bh 
								where family_id = $1 and object_id = $2 and action_id = '2'
								and abs(extract(epoch from (date - current_timestamp::timestamp(0)))) < 10 LIMIT 1",
								true, array(User::familyId(), $dbRecordId)
							);
							if ($psbCheckId && $psbRecId) {
								/**
								 * Логи для отладки.
								 */
								Internal::toLog("', psbCheckId: '$psbCheckId', psbRecId: '$psbRecId'", '/var/log/drebedengi/check.skleyka.log');
								break;
							}
							$nearRecord = $nearRec;
							break;
						}

						if ($nearRecord) {
							$nearRecordId = $nearRecord->get(Business::ID);
							$curRecord->
							set(v2_BudgetRecord::DATE, $nearRecord->get(v2_BudgetRecord::DATE))->
							set(v2_BudgetRecord::OBJECT_ID, $nearRecord->get(v2_BudgetRecord::OBJECT_ID));

							if ($checkObj->get(BudgetCheck::CH_PLACE_ID) == -1) {
								$curRecord->set(v2_BudgetRecord::ACCOUNT_ID, $nearRecord->get(v2_BudgetRecord::ACCOUNT_ID));
								$checkObj->set(BudgetCheck::CH_PLACE_ID, $nearRecord->get(v2_BudgetRecord::ACCOUNT_ID))->daoSave();
							}

							DB::query("UPDATE budget_record_planned SET entered_id = $1 WHERE entered_id = $2", array($dbRecordId, $nearRecordId));
							/**
							 * Теги нельзя переклеить просто update-ом, потому что у QR операции уже могут быть
							 * такие же теги как и в СМС. Сработает ограничение БД.
							 */
							$tagIds = DAO::fillArray(
								"DELETE FROM budget_record_tags WHERE budget_record_id = $1 RETURNING budget_tag_id",
								null, array($nearRecordId)
							);
							foreach ($tagIds as $tagId) {
								$tgParams = array($dbRecordId, $tagId['budget_tag_id']);
								$isExist  = DAO::getCount("SELECT budget_tag_id FROM budget_record_tags WHERE budget_record_id = $1 AND budget_tag_id = $2", true, $tgParams);
								if (!$isExist) {
									DB::query("INSERT INTO budget_record_tags(budget_record_id, budget_tag_id) VALUES ($1, $2)", $tgParams);
								}
							}

							/**
							 * record_id в push_history "пропадёт", если СМС и QR операция приходят вместе в одной синхронизации, т.к.
							 * в этой точке связи в push_history нет и операцию удаляем. Пока не страшно, record_id используется только
							 * для статистики в алгоритме автоопределения счетов.
							 */
							DB::query("UPDATE push_history SET record_id = $1 WHERE record_id = $2", array($dbRecordId, $nearRecordId));

							$comment = $nearRecord->get(v2_BudgetRecord::COMMENT)." ".$curRecord->get(v2_BudgetRecord::COMMENT);
							$curRecord->set(v2_BudgetRecord::COMMENT, $comment)->daoSave();

							$nearRecord->daoDrop();
							/**
							 * Отправляем клиенту команду на удаление СМС операции, чтобы пользователь не увидел дубля даже
							 * на короткое время между синхронизациями.
							 */
							$deleteSmsRecId = $nearRecordId;
						}
					}
				}

				$retArray[] = array(
					'server_id'      => $nextIdVal,
					'client_id'      => $checkToRec['client_id'],
					'deleteSmsRecId' => $deleteSmsRecId,
					'status'         => 'inserted'
				);
			}
			/**
			 * Перед тем, как успешно отправить список связок, надо грохнуть все
			 * клиентско-серверные связки данного типа, от прошлой синхры и добавить текущие.
			 */
			$delAppIdSql = $appId ? "AND app_instance_id = ".DB::esc($appId) : '';
			DB::query("DELETE FROM budget_sync_last
	    		WHERE user_id = $1
				AND api_id = $2
				AND object_type = $3
				$delAppIdSql",
				array(User::id(), Session::get(Constant::API_ID), BudgetHistory::TYPE_CHECK_TO_REC)
			);
			foreach ($retArray as $retValue) {
				if ($retValue['status'] != 'inserted') {
					continue;
				}
				BudgetSyncLast::create()->
				set(BudgetSyncLast::USER_ID, User::id())->
				set(BudgetSyncLast::API_ID, Session::get(Constant::API_ID))->
				set(BudgetSyncLast::OBJECT_TYPE, BudgetHistory::TYPE_CHECK_TO_REC)->
				set(BudgetSyncLast::CLIENT_ID, $retValue['client_id'])->
				set(BudgetSyncLast::SERVER_ID, $retValue['server_id'])->
				set(BudgetSyncLast::INSTANCE_ID, $appId)->
				set(BudgetSyncLast::DATE, 'now()')->
				daoSave();
			}
			DB::transactionCommit();
		}
		catch (Exception $e) {
			DB::transactionRollback();
			$periodClosed = Date::periodClosed($e, false);
			if ($periodClosed) {
				throw new SoapFault('188', $periodClosed);
			}
			$conflictMsg = 'Для восстановления синхронизации, ';
			$conflictMsg .= ($apiId == 'vandr_api' ? "необходимо перезайти через меню\n'Ещё->Сменить пользователя'." : 'приложение необходимо переустановить.');
			if (mb_strpos($e->getMessage(), 'create_only_access') !== false) {
				throw new SoapFault('188', "Настройки прав доступа не позволяют изменять операции\n".$conflictMsg);
			}
			$accessErrorMsg = array('forbidden_type_', 'el_forbidden', 'not_owner_obj', 'readonly_access', 'not_owner_rec');
			foreach ($accessErrorMsg as $errorMsg) {
				if (strpos($e->getMessage(), $errorMsg) !== false) {
					throw new SoapFault('188', "Для обработки чека не хватает прав доступа.\n".$conflictMsg);
				}
			}
			Debug::mail($e);
			throw new SoapFault('125', $e->getMessage());
		}
		return $retArray;
	}

	/**
	 * Get user right access for multi user mode; Returns 0 - full access, 1 - limited access; For limited access it is neccessary to use getBalance method to display user balance in the application; In this case local transactions must simple increment balance, but not compute that as sum of all transactions;
	 *
	 * @param string $apiId
	 * @param string $login
	 * @param string $pass
	 * @return string
	 */
	public function getRightAccess($apiId, $login, $pass)
	{
		$this->login($apiId, $login, $pass);

		$access = DAO::getCount("SELECT acc_level FROM users WHERE id = $1", true, array(User::id()));
		$access = ($access > 0 ? '1' : '0');
		return "$access";
	}

	/**
	 * Internal method; Don't use it;
	 *
	 * @param string $url
	 * @return array
	 */
	public function getServerSubs($url)
	{
		return Internal::getServerSubs($url);
	}

	public function getSubscriptionStatus($apiId, $login, $pass)
	{
		$this->login($apiId, $login, $pass);
		return Internal::getSubscriptionStatus($apiId, $login, $pass);
	}

	public function setPaymentTransaction($apiId, $login, $pass, $transactionReceipt, $amount)
	{
		$this->login($apiId, $login, $pass);

		if (Billing::TARIF_VANDR == $apiId) {
			return Internal::setPaymentTransactionAndroid($apiId, $login, $pass, $transactionReceipt);
		}

		Internal::setPaymentTransaction($apiId, $login, $pass, $transactionReceipt, $amount);

		return $this->getExpireDate($apiId, $login, $pass);
	}

	/**
	 * Try to parse text data as records; [def..] default field values if no rules detected; [list] => array (indexes must be 0,1,2...N) of strings to parse (UTF8); Returns the array of array - data for records;
	 *
	 * @param string $apiId
	 * @param string $login
	 * @param string $pass
	 * @param array $list
	 * @return array
	 */
	public function parsePushData($apiId, $login, $pass, $list)
	{
		/**
		 * Функция платная
		 */
		try {
			$this->getAccessStatus($apiId, $login, $pass);
		}
		catch (SoapFault $e) {
			if ($e->faultcode != '87') { // trial
				throw $e;
			}
		}

		/**
		 * @todo Добавить поддержку серверной версии
		 */
		if (defined('IS_SERVER')) {
			return array();
		}

		/**
		 * Проверки
		 */
		$this->checkEmptyArray($list, '116', 'String list');

		/**
		 * Уведомления приходят "в сыром" виде как есть. Клиент их никак не обрабатывает, т.к. на сервре это делать удобнее.
		 * Поэтому сначала нужно на основе присланных данных составить список строк (СМС или пушей), в которых предположительно
		 * содержатся расходные операции, который затем скармливать как и раньше в ImportRule::parse
		 */
		$blackListTitle = array(
			'приложение "сообщения" выполняется' => '',
			'messaging is running'               => '',
			'приложение "сообщения" запущено'    => '',
			'инициализация…'                     => '',
			'ваши средства в безопасности'       => '',
			'проверка устройства'                => '',
		);

		$smsPkg = array(
			'com.android.mms'                   => '',
			'com.android.mms.service'           => '',
			'com.google.android.apps.messaging' => '',
			'com.samsung.android.messaging'     => '',
			'com.android.messaging'             => '',
			'rpkandrodev.yaata'                 => '',
			'com.mysms.android.sms'             => '',
			'com.textra'                        => '',
			'com.p1.chompsms'                   => '',
			'com.truecaller'                    => '',
			'com.asus.message'                  => '',
			'com.htc.sense.mms'                 => '',
			'com.lenovo.ideafriend'             => '',
			'com.android.contacts'              => '',
			'com.jb.gosms'                      => '',
			'com.calea.echo'                    => '',
			'com.verizon.messaging.vzmsgs'      => '',
			'com.handcent.app.nextsms'          => '',
			'com.sonyericsson.conversations'    => '',
			'com.thinkyeah.message'             => '',
			'com.oneplus.mms'                   => '',
			'com.transsion.smartmessage'        => '',
		);

		$bankPkg = array(
			'com.webmoney.my'                        => '',
			'ru.vtb24.mobilebanking.android'         => '',
			'ru.rocketbank.r2d2'                     => '',
			'com.idamob.tinkoff.android'             => '',
			'ru.sberbankmobile'                      => '',
			'ru.sberbank.sa.mobile'                  => '',
			'com.ftband.mono'                        => '',
			'ua.privatbank.ap24'                     => '',
			'com.google.android.gms'                 => '',
			'com.samsung.android.spay'               => '',
			'ru.alfabank.mobile.android'             => '',
			'ru.alfabank.oavdo.amc'                  => '',
			'ru.sovcomcard.halva.v1'                 => '',
			'ru.gazprombank.android.mobilebank.app'  => '',
			'ru.raiffeisennews'                      => '',
			'ru.raiffeisen.rmobile'                  => '',
			'rs.raiffeisen.mobile'                   => '',
			'wit.android.bcpbankingapp.millenniumpl' => '',
			'kz.kaspi.mobile'                        => '',
			'ru.yandex.money'                        => '',
			'ru.simpls.brs2.mobbank'                 => '',
			'ru.nspk.mirpay'                         => '',
			'ru.nspk.sbpay'                          => '',
			'com.tochka.bank'                        => '',
			'pl.aliorbank.aib'                       => '',
			'ru.zhuck.webapp'                        => '',
			'ru.sberbank_sbbol'                      => '',
			'ru.sberbank.iconwhite'                  => '',
			'ru.bankuralsib.mb.android'              => '',
			'ru.rshb.dbo'                            => '',
			'ru.advisa.smartsms'                     => '',
			'ru.broker.my'                           => '',
			'com.openbank'                           => '',
			'ru.yoo.money'                           => '',
			'ru.vtb24.biz.client.android'            => '',
			'kz.homecredit.ibank'                    => '',
			'ru.kykyryza'                            => '',
			'ru.sberbankmobile_alpha'                => '',
			'md.maib.maibank'                        => '',
			'ru.sberbank_sbbol'                      => '',
			'modulbank.ru.app'                       => '',
			'ua.alfabank.mobile.android'             => '',
			'ru.mw'                                  => '',
			'com.revolut.revolut'                    => '',
			'logo.com.mbanking'                      => '',
			'ru.letobank.prometheus'                 => '',
			'com.bbva.bbvacontigo'                   => '',
			'com.bank.vr'                            => '',
			'com.denizbank.mobildeniz'               => '',
			'ru.unicredit.android'                   => '',
			'am.imwallet.android'                    => '',
			'com.yandex.bank'                        => '',
			'ru.ozon.app.android'                    => '',
			'by.bgpb.mbanking'                       => '',
			'ru.ozon.fintech.finance'                => '',
			'by.alfabank.insync3'                    => '',
			'by.alfabank.insync'                     => '',
			'ua.alfabank.mobile.android'             => '',
			'ru.ozon.fintech.sme'                    => '',
			'com.google.android.apps.walletnfcrel'   => '',
		);

		$newList          = array();
		$groupedSmsList   = array();
		$smsListHistIdx   = array();
		$smsPushHistIdx   = array();
		$isPushLink       = false;
		$retArray         = array();
		$retArrayDebug    = array();
		$isDebug          = false;
		$this->pushErrors = array();
		$floodPkg         = array();
		$pkgCount         = array();
		$workTimeArray    = null;
		$cancelPattern    = '/(отмена|otmena|otmenen|otkaz|отказ|отклонен|отклонён|Недостатньо|Недостаточно|Возврат)/iu';

		try {
			DB::transactionBegin();

			$smsDefaults = BudgetSettings::getValue('android_sms_to_push');
			$smsDefaults = json_decode($smsDefaults, true);
			$userSrcPref = Session::get(Constant::IS_MULTIPLE_USERS) ? ' '.Session::get(User::NICK).' ' : ' ';

			/**
			 * Предварительный проход для определения спаммеров:
			 * - Больше 20 запросов в минуту от одного приложения.
			 * Это скорее всего не финансовые уведомления.
			 */
			PushData::getBlackList();
			$whitePkg = array_merge($smsPkg, $bankPkg);
			$pushId   = '';
			foreach ($list as $jKey => $jsonLine) {
				$nData  = json_decode($jsonLine, true);
				$nPkg   = mb_substr(mb_strtolower(trim(ArrayUtils::safeGet($nData, 'pkg'))), 0, 128);
				$msec   = ArrayUtils::safeGet($nData, 'date');
				$pushId = ArrayUtils::safeGet($nData, 'pushId');
				if (isset($whitePkg[$nPkg]) || empty($msec) || isset(PushData::$blackListPkg[$nPkg])) {
					continue;
				}
				isset($pkgCount[$nPkg]) ? '' : $pkgCount[$nPkg] = array();
				$pkgCount[$nPkg][] = $msec;
				$msecPrev20        = ArrayUtils::safeGet($pkgCount[$nPkg], (count($pkgCount[$nPkg]) - 20));
				if ($msecPrev20 && (($msec - $msecPrev20) / 1000) < 60) {
					$floodPkg[$nPkg] = '';
					PushData::addToBlackList($nPkg);
				}
			}

			/**
			 * Связку пушей и операций пересортируем в обратном порядке, т.к. Андроид клиент её присылает "задом наперёд"
			 */
			$pushId ? $list = array_reverse($list) : '';

			foreach ($list as $jKey => $jsonLine) {
				if (empty($jsonLine)) {
					continue;
				}
				$noteSmsList       = array();
				self::$lastMsgList = array();
				$notificationData  = json_decode($jsonLine, true);
				$nDate             = ArrayUtils::safeGet($notificationData, 'date');
				$nHistId           = ArrayUtils::safeGet($notificationData, 'histId');
				if ($nDate) {
					$msec                     = mb_substr($nDate, -3);
					$notificationData['date'] = date("Y-m-d H:i:s", floor($nDate / 1000)).".$msec";
				}
				if (empty($nHistId)) {
					/**
					 * ВОзможно отправка начальных данных при переходе с СМС на пуш.
					 * Соответствия СМС номеров и полей по умолчанию
					 */
					if (isset($notificationData[0]['smsNum'])) {
						BudgetSettings::save('android_sms_to_push', json_encode($notificationData, JSON_UNESCAPED_UNICODE));
						/**
						 * После использования этого значения по назначению - удалить настройку
						 */
						continue;
					}
					/**
					 * Возможно связка серверной транзакции и пушки
					 */
					$pushId     = ArrayUtils::safeGet($notificationData, 'pushId');
					$ruleId     = ArrayUtils::safeGet($notificationData, 'ruleId');
					$recordId   = ArrayUtils::safeGet($notificationData, 'recordId');
					$recordObj  = v2_BudgetRecord::create($pushId ? $recordId : null);
					$recordDate = $recordObj->get(v2_BudgetRecord::DATE);
					$pushValues = $pushId ? DAO::fillArray("SELECT message,date,source FROM push_history WHERE id = $1", null, array($pushId)) : '';
					if ($pushId && $recordDate && $pushValues) {
						$isPushLink = true;
						$secondId   = DAO::getCount("
    						SELECT (CASE WHEN t_all.record1 = $1 THEN t_all.record2 ELSE t_all.record1 END) AS second_id
    						FROM budget_record_links AS t_all WHERE (record1 = $1 OR record2 = $1) AND type = $2",
							true, array($recordId, BudgetRecordLinks::TYPE_MOVE)
						);

						/**
						 * 14 сен 2022 Переносим сюда основную логику отсева дублей от СМС/чеков
						 * Так же и логку отлова "отмены операции". В общем случае считаем, что отмена приходит
						 * после создания самой операции.
						 */
						$message     = $pushValues[0]['message'];
						$pushDate    = $pushValues[0]['date'];
						$pushSource  = $pushValues[0]['source'];
						$isCancelMsg = preg_match($cancelPattern, $message, $wordMatch);
						$recordSum   = abs($recordObj->get(v2_BudgetRecord::SUM));
						$msgType     = $recordObj->get(v2_BudgetRecord::SUM) > 0 ? v2_Budget::TYPE_INCOME : v2_Budget::TYPE_WASTE;
						$recIds      = array($recordId);
						if ($secondId) {
							$msgType  = v2_Budget::TYPE_MOVE;
							$recIds[] = $secondId;
						}
						if ($isCancelMsg) {
							$recList     = DAO::fillArray("
								SELECT record_id, rule_id, message, type FROM budget_record AS br
								JOIN push_history ON (record_id = br.id)
								JOIN budget_object ON (budget_object.id = br.budget_object_id)
								WHERE br.budget_family_id = $1 AND abs(difference) = $2 AND br.id NOT IN (".DB::in($recIds).")
								AND operation_date <= '$recordDate'
								AND abs(extract(epoch from (operation_date - '$recordDate'::timestamp(0)))) < 120*60
								ORDER BY operation_date DESC LIMIT 1",
								null, array(User::familyId(), $recordSum)
							);
							$prevMessage = 'prev record not found';

							/**
							 * Если обе операции определены как "доход", то предыдущую не удаляем.
							 */
							$isBothIncome = ($msgType == v2_Budget::TYPE_INCOME && !empty($recList) && $recList[0]['type'] == v2_Budget::TYPE_INCOME);
							$isBothIncome ? $prevMessage = "isBothIncome: '{$recList[0]['message']}'" : '';

							/**
							 * Если предыдущую операцию не нашли, то возможно одновременное поступление уведомлений об отмене и покупке,
							 * или даже "отмена" приходит раньше на нес-ко секунд. Проверим следующую по списку операцию, не там ли "покупка"?
							 * Берём максимум 10 сек разницы по времени.
							 */
							$nextNotifPref = '';
							$nextNotifData = json_decode(ArrayUtils::safeGet($list, $jKey + 1), true);
							if (!$recList && $nextNotifData) {
								$recList = DAO::fillArray("
									SELECT br.id AS record_id, rule_id, message, type FROM budget_record AS br
									JOIN push_history ON (push_history.id = $1)
									JOIN budget_object ON (budget_object.id = br.budget_object_id)
									WHERE br.id = $2 AND abs(difference) = $3
									AND abs(extract(epoch from (operation_date - '$recordDate'::timestamp(0)))) < 10",
									null, array(
										ArrayUtils::safeGet($nextNotifData, 'pushId'),
										ArrayUtils::safeGet($nextNotifData, 'recordId'),
										$recordSum
									)
								);
								$recList ? $nextNotifPref = '[Следующая] ' : '';
							}

							$deleteAllowed = true;
							if (preg_match('/(отмена|otmena|otmenen|Возврат)/iu', $message, $wordMatch) && !$recList) {
								/**
								 * Вероятно частичный возврат, не удаляем
								 */
								if ($msgType == v2_Budget::TYPE_INCOME) {
									$deleteAllowed = false;
								}
								/**
								 * Перемещение с использованием правила пользователя, вероятно что-то нужное
								 */
								if ($secondId && $ruleId) {
									$deleteAllowed = false;
								}
							}

							if (!empty($recList) && !$isBothIncome) {
								$rItem = $recList[0];
								if ($rItem['type'] == v2_Budget::TYPE_MOVE) {
									$rItemSecondId = DAO::getCount("
										SELECT (CASE WHEN t_all.record1 = $1 THEN t_all.record2 ELSE t_all.record1 END) AS second_id
										FROM budget_record_links AS t_all WHERE (record1 = $1 OR record2 = $1) AND type = $2",
										true, array($rItem['record_id'], BudgetRecordLinks::TYPE_MOVE)
									);
									DB::query("DELETE FROM budget_record_links WHERE record1 = $1 OR record2 = $1", array($rItem['record_id']));
									DB::query("DELETE FROM budget_record WHERE id = $1", array($rItemSecondId));
								}
								DB::query("DELETE FROM budget_record WHERE id = $1", array($rItem['record_id']));
								$prevMessage = $nextNotifPref."rule_id = '{$rItem['rule_id']}', type = '{$rItem['type']}', prevMsg = '".$rItem['message']."', recId = '{$rItem['record_id']}'";
							}

							if ($secondId && $deleteAllowed) {
								DB::query("DELETE FROM budget_record_links WHERE record1 = $1 OR record2 = $1", array($recordId));
								DB::query("DELETE FROM budget_record WHERE id = $1", array($secondId));
							}

							$message     = str_replace(array("\n", "\r"), ' ', $message);
							$prevMessage = str_replace(array("\n", "\r"), ' ', $prevMessage);
							$feedBody    = ($deleteAllowed ? '' : '[Не удаляем] ')."rule_id = '$ruleId', type = '$msgType', msg = '$message', recId '$recordId'\n$prevMessage\n";

							Internal::toLog("\n$feedBody"."########################################\n", "/var/log/drebedengi/web.sms.new.cancel.log");

							/**
							 * Если
							 * - в строке есть слова типа "Отмена|Возврат"
							 * - не найдена пред. операция
							 * - тип операции "доход" или тип операции "перемещеине" и сработало пользовательское правило
							 * то может быть это частичный возврат или что-то нужное пользователю.
							 * Удалять такую отмену не надо, пусть будет доход
							 */
							if ($deleteAllowed) {
								$recordObj->daoDrop();
								continue;
							}
						}
						/**
						 * Если это не отмена, то запускаем логику проверки дублей
						 */
						else {
							$recList = DAO::fillArray("
								SELECT record_id, rule_id, message, source FROM budget_record AS br
								JOIN push_history ON (record_id = br.id)
								JOIN budget_object ON (budget_object.id = br.budget_object_id)
								WHERE br.budget_family_id = $1 AND abs(difference) = $2 AND br.id <> $3
								AND budget_object.type = $4
								AND abs(push_history.date - $5) < 6*3600*1000
								AND br.id NOT IN (
									SELECT record_id FROM budget_check_record WHERE record_id = br.id
								)
								ORDER BY abs(push_history.date - $5) LIMIT 1",
								null, array(User::familyId(), $recordSum, $recordId, $msgType, $pushDate)
							);
							if ($recList) {
								$dblRecId = self::getDoubleId(
									$recordId, $message, $pushSource, $ruleId,
									$recList[0]['record_id'], $recList[0]['message'], $recList[0]['source'], $recList[0]['rule_id']
								);
								if ($dblRecId) {
									if ($msgType == v2_Budget::TYPE_MOVE) {
										$halfs = DAO::fillArray("
											SELECT * FROM budget_record_links WHERE (record1 = $1 OR record2 = $1) AND type = $2",
											null, array($dblRecId, BudgetRecordLinks::TYPE_MOVE)
										);
										DB::query("DELETE FROM budget_record_links WHERE record1 = $1 OR record2 = $1", array($dblRecId));
										DB::query("DELETE FROM budget_record WHERE id = $1 OR id = $2",
											array($halfs[0]['record1'], $halfs[0]['record2'])
										);
									}
									DB::query("DELETE FROM budget_record WHERE id = $1", array($dblRecId));
								}
								/**
								 * Если признали дублем и удалили операцию для текущего уведомления - дальше ничего делать не надо
								 */
								if ($dblRecId == $recordId) {
									continue;
								}
							}
						}

						/**
						 * Если это перемещение, то нужно вытянуть вторую полопвинку, т.к. клиент связывает другую
						 */
						$secondId ? $recordId = $secondId : '';

						/**
						 * Если это расходная операция, запустим логику QR операции.
						 */
						$isDeleteSmsRec = $this->processQr($recordId, $pushId, $ruleId, $secondId);

						/**
						 * В Android версиях с QR ( >= 2.4.15) в ответ отдаётся признак, нужно ли удалять
						 * операцию, если она дублирует QR-овскую. Просто чтобы пользователь не увидел дубля
						 * в интерфейсе, в промежуток времени до следующей синхронизацией
						 */
						if (self::isMoreVersion(array(2, 4, 15)) && $isDeleteSmsRec) {
							$retArray[] = array(
								'type'    => (string)ImportRule::ACTION_NONE,
								'hist_id' => "$recordId",
							);
						}
						continue;
					}
					$this->addError($jKey, empty($notificationData) ? $jsonLine : $notificationData, 'no histId in the input data');
					continue;
				}

				/**
				 * Шаг 1: Получиить список сообщений, (их может быть несколько) сгруппированных по
				 * источникам, содержащихся в данном уведомлении.
				 *
				 * @todo Пока считаем, что от в одном уведомлении, от одного источника содержится одно
				 *      сообщение. Но алгоритм можно легко расширить и для обработки нескольких сообщений
				 *      от одного источника. Если понадобится - сделаем.
				 */
				$nPkg   = mb_strtolower(trim(ArrayUtils::safeGet($notificationData, 'pkg')));
				$nText  = trim(ArrayUtils::safeGet($notificationData, 'text'));
				$nLines = ArrayUtils::safeGet($notificationData, 'textLines');
				$nTitle = trim(ArrayUtils::safeGet($notificationData, 'title'));

				/**
				 * Вырезаем лишнее из заголовка.
				 */
				if ('com.asus.message' == $nPkg) { // У этого номер ДО двоеточия
					$nTitle = preg_replace('/(:.*)$/', '', $nTitle);
				}
				$nSmsTitle = PushData::stripSmsTitle($nTitle);

				/**
				 * Игноры флудеров и вручную добавленных на сервере программ
				 */
				if (isset($floodPkg[$nPkg]) || isset(PushData::$blackListPkg[$nPkg])) {
					if (self::isMoreVersion(array(2, 5, 12))) {
						$retArray[] = array(
							'type'    => (string)ImportRule::ACTION_FLOOD,
							'hist_id' => "$nHistId",
							'pkg'     => "$nPkg",
						);
					}
					else {
						ArrayUtils::push($groupedSmsList['DUP'], 'DUP: [a]');
						ArrayUtils::push($smsListHistIdx['DUP'], $nHistId);
						ArrayUtils::push($smsPushHistIdx['DUP'], '1');
					}
					continue;
				}

				/**
				 * Игноры по чёрным спискам
				 */
				if (isset($blackListTitle[$nSmsTitle]) || (isset($smsPkg[$nPkg]) && isset(PushData::$blackListSms[$nSmsTitle]))) {
					ArrayUtils::push($groupedSmsList['DUP'], 'DUP: [b]');
					ArrayUtils::push($smsListHistIdx['DUP'], $nHistId);
					ArrayUtils::push($smsPushHistIdx['DUP'], '1');
					continue;
				}

				/**
				 * Мониторинг для отладки
				 */
				$newList[] = $notificationData;

				if ($nLines && !is_array($nLines)) {
					$this->addError($jKey, $notificationData, 'textLines must be an array');
					ArrayUtils::push($groupedSmsList['DUP'], 'DUP: [c]');
					ArrayUtils::push($smsListHistIdx['DUP'], $nHistId);
					ArrayUtils::push($smsPushHistIdx['DUP'], '1');
					continue;
				}
				$smsSource = "$nPkg:$nSmsTitle";
				/**
				 * Основной "старый" СМС месенджер. СМС могут идти друг за другом и складываться в одно уведомление,
				 * разделяемые переводом строк. Могут быть и сообщения с переводом строки.
				 */
				if (in_array($nPkg, array('com.android.mms', 'com.p1.chompsms', 'com.android.mms.service'))) {
					$newLineSym = ('com.p1.chompsms' == $nPkg ? $newLineSym = " | " : "\n");
					$isN        = mb_strpos($nText, $newLineSym);
					if ($nLines) {
						foreach ($nLines as $lineId => $textLine) {
							$textLine    = str_replace('…', ' ', trim($textLine));
							$spacePos    = mb_strpos($textLine, '  ');
							$spaceOffset = 2;
							if (!$spacePos && empty($nText)) {
								$spacePos    = mb_strpos($textLine, ' ');
								$spaceOffset = 1;
							}
							if (!$spacePos) {
								/**
								 * Кейс типа {"pkg":"com.android.mms",
								 * "text":"Новых сообщений: 2","textLines":["Новых сообщений: 2"],
								 * "title":"Елена Владимировна Аль-Зейди"}
								 */
								if (count($nLines) == 1 &&
									mb_stripos($textLine, 'Новых сообщений') !== false ||
									mb_stripos($textLine, 'new messages') !== false
								) {
									$noteSmsList["$nPkg:$nTitle"] = $textLine;
									continue;
								}

								$nums = str_replace('、', ',', $nText);
								if (mb_strpos($nums, ',') === false) {
									$this->addError($jKey, $notificationData, "not detected num for '$textLine', nums = $nums");
									continue;
								}
								$nums   = explode(',', $nums);
								$num    = trim(ArrayUtils::safeGet($nums, $lineId));
								$numInv = trim(ArrayUtils::safeGet($nums, count($nLines) - $lineId - 1));
								if ($num && mb_strpos($textLine, $num) === 0) {
									$spacePos = mb_strlen($num);
								}
								elseif ($numInv && mb_strpos($textLine, $numInv) === 0) {
									$spacePos = mb_strlen($numInv);
								}
								else {
									$this->addError($jKey, $notificationData, "not detected num[1] for '$textLine', num = $num");
									continue;
								}
								$spaceOffset = 0;
							}
							$num = PushData::stripSmsTitle(mb_substr($textLine, 0, $spacePos));
							$msg = trim(mb_substr($textLine, $spacePos + $spaceOffset));
							if (isset(PushData::$blackListSms[$num]) || empty($msg)) {
								continue;
							}
							$noteSmsList["$nPkg:$num"] = $msg;
						}
					}
					elseif ($isN) {
						$lastMessage = $this->getLastPush($smsSource);
						$lastMsgPos  = mb_strrpos($nText, $lastMessage.$newLineSym);
						if ($lastMsgPos === false || empty($lastMessage)) {
							/**
							 * Считаем, что одно многострочное сообщение
							 */
							$noteSmsList[$smsSource] = $nText;
						}
						else {
							$curMessage = trim(mb_substr($nText, $lastMsgPos + mb_strlen($lastMessage.$newLineSym)));
							$curMessage ? $noteSmsList[$smsSource] = $curMessage : '';
						}
					}
					elseif ($nSmsTitle && $nText) { // Иногда бывают "пустые" уведомления от СМС.
						$noteSmsList[$smsSource] = $nText;
					}
				}
				/**
				 * Кейс когда СМС друг за другом идут в разных уведомлениях, причём не прочитанные повторяются по
				 * многу раз, в разных вариантах.
				 */
				elseif (isset($smsPkg[$nPkg])) {
					if ($nLines && in_array($nPkg, array('com.sonyericsson.conversations', 'com.calea.echo', 'com.handcent.app.nextsms'))) {
						foreach ($nLines as $idx => $textLine) {
							if (trim($textLine) == '') {
								continue;
							}
							$noteSmsList[$smsSource] = $textLine;
						}
					}
					elseif ($nLines && !in_array($nPkg, array('com.textra', 'com.truecaller', 'com.mysms.android.sms', 'com.calea.echo'))) {
						$nums = $nText ? explode(',', $nText) : array();
						foreach ($nLines as $idx => $textLine) {
							$textLine = preg_replace('/\x{FEFF}/u', '', trim($textLine));
							$found    = false;
							if ($nums) {
								foreach ($nums as $num) {
									if (trim($num) == '' || mb_strpos($textLine, trim($num)) !== 0) {
										continue;
									}
									$sNum = PushData::stripSmsTitle($num);
									if (isset(PushData::$blackListSms[$sNum])) {
										$found = true;
										break;
									}
									$msgText = trim(mb_substr($textLine, mb_strlen($num)));
									$found   = true;
									if (empty($msgText)) {
										continue;
									}
									$noteSmsList["$nPkg:$sNum"] = $msgText;
									break;
								}
							}
							else {
								$textLine = str_replace('‬: ', '  ', $textLine);
								$spacePos = mb_strpos($textLine, '  ');
								if (!$spacePos) {
									$this->addError($jKey, $notificationData, "not detected num for '$textLine'");
									continue;
								}
								$num = PushData::stripSmsTitle(mb_substr($textLine, 0, $spacePos));
								if (isset(PushData::$blackListSms[$num])) {
									continue;
								}
								$noteSmsList["$nPkg:$num"] = trim(mb_substr($textLine, $spacePos + 2));
								$found                     = true;
							}
							$found ? '' : $this->addError($jKey, $notificationData, "Number in 'text' not found in line '$textLine'");
						}
					}
					elseif ($nSmsTitle && $nText) { // Иногда бывают "пустые" уведомления от СМС.
						$noteSmsList[$smsSource] = $nText;
					}
				}
				/**
				 * У банковских приложений title приклеиваем к тексту, там полезная информация для трат
				 */
				elseif ($nText || $nTitle) {
					$noteSmsList[$nPkg] = "$nText $nTitle";
				}

				/**
				 * Шаг 2: Выкидываем дубли (то, что равно предыдущему сообщению) и сохраняем НОВЫЕ
				 * сообщения, содержащиеся в данном уведомлении.
				 */
				$isAnyMessage = false;
				foreach ($noteSmsList as $source => $message) {
					if (empty(trim($message))) {
						$this->addError($jKey, $notificationData, "empty message from '$source'");
						$isAnyMessage = true;
						continue;
					}
					if (empty(trim($source))) {
						$this->addError($jKey, $notificationData, "empty source for message '$message'");
						$isAnyMessage = true;
						continue;
					}
					$source      = mb_substr($source, 0, 64);
					$message     = mb_substr($message, 0, 1024);
					$message     = preg_replace('/\x{FEFF}/u', '', $message);
					$lastMessage = empty(self::$lastMsgList[$source]) ? $this->getLastPush($source) : self::$lastMsgList[$source]['msg'];
					if ($lastMessage === $message) {
						$isKostyl = false;
						/**
						 * Если разница по времени между одинаковыми сообщениями более 3 часов - считаем, что это разные уведомления,
						 * если только не определён в уведомлениях одинаковый баланс
						 */
						$firstBal  = '';
						$firstAcc  = '';
						$firstCur  = '';
						$firstType = '';
						ImportRule::parseSumAndCurrency(
							str_replace(array("\n", "\r"), ' ', $message), $firstCur, $firstType, $firstBal, $firstAcc
						);
						if (floor(($nDate - self::$lastMsgList[$source]['date']) / 1000) > 3600 * 3 && !$firstBal) {
							$isKostyl = true;
						}
						/**
						 * Запрет фильтрации дублей для некоторых пользователей
						 */
						User::id() == '1000000284951' && $source == 'ru.nspk.mirpay' ? $isKostyl = true : '';
						User::id() == '1000000072571' && $source == 'com.google.android.gms' ? $isKostyl = true : '';
						if (!$isKostyl) {
							continue;
						}
					}
					/**
					 * Отсеиваем дубли так же и по дате, она с точностью до мсек.
					 */
					$isSameDate = DAO::getCount("
						SELECT id FROM push_history WHERE user_id = $1 AND date = $2 AND message = $3 LIMIT 1",
						TRUE, array(User::id(), $nDate, $message)
					);
					if ($isSameDate) {
						continue;
					}
					$nextIdVal = DB::query("SELECT nextval('push_history_id_seq')");
					$nextIdVal = pg_fetch_result($nextIdVal, 0, 0);
					DB::query("INSERT INTO push_history (id, user_id, source, message, date) VALUES ($5, $1, $2, $3, $4)",
						array(User::id(), mb_strtolower(trim($source)), $message, $nDate, $nextIdVal)
					);

					ArrayUtils::push($groupedSmsList[$source], str_replace(array("\n", "\r", '[', ']'), array(' ', '', '(', ')'), $message));
					ArrayUtils::push($smsListHistIdx[$source], $nHistId);
					ArrayUtils::push($smsPushHistIdx[$source], $nextIdVal);
					$isAnyMessage = true;
				}

				/**
				 * Если в уведомлении не нашли подходящего для распознавания текста - помечаем как шлак
				 */
				if (!$isAnyMessage) {
					ArrayUtils::push($groupedSmsList['DUP'], 'DUP: [d]');
					ArrayUtils::push($smsListHistIdx['DUP'], $nHistId);
					ArrayUtils::push($smsPushHistIdx['DUP'], '1');
				}
			}

			foreach ($groupedSmsList as $source => $smsList) {
				/**
				 * Поля по умолчанию определяем сами для каждого источника. Сначала смотрим нет ли полей по
				 * умолчанию от СМС настроек. Временно. Когда СМС "забудутся" - удалить оставить только
				 * автоопределение на основе статистики из push_history
				 */
				$srcConcat = explode(':', mb_strtolower($source));
				$srcConcat = trim($srcConcat[count($srcConcat) > 1 ? 1 : 0]);
				$defParams = PushData::getDefParams($srcConcat, $smsDefaults);
				$sourceId  = DAO::getCount("SELECT id FROM cluster_place_source WHERE $1 = ANY(alias)", true, array($srcConcat));

				/**
				 * Название источника приклеиваем в конец строки. Это тоже может быть использовано для правил.
				 * А так же для логики стоп-слов на клиенте.
				 */
				$smsTextPref = $userSrcPref.str_replace('+', '', $srcConcat); // Плюсик может привести к неверному срабатыванию
				foreach ($smsList as $key => $smsText) {
					/**
					 * Кастомные костыли для форматов некоторых приложений
					 */
					if ('co.uk.getmondo' == $source) {
						$smsText = preg_replace('/spent.*today/', 'spent ... today', $smsText);
						$smsText = preg_replace('/You have.*left in your/', 'You have ... left in your', $smsText);
					}
					$smsList[$key] = $smsText.$smsTextPref;
				}

				/**
				 * Скармливаем парсеру полученный список сообщений. Уже без дублей. Сообщения сгруппированы по источникам.
				 */
				$parsedArray = ImportRule::parse($smsList, User::id(), $defParams[0], $defParams[1], $defParams[2], $defParams[3]);

				/**
				 * Обработка результата парсинга
				 */
				foreach ($parsedArray as $key => $retItem) {
					is_array($retItem) ? '' : $retItem = array();
					/**
					 * Все не распознанные операции НЕ показываем в истории у пользователя, т.к.
					 * там будет очень много шлака.
					 */
					empty($retItem) ? $retItem['type'] = (string)ImportRule::ACTION_DUP : '';

					$retItem['hist_id']             = (String)$smsListHistIdx[$source][$key];
					$retItem['server_push_hist_id'] = (String)$smsPushHistIdx[$source][$key];
					$retItem['src_text']            = $smsList[$key];

					/**
					 * Если дата не распозналась из текста СМС, подсталяем дату уведомления иначе
					 * может не сработать отсев дублей, если обработка СМС идёт с задержкой
					 */
					if (ArrayUtils::safeGet($retItem, 'is_defaultdate') == 'true') {
						$retItem['date'] = DAO::getCount(
							"SELECT to_timestamp(date/1000)::timestamp without time zone FROM push_history WHERE id = $1",
							false, array($retItem['server_push_hist_id'])
						);
					}
					/**
					 * Если дата определена из текста и она очень давнишняя или очень будущая (больше 32 дней)
					 * это вероятно ложная операция, которая поломает у пользователя (ретро/пер)спективу.
					 * Экспериментально запрещаем.
					 */
					elseif (isset($retItem['date']) && abs(strtotime($retItem['date']) - time()) > 3600 * 24 * 32) {
						$retItem['type'] = (string)ImportRule::ACTION_NONE;
//						Debug::mail("Invalid date testing: $login\n"
//							."smsText = ".print_r($smsList[$key], true)."\n"
//							."retItem = ".print_r($retItem, true)."\n"
//						);
					}

					/**
					 * Используем белые списки источников сообщений. Если операция распозналась не по правилу пользователя,
					 * но источника у нас ещё нет - игнорируем её и шлём нам для анализа (может надо добавить источник).
					 */
					if ($retItem['type'] > ImportRule::ACTION_NONE) {
						$debugPref = '';
						if (!$sourceId && empty($retItem['is_user_rule'])) {
							$debugPref = 'NEW SOURCE';
							/**
							 * Не обращаем внимание на некоторые явно ненужные уведомления
							 */
							$nSrcIgnores = array('br.eti.fml.satoshi', 'заказ', 'Оплата по', 'com.ifttt.ifttt', 'dromfilter', 'PLASHTANE', 'билайн', 'Здравствуйте', 'ru.yandex.uber_by', 'Starbucks', '+7 ', 'platformofd', 'BTC', 'Isurio', 'бонус', 'лицевой', 'счет', 'кредит', 'подарок', 'курс');
							foreach ($nSrcIgnores as $ignoreStr) {
								if (mb_stripos($retItem['src_text'], $ignoreStr) !== false) {
									$debugPref = '';
									break;
								}
							}
							$retItem['type'] = (string)ImportRule::ACTION_NONE;
						}

						if ($debugPref) {
//			    			Debug::mail(
//								"Parsing result new source $debugPref\nsrcConcat: '$srcConcat'\n".
//								"Result ".print_r($retItem, true)
//							);
						}
					}

					if ($retItem['type'] == ImportRule::ACTION_PLANNED) {
						/**
						 * Подтверждаем планируемую операцию.
						 */
						$entered_id      = BudgetRecordPlanned::create($retItem['planned_id'])->doEnter($retItem);
						$retItem['type'] = (string)ImportRule::ACTION_NONE;

						/**
						 * Логика слияния с QR - такая же как и при создании операции из СМС
						 */
						$this->processQr($entered_id, $retItem['server_push_hist_id'], $retItem['rule_id'], null);
					}

					/**
					 * Костыль для сообщений об отмене/отказе. Если у пользователя сработало правило "ничего не делать",
					 * надо насильно операцию всё-таки создать, чтобы логика отмены сработала в момент сопоставления
					 * pushId => $recordId (см. код выше в этом же методе)
					 */
					$isCancelMsg = preg_match($cancelPattern, $retItem['src_text'], $wordMatch);
					if ($isCancelMsg && $retItem['type'] == ImportRule::ACTION_NONE && isset($retItem['sum'])) {
						$retItem['type']          = (string)v2_Budget::TYPE_INCOME;
						$retItem['sum']           = abs($retItem['sum']);
						$retItem['cat_id']        = (string)$defParams[2];
						$retItem['place_from_id'] = (string)$defParams[0];
					}

					isset($retItem['sum']) ? $retItem['sum'] = (string)$retItem['sum'] : '';
					$retArray[] = $retItem;

					/**
					 * Костыли для сервиса учёта времени.
					 */
					if (ArrayUtils::safeGet($retItem, 'rule_id') == '10314') {
						$workTimeArray = array(round(abs($retItem['sum']) / 100, 2), $retItem['date']);
					}
				}
			}

			/**
			 * Отладка
			 */
//			if (!empty($this->pushErrors)) {
//				Debug::mail("Parsing result for: $login\n"
//					."pushErrors = ".print_r($this->pushErrors, true)."\n"
//					."retArrayDebug = ".print_r($retArrayDebug, true)."\n"
//					."retArray = ".print_r($retArray, true)."\n"
//				);
//			}
//			if (!empty($floodPkg)) {
//				foreach ($pkgCount as $pkg => $floodItem) {
//					foreach ($floodItem as $key => $msec) {
//						$pkgCount[$pkg][$key] = $msec.", ".date("Y-m-d H:i:s", floor($msec / 1000));
//					}
//				}
//				Debug::mail("Parsing result for flooders: $login\n"
//					."floodPkg = ".print_r($floodPkg, true)."\n"
//					."pkgCount = ".print_r($pkgCount, true)."\n"
//					."retArray = ".print_r($retArray, true)."\n"
//				);
//			}

			/**
			 * Послать команду с сервера на отключение/включение алгоритма СМС
			 * Переход на пуши в Андроид приложении.
			 * Удалить уже где-нить через годик в 2020-09-1.
			 */
			if (self::isMoreVersion(array(2, 4, 5)) && !$isPushLink) {
				$retArray[] = array(
					'type'            => '-2',
					'is_autodetected' => 'true', //  Это у нас будет is_read_sms_disabled
				);
			}

			DB::transactionCommit();

			/**
			 * Костыли для сервиса учёта времени.
			 */
			if ($workTimeArray) {
				DB::disconnect();
				DB::setConnectParameter('base', 'worktime');
				/**
				 * Вносим оплату в учёт-времени
				 */
				DB::query("INSERT INTO worker_pay (worker_id, date, payed) VALUES ('33', $2, $1)", $workTimeArray);
				DB::disconnect();
				DB::setConnectParameter('base', null);
			}
		}
		catch (BaseException $e) {
			DB::transactionRollback();
			$conflictMsg = 'Для восстановления синхронизации, ';
			$conflictMsg .= ($apiId == 'vandr_api' ? "необходимо перезайти через меню\n'Ещё->Сменить пользователя'." : 'приложение необходимо переустановить.');
			if (mb_strpos($e->getMessage(), 'create_only_access') !== false) {
				throw new SoapFault('188', "Настройки прав доступа не позволяют изменять операции\n".$conflictMsg);
			}
			$accessErrorMsg = array('forbidden_type_', 'el_forbidden', 'not_owner_obj', 'readonly_access', 'not_owner_rec');
			foreach ($accessErrorMsg as $errorMsg) {
				if (strpos($e->getMessage(), $errorMsg) !== false) {
					throw new SoapFault('188', "Доступ к скрытым элементам запрещён\n".$conflictMsg);
				}
			}
			self::checkNameUniq($e->getMessage());
			throw $e;
		}

		return $retArray;
	}

	/**
	 * Если есть QR-овская операция на ту же сумму и дату плюс-минус MINUTE_DIFF мин, то
	 * СМС-очную операцию удаляем, взяв из неё:
	 * 1. Для базовой QR операции, когда чек не обработан или обработан, но позиция одна
	 *    - счёт, если budget_check.change_place_id = -1
	 *    - комментарий, к которому в конец приклеиваем QR тег [чек 12312]
	 *    - категорию, если распознанная при обработке QR менее приоритетна.
	 * 2. Для всех операций если их больше одной, когда чек обработан
	 *    - Счёт, если budget_check.change_place_id = -1
	 *    - Теги, если они есть
	 *    - Категорию, если распознанная при обработке QR менее приоритетна.
	 */
	private function processQr($recordId, $pushId, $ruleId, $secondId)
	{
		$recordObj = v2_BudgetRecord::create($recordId);
		$recFamId  = $recordObj->get(v2_BudgetRecord::FAMILY_ID);
		$srcName   = DAO::getCount("
	    	SELECT regexp_replace(source, '^(.*:)', '') FROM push_history WHERE id = $1",
			true, array($pushId)
		);
		if (!$recordObj->get(Business::ID) || !$srcName) {
			return false;
		}
		if ($recFamId != User::familyId()) {
			throw new BaseException("Record family id ($recFamId) not equal session family id");
		}

		$linkRecordId    = $recordId;
		$isDeleteSmsRec  = false;
		$curRecDate      = $recordObj->get(v2_BudgetRecord::DATE);
		$curRecComment   = $recordObj->get(v2_BudgetRecord::COMMENT);
		$nearCheckParams = DAO::fillArray("
			SELECT id, state, change_place_id FROM budget_check WHERE qr_sum = $2 AND user_id IN (
				SELECT id FROM users WHERE family_id = $1
			)
			AND abs(extract(epoch from (qr_date - '$curRecDate'::timestamp(0)))) < ".BudgetCheck::MINUTE_DIFF."*60
			ORDER BY abs(extract(epoch from (qr_date - '$curRecDate'::timestamp(0)))) LIMIT 1",
			null, array(User::familyId(), abs($recordObj->get(v2_BudgetRecord::SUM)))
		);

		/**
		 * Слияние с QR актуально только для расходной операции
		 */
		if (!$secondId && $recordObj->get(v2_BudgetRecord::SUM) < 0 && !empty($nearCheckParams)) {
			$tagIds       = array();
			$nearRecordId = null;
			$tagsString   = '';
			$chPlaceId    = null;
			$qrRecordList = v2_BudgetRecord::create()->daoMakeObjectList("
				SELECT * FROM budget_record WHERE budget_family_id = $1 AND id IN (
					SELECT record_id FROM budget_check_record WHERE check_id = $2
				)",
				false, array(User::familyId(), $nearCheckParams[0]['id'])
			);

			if ($qrRecordList) {
				$tagIds = DAO::fillArray(
					"DELETE FROM budget_record_tags WHERE budget_record_id = $1 RETURNING budget_tag_id", null,
					array($recordId)
				);
			}

			$firstComment = trim(preg_replace('/\[[^\]]+\]/u', '', $curRecComment));
			empty($firstComment) ? '' : $firstComment .= ' ';

			foreach ($qrRecordList as $qrRecord) {
				$nearRecordId = $qrRecord->get(Business::ID);
				/**
				 * Запоминать связку в push_history можно тоже только тогда, когда чек не разбит
				 * на множество позиций. В противном случае с какой позицией связывать пушку и зачем?
				 */
				$linkRecordId = count($qrRecordList) == 1 ? $nearRecordId : null;
				$qrComment    = '';

				if ($nearCheckParams[0]['change_place_id'] == -1) {
					$chPlaceId = $recordObj->get(v2_BudgetRecord::ACCOUNT_ID);
					$qrRecord->set(v2_BudgetRecord::ACCOUNT_ID, $chPlaceId);
				}

				/**
				 * Переклеиваем связки тегов
				 */
				$tmpStr = '';
				foreach ($tagIds as $tagId) {
					$tgParams = array($nearRecordId, $tagId['budget_tag_id']);
					$isExist  = DAO::getCount("SELECT budget_tag_id FROM budget_record_tags WHERE budget_record_id = $1 AND budget_tag_id = $2", true, $tgParams);
					if (!$isExist) {
						DB::query("INSERT INTO budget_record_tags(budget_record_id, budget_tag_id) VALUES ($1, $2)", $tgParams);
					}
					if ($tagsString) {
						continue;
					}
					$tagName = DAO::getCount(
						"SELECT name FROM budget_tags WHERE id = $1",
						false, array($tagId['budget_tag_id'])
					);
					$tmpStr  .= " [$tagName]";
				}
				$tagsString = $tmpStr;
				empty($tagsString) ? '' : $tagsString .= ' ';

				/**
				 * Категорией по умолчанию считаем категорию заданную правилом обработка СМС
				 */
				$defCategoryId = $recordObj->get(v2_BudgetRecord::OBJECT_ID);

				if (count($qrRecordList) == 1) {
					/**
					 * Чек либо не обработан, либо обработан, но позиция одна.
					 * Слияние QR операции с операцией, подтверждающей планируемую,
					 * актуально только в этом случае.
					 */
					DB::query(
						"UPDATE budget_record_planned SET entered_id = $1 WHERE entered_id = $2",
						array($nearRecordId, $recordId)
					);
				}
				else if ($qrRecord->get(v2_BudgetRecord::OBJECT_ID) != Family::getFirstCategoryId()) {
					$defCategoryId = $qrRecord->get(v2_BudgetRecord::OBJECT_ID);
				}

				$qrComment    = $firstComment.$tagsString.$qrRecord->get(v2_BudgetRecord::COMMENT);
				$firstComment = '';

				/**
				 * @todo
				 * Когда заработает алгоритм определения категорий при обработке позиций QR, тут
				 * нужно делать сравнение приоритетов "правило обработки" > "алгоритм QR" и только
				 * тогда менять категорию. Ввести понятие приоритета.
				 *  - 11 авг 2020. Категории по QR определились хорошо, а после поступления СМС
				 *    они все "затёрлись" категорией из СМС, получилось плохо. Пробуем пока, если
				 *    позиция не одна, затирать только категорию "Без категории", отдавая приоритет
				 *    алгоритму определения категорий из QR.
				 */
				$qrRecord->
				set(v2_BudgetRecord::OBJECT_ID, $defCategoryId)->
				set(v2_BudgetRecord::COMMENT, $qrComment)->
				daoSave();
			}
			/**
			 * Если "ближайшей" QR-овской операцией оказалась та же самая операция, то поля меняем,
			 * но удалять не удаляем.
			 */
			if ($nearRecordId != $recordId) {
				$isDeleteSmsRec = true;
				DB::query("DELETE FROM budget_record WHERE id = $1", array($recordId));
			}
			if ($chPlaceId) {
				DB::query("UPDATE budget_check SET change_place_id = $1 WHERE id = $2", array($chPlaceId, $nearCheckParams[0]['id']));
			}
		}

		/**
		 * Если созданная операция имеет счёт отличный от определённого по умолчанию, то считаем его уже
		 * изменённым и полезным для будущих сопоставлений и кластеризации. Значит сработало пользовательское правило.
		 */
		if ($linkRecordId) {
			$smsDefaults = BudgetSettings::getValue('android_sms_to_push');
			$smsDefaults = json_decode($smsDefaults, true);
			$linkParam   = array($linkRecordId, $pushId, $ruleId ? $ruleId : null);
			$sqlSet      = '';
			$defParams   = PushData::getDefParams($srcName, $smsDefaults, false);
			$accountId   = $recordObj->get(v2_BudgetRecord::ACCOUNT_ID);
			if ($accountId != $defParams[0]) {
				$linkParam[] = $accountId;
				$sqlSet      = ', changed_account_id = $4';
			}
			/**
			 * Проверить, трата могла быть уже удалена.
			 */
			if (!DAO::getCount("SELECT id FROM budget_record WHERE id = $1", true, array($linkRecordId))) {
				$linkParam[0] = null;
			}
			DB::query("UPDATE push_history SET rule_id = $3, record_id = $1$sqlSet WHERE id = $2", $linkParam);
		}
		return $isDeleteSmsRec;
	}

	private function addError($lineId, $srcData, $errText)
	{
		empty($this->pushErrors[$lineId]) ? $this->pushErrors[$lineId] = array('src' => $srcData) : '';
		$this->pushErrors[$lineId][] = $errText;
	}

	private function getLastPush($srcName = null)
	{
		$param = array(User::id());
		/**
		 * Некоторые строки в БД попадают с BOM. Нужно его вырезать, чтобы сравнение строк работало
		 * корректно. Иначе у пользователей будут дубли.
		 */
		$sql = "SELECT regexp_replace(message, '[\uFEFF]', '') AS msg, date FROM push_history WHERE user_id = $1";
		if ($srcName) {
			$srcName = mb_substr($srcName, 0, 64);
			$param[] = mb_strtolower(trim($srcName));
			$sql     .= " AND source = $2";
		}

		$sql     .= " ORDER BY id DESC LIMIT 1";
		$msgDate = DAO::fillArray($sql, null, $param);
		if (isset($msgDate[0])) {
			self::$lastMsgList[$srcName] = array('msg' => $msgDate[0]['msg'], 'date' => $msgDate[0]['date']);
			return $msgDate[0]['msg'];
		}
		return '';
	}

	/**
	 * Try to parse text data as records; [def..] default field values if no rules detected; [list] => array (indexes must be 0,1,2...N) of strings to parse (UTF8); Returns the array of array - data for records;
	 *
	 * @param string $apiId
	 * @param string $login
	 * @param string $pass
	 * @param string $defPlaceFromId
	 * @param string $defCatId
	 * @param string $defSrcId
	 * @param string $defPlaceToId
	 * @param array $list
	 * @return array
	 */
	public function parseTextData($apiId, $login, $pass, $defPlaceFromId, $defCatId, $defSrcId, $defPlaceToId, $list)
	{
		/**
		 * Функция платная
		 */
		try {
			$this->getAccessStatus($apiId, $login, $pass);
		}
		catch (SoapFault $e) {
			if ($e->faultcode != '87') { // trial
				throw $e;
			}
		}

		/**
		 * Проверки
		 */
		$this->checkEmptyArray($list, '116', 'String list');
		$this->checkNumeric($defPlaceFromId, 'defPlaceFromId');
		$this->checkNumeric($defCatId, 'defCatId');
		$this->checkNumeric($defSrcId, 'defSrcId');
		$this->checkNumeric($defPlaceToId, 'defPlaceToId');

		foreach ($list as $key => $line) {
			$list[$key] = str_replace(array("\n", "\r", '[', ']'), array(' ', '', '(', ')'), $line);
		}

		$isAnyQr  = false;
		$retArray = ImportRule::parse($list, User::id(), $defPlaceFromId, $defCatId, $defSrcId, $defPlaceToId);

		/**
		 * Надо проверить нет ли специальных планируемых операций в списке.
		 * Подтерждаем их на этом этапе, а клиентам отдаёт "Игнор".
		 */
		foreach ($retArray as $key => $retItem) {
			if ($retItem && isset($retItem['is_qr'])) {
				$isAnyQr = true;
			}
			if ($retItem && $retItem['type'] == ImportRule::ACTION_PLANNED) {
				// Подтверждаем планируемую операцию.
				BudgetRecordPlanned::create($retItem['planned_id'])->doEnter($retItem);

				$retArray[$key] = array(
					'type' => (string)ImportRule::ACTION_NONE,
				);
			}
		}

		/**
		 * Для анализа копим базу СМС, чтобы определять стандартные паттерны
		 */
		$version = self::getVersion();
		$version = is_array($version) ? implode('.', $version) : '';
		if (!defined('IS_SERVER') && ImportRule::IS_COPY_SMS) {
			foreach ($list as $sms) {
				$sms = mb_substr($version.' sms: '.$sms, 0, 512);
				DB::query("INSERT INTO sms_patterns (name,user_id) VALUES ($1, $2)", array($sms, User::id()));
			}
		}

		/**
		 * Отладка, удалить
		 */
//		if ($isAnyQr) {
//			Debug::mail(print_r($list, true)."\n\n".print_r($retArray, true));
//		}

		return $retArray;
	}

	private function getRefId()
	{
		$refId = ArrayUtils::safeGet($_SERVER, 'HTTP_USER_AGENT');
		$refId = explode(',', $refId);
		$refId = explode(' ', ArrayUtils::safeGet($refId, 2));
		if (ArrayUtils::safeGet($refId, 1) == 'ref.') {
			$refId = ArrayUtils::safeGet($refId, 2);
			$refId = (!$refId ? null : $refId);
		}
		else {
			$refId = null;
		}
		return $refId;
	}

	public static function getAppId()
	{
		$appId = ArrayUtils::safeGet($_SERVER, 'HTTP_USER_AGENT');
		$appId = explode(',', $appId);
		$appId = explode(' ', ArrayUtils::safeGet($appId, 1));
		if (ArrayUtils::safeGet($appId, 1) == 'app.') {
			$appId = ArrayUtils::safeGet($appId, 2);
			$appId = (!$appId ? null : $appId);
		}
		else {
			$appId = null;
		}
		return $appId;
	}

	public static function getVersion()
	{
		$versionString = ArrayUtils::safeGet($_SERVER, 'HTTP_USER_AGENT');
		$versionString = explode(',', $versionString);
		$versionString = explode(' ', ArrayUtils::safeGet($versionString, 0));
		return explode('.', ArrayUtils::safeGet($versionString, 1));
	}

	public static function isMoreVersion($versionArray)
	{
		$versionString = self::getVersion();

		if (count($versionString) >= 3 && (
				ArrayUtils::safeGet($versionString, 0) > $versionArray[0] ||
				(ArrayUtils::safeGet($versionString, 0) == $versionArray[0] && ArrayUtils::safeGet($versionString, 1) > $versionArray[1]) ||
				(ArrayUtils::safeGet($versionString, 0) == $versionArray[0] && ArrayUtils::safeGet($versionString, 1) == $versionArray[1] &&
					ArrayUtils::safeGet($versionString, 2) >= $versionArray[2]
				)
			)
		) {
			return true;
		}
		return false;
	}

	public static function getDoubleId(
		$firstRecId, $firstMessage, $firstSource, $firstRuleId,
		$secondRecId, $secondMessage, $secondSource, $secondRuleId
	)
	{
		$logPrefix     = '';
		$returnId      = null;
		$firstCur      = '';
		$firstType     = '';
		$firstBal      = '';
		$firstAcc      = '';
		$secondCur     = '';
		$secondType    = '';
		$secondBal     = '';
		$secondAcc     = '';
		$isSentMail    = false;
		$firstMessage  = str_replace(array("\n", "\r"), ' ', $firstMessage);
		$secondMessage = str_replace(array("\n", "\r"), ' ', $secondMessage);
		ImportRule::parseSumAndCurrency($firstMessage, $firstCur, $firstType, $firstBal, $firstAcc);
		ImportRule::parseSumAndCurrency($secondMessage, $secondCur, $secondType, $secondBal, $secondAcc);

		try {
			/**
			 * Если валюты определены и не совпадают - это не дубли
			 */
			if ($firstCur && $secondCur && $firstCur != $secondCur) {
				$logPrefix = "Валюты: '$firstCur' != '$secondCur'";
				return null;
			}

			/**
			 * Если определены и не совпадают балансы, а счёт не определён в обоих или совпадает,
			 * это тоже не дубли.
			 */
			if (is_numeric($secondBal) && is_numeric($firstBal) && $firstBal != $secondBal) {
//				$logPrefix = "Баланс: '$firstBal' != '$secondBal'";
				return null;
			}

			/**
			 * Если тексты сообщений одинаковы с точностью до символа и разница между ID-ами небольшая - дубли
			 * 'Сумма 346,07 ₽. Карта Mir  • • • • 1696 Оплата покупки':216049837
			 * 'Сумма 346,07 ₽. Карта Mir  • • • • 1696 Оплата покупки':216049840
			 */
			$isDouble = false;
			if ($firstMessage == $secondMessage && abs($firstRecId - $secondRecId) < 50) {
				$isDouble = true;
			}

			/**
			 * Если счёт один и тот же и в обоих уведомлениях есть слово "перевод" и оба они от одного источника
			 * и похожи друг на друга (один формат), то это скорее всего разные переводы на одну и ту же сумму.
			 */
			$isBalEqual = (is_numeric($firstBal) && $firstBal == $secondBal);
			if (
				$secondAcc && $firstAcc == $secondAcc && $firstSource == $secondSource && !$isDouble &&
				mb_substr($firstMessage, 0, 4) == mb_substr($secondMessage, 0, 4) && !$isBalEqual &&
				mb_stripos($firstMessage, 'перевод') !== false && mb_stripos($secondMessage, 'перевод') !== false
			) {
				$logPrefix = "Разные переводы";
				return null;
			}

			/**
			 * Если балансы и счета определены и совпадают - однозначно дубли
			 */
			if (is_numeric($secondBal) && $firstBal == $secondBal && $secondAcc && $firstAcc == $secondAcc) {
				$isDouble = true;
			}

			/**
			 * Так же НЕ считаем дублями очень похожие сообщения с одинаковыми суммами, но разными счетами и балансами
			 * *6557 Списание 99p ПРОМСВЯЗЬБАНК Баланс 70'850.23p psbank
			 * *6556 Списание 99p ПРОМСВЯЗЬБАНК Баланс 12'100.98p psbank
			 *
			 * Покупка на 65 RUB, карта *3823  Доступно 501 874.51 RUB Метро Санкт-Петербург com.idamob.tinkoff.android
			 * Покупка на 65 RUB, карта *0589  Доступно 501 939.51 RUB Метро Санкт-Петербург com.idamob.tinkoff.android
			 *
			 * Но при этом, если балансы оба определены и равны - считаем дублями.
			 * Покупка на 404 ₽. Ozon. Баланс 1 953.64 ₽ Ozon Банк
			 * Покупка на 404 ₽. Ozon. Баланс 1 953.64 ₽ Ozon Банк
			 *
			 * Так же, если один из балансов определён, а другой нет - считаем дублями.
			 * Покупка на 7 005 ₽. Ozon. Баланс 216 ₽ Ozon Банк
			 * Покупка на 7 005 ₽. Ozon Ozon Банк
			 *
			 * Так же, если оба баланса не определены, и хотя бы один счёт не определён - считаем дублями
			 * Вклад на 53 715.21 ₽ продлен до 5 октября 2024 года Пролонгация вклада
			 * Вклад на 53 715.21 ₽ продлен до 5 октября 2024 года Пролонгация вклада
			 */
			$dblPattern = '/(списание|покупка|Поступление|Зачисление|Баланс|Остаток|Оплата|Пролонгация)/iu';
			$oneMsg     = preg_match($dblPattern, $firstMessage, $wordMatch1, PREG_OFFSET_CAPTURE);
			$otherMsg   = preg_match($dblPattern, $secondMessage, $wordMatch2, PREG_OFFSET_CAPTURE);
			$isSimilar  = false;
			if ($oneMsg && $oneMsg == $otherMsg && $wordMatch1[0][1] == $wordMatch2[0][1]) {
				$isSimilar = true;
				if (is_numeric($secondBal) && $firstBal == $secondBal) {
					$isDouble = true;
				}
				if ((is_numeric($secondBal) && !is_numeric($firstBal)) || (!is_numeric($secondBal) && is_numeric($firstBal))) {
					$isDouble = true;
				}
				if (!is_numeric($secondBal) && !is_numeric($firstBal) && (!$firstAcc || !$secondAcc)) {
					$isDouble = true;
				}
			}
			if ($isSimilar && !$isDouble) {
				$logPrefix = "Похожие сообщения: '{$wordMatch1[0][0]}'";
				return null;
			}

			/**
			 * Если дошли до сюда, значит какое-то из уведомлений уже точно надо считать дублем.
			 * По умолчанию считаем дублем текущее уведомление $firstMessage
			 */
			$returnId = $firstRecId;

			/**
			 * Если оба уведомления определены правилом или оба без правил, но одно из них от gms/splay - считаем его дублем
			 */
			$gmsSpay    = array('com.google.android.gms', 'com.samsung.android.spay');
			$isBothRule = $firstRuleId && $secondRuleId ? true : false;
			if ($isBothRule || (!$firstRuleId && !$secondRuleId)) {
				if (in_array($secondSource, $gmsSpay)) {
					$returnId = $secondRecId;
				}
				/**
				 * Если обе операции определены правилами, надо выбрать из них ту, где категория не "Без категории"
				 */
				elseif ($isBothRule) {
					$firstRecObj   = v2_BudgetRecord::create($firstRecId);
					$defCategoryId = Family::getFirstCategoryId();
					if ($firstRecObj->get(v2_BudgetRecord::OBJECT_ID) == $defCategoryId) {
						$returnId = $secondRecId;
					}
				}
			}
			/**
			 * Если какое-то одно из уведомлений определено правилом, оставляем его и дублем считаем другое
			 */
			elseif ($firstRuleId) {
				$returnId = $secondRecId;
			}

			return $returnId;
		}
		finally {
			$returnId ? $logPrefix = "Удаляем: '$returnId'" : '';
			if ($logPrefix) {
				$logMsg  = "$logPrefix\n1: '$firstMessage':$firstRecId\n2: '$secondMessage':$secondRecId\n";
				$logFile = $returnId ? "double" : "notdouble";
				Internal::toLog("\n$logMsg", "/var/log/drebedengi/web.sms.new.$logFile.log");
				$isSentMail ? Debug::mail($logMsg) : '';
			}
		}
	}

	public static function checkNameUniq($errMsg, $isShortMsg = false, $isException = true)
	{
		$mPos = mb_strpos($errMsg, 'el_name_uniq');
		if ($mPos == false) {
			return '';
		}
		$errMsg     = mb_substr($errMsg, $mPos);
		$elData     = explode(':', $errMsg);
		$elType     = $elData[2];
		$elName     = $elData[1];
		$elTypeName = 'Элемент не уникален';
		switch ($elType) {
			case v2_Budget::TYPE_INCOME:
				$elTypeName = "Источник дохода '$elName' не уникален";
				break;
			case v2_Budget::TYPE_WASTE:
				$elTypeName = "Категория расхода '$elName' не уникальна";
				break;
			case v2_Budget::TYPE_MOVE:
				$elTypeName = $elData[3] == 't' ? 'Долговой счёт' : 'Счёт';
				$elTypeName .= " '$elName' не уникален";
				break;
			case v2_Budget::TYPE_GROUP_PLACE:
				$elTypeName = "Название группы '$elName' не уникально";
				break;
		}
		$errMsg = "$elTypeName.\nИзмените название";
		if ($isShortMsg) {
			$errMsg = "'$elName' не уникален.<br>Измените название";
		}
		if ($isException) {
			$errMsg .= ' в справочниках веб-версии.';
			throw new SoapFault('188', $errMsg);
		}
		if (!$isShortMsg) {
			$errMsg .= " пожалуйста.";
		}
		return $errMsg;
	}
}

?>
